/*
 * SmoothB9.cpp
 * Dennis Lew (dennis_L76@yahoo.ca)
 *
 * ROBOT     : SmoothB4
 * DATA FILE : None
 *
 * Uses the OOP structure, call 'getSBv4r0Instance()' to get a
 * new instance.  All instances of SmoothB4 in a race will share
 * the same path data.
 *    
 * If you want to use any part of this code in your own robot, or
 * are curious about anything I've done here, just let me know.
 *
 * <CHANGES>
 * v1.4: -fixed NULL pointer bug when trying to load data
 *       -improved collision detection
 *       -robot will not actively try to pass other cars (UNDER DEVELOPMENT)
 *       -robot should be able to reliably finish races now
 *
 * v1.5: -bugs introduced by earlier improvements have been fixed
 *       -robot runs better on tracks with no data
 *
 * v1.6: -complete rewrite of the anti-collision code (MUCH BETTER)
 *
 * v2.0: -complete rewrite of the path creation routine (MUCH BETTER, BUT NEEDS MORE DEVELOPMENT)
 *       -new data file format (SMALLER THAN OLD VERSION)
 *       -improved control routine
 *       -improved anti-collision code
 *       -functional passing code (FINALLY!)
 *
 * v2.1: -changes for RARS v0.9
 *
 * v2.2: -defaults to using a constant string for its name
 *       -better control parameters for surfaces 0 and 2
 *       -small adjustments to anti-collision code
 *
 * v2.3: -rewrite of the path creation routine (WORKS ON ALL TRACKS!)
 *       -this routine takes longer than the last one:
 *           Avg Time (P200MMX): v2.2 =  0.3s
 *                               v2.3 =  1.3s
 *           Max Time (P200MMX): v2.2 = 17.7s
 *           (kimla04.trk)       v2.3 = 20.5s
 *           longrand.trk Time : v2.2 = 1min (cannot complete lap)
 *                               v2.3 = 9min
 *       -in most cases, this routine will create a faster path
 *       -small adjustments to the control routine, for better stability
 *
 * v3.0: -added elliptical components to the path for better performance
 *       -the routine is not efficient, it adds about 0.7s/turn to the
 *        path creation routine (STILL NEEDS WORK)
 *       -made robot more aggresive in traffic
 *
 * v3.1: -various improvements to the anti-collision code 
 *        (back-ported to v2.3)
 *       -removed an unused member in the SBv3r0 class.
 *       -replaced fuel and damage handling code.
 *
 * v3.2: -new and improved anti-collision code
 *
 * v4.0: -new path creation routine is based on K1999/K2000 routine
 *       -new and improved control routine
 *       -new anti-collision code (back ported to v3.2)
 *
 * v4.1: -new and improved passing routine
 *       -bunch of changes to make robot work better in traffic
 *       -a few more changes to make robot work better in traffic (as good as its gonna get!)
 *       -bug fixes
 *
 * v4.2: -changes to path creation routine
 *       -more bug fixes
 */
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "os.h"
#include "car.h"
#include "track.h"

#ifndef SB4_DLL
// Set the following macro to the number of SmoothB
// robots that will be used.
#define SB4_CARS 2
// Set the following macro to 1 if the default name
// will be used, or 0 if not
#define SB4_USE_DEFAULT_NAME 1

// CONSTANTS
static const char *SB4_AUTHOR           = "Dennis Lew (dennis_L76@yahoo.ca)",
                  *SB4_DESCRIPTION      = "v4.2: Oct 11, 2003",
                  *SB4_NAME[1+SB4_CARS] = {"SmoothB4",    // the default name
                                           "SBv4r2-1",    // names for the other robots that may be used
                                           "SBv4r2-2"},
                  *SB4_BITMAP2D         = "car_gray_green",
                  *SB4_MODEL3D          = NULL;
static const int   SB4_NOSE             = oLIGHTGRAY,
                   SB4_TAIL             = oGREEN;
#endif
static const char *SB4_DATA_FILE  = "no_data_yet.dat";
static const float MY_PI          =    3.141592652f,
                   MY_PI21        =    2.0f * MY_PI,
                   MY_PI12        =    0.5f * MY_PI,
                   SB4_BRK[3]     =   {0.902f,
                                       1.0f,
                                       1.0f},
                   SB4_SPD[3]     =   {0.965f,
                                       0.995f,
                                       0.998f},
                   SB4_MAX_SPD    =  400.0f,
                   SB4_FC         =    0.000315f,  // fuel / ft
                   SB4_IN_WID     =    1.5f,
                   SB4_OUT_WID    =    5.0f,
                   SB4_MIN_CHANGE =    0.025f,
                   SB4_STEP       =    0.2f,
                   SB4_MIN_LEN    =   10.0f,
                   SB4_MIN_CURV   =    0.0001f;
static const int   SB4_MAX_RNDS   = 1000;

// GLOBAL VARIABLES
static float MAX_ACCEL, MAX_PWR, BRK2, SPD;

// STATIC FUNCTIONS
//-----START spd_func.h-----//
#define SB4_car_mass(fuel) \
   (M + (fuel/g))
#define SB4_drag(damage) \
   (DRAG_CON * (1.0 + ((double)(damage << 1) / (double)MAX_DAMAGE)))
#define SB4_drag_accel(fuel, damage) \
   (float)(SB4_drag(damage) / SB4_car_mass(fuel))

float  SB4_crn_spd(float curv);
float  SB4_brkdist(float vi, float vf, float max_a, float drag_accel);
float  SB4_vf(float vi, float brkdist, float max_a, float drag_accel);
float  SB4_slip(float u);
double SB4_fuel_needed(Stage s, int lapsleft, float fpl, float cur_fuel);
//-----END spd_func.h-----//

// CLASSES
//-----START Primitive.h-----//
class SB4Primitive
{
public:
   virtual float dist2Point(const SB4Primitive &chk_pt) const = 0;
   
protected:
   // CONSTRUCTOR
   SB4Primitive() {}
};
//-----END Primitive.h-----//

//-----START Vector.h-----//
class SB4Vector : public SB4Primitive
{
public:
   // MEMBERS
   float x, y;

   // CONSTRUCTORS
   SB4Vector()
      : x(0.0f), y(0.0f)
   {}
   SB4Vector(float vx, float vy)
      : x(vx), y(vy)
   {}
   SB4Vector(float angle)
      : x((float)cos((double)angle)), y((float)sin((double)angle))
   {}
   SB4Vector(const SB4Vector &v)
      : x(v.x), y(v.y)
   {}   
   // METHODS
   void  add(const SB4Vector &v);
   void  sub(const SB4Vector &v);
   float mag() const;
   float mag2() const;
   float dist2Point(const SB4Vector &chk_pt) const;
   float dist2Point2(const SB4Vector &chk_pt) const;
   float ang() const;
   float ang(const SB4Vector &v) const;
   void  scale(float factor);
   void  normalize();
   float dot(const SB4Vector &v) const;
   void  rotate(float angle);

   float dist2Point(const SB4Primitive &chk_pt) const
      {return dist2Point((const SB4Vector&)chk_pt);}
};
//-----END Vector.h-----//

//-----START Line.h-----//
class SB4Line : public SB4Primitive 
{
public:
   // MEMBERS
   SB4Vector pt;
   SB4Vector dir;

   // CONSTRUCTORS
   SB4Line()
      : pt(), dir()
   {}
   SB4Line(float px, float py, float dx, float dy)
      : pt(px, py), dir(dx, dy)
   { dir.normalize(); }
   SB4Line(float px, float py, float d_ang)
      : pt(px, py), dir(d_ang)
   {}
   SB4Line(const SB4Line &l)
      : pt(l.pt), dir(l.dir)
   {}
   // METHODS
   void  getPoint(SB4Vector &ret_pt, float dist) const;
   void  movePoint(float dist);
   float dist2Point(const SB4Vector &chk_pt) const;
   int   intersect(SB4Vector &ret_pt, const SB4Line &l) const;

   float dist2Point(const SB4Primitive &chk_pt) const
      {return dist2Point((const SB4Vector&)chk_pt);}
};
//-----END Line.h-----//

//-----START Circle.h-----//
class SB4Circle : public SB4Primitive
{
public:
   // MEMBERS
   SB4Vector c;
   float     R;

   // CONSTRUCTORS
   SB4Circle()
      : c(), R(0.0f)
   {}
   SB4Circle(float rad)
      : c(), R(rad)
   {}
   SB4Circle(float cx, float cy, float rad)
      : c(cx, cy), R(rad)
   {}
   SB4Circle(const SB4Circle &cc)
      : c(cc.c), R(cc.R)
   {}
   // METHODS
   float dist2Point(const SB4Vector &chk_pt) const;
   void  getPoint(SB4Vector &ret_pt, float angle) const;
   float ang(SB4Vector &pt) const;

   float dist2Point(const SB4Primitive &chk_pt) const
      {return dist2Point((const SB4Vector&)chk_pt);}
};
//-----END Circle.h-----//

//-----START TrkSeg.h-----//
class SB4TrkSeg
{
public:
   // MEMBERS
   float         Sdst, Fdst;  // Start and End locations (measured on centre line of trk)
   float         len, rad;
   SB4Primitive *left_side;
   int           Npts;
   
   // METHODS
   virtual void  get_norm(SB4Line &ret_n, float trk_dist) const = 0;
   virtual float get_trk_dist(const SB4Vector &chk_pt) const    = 0;

   void  getPoint(SB4Vector &ret_pt, float trk_dist, float to_lft) const;
   float to_left(const SB4Vector &chk_pt) const;
   int   same_dir(const SB4TrkSeg &seg) const;   

protected:
   // CONSTRUCTORS
   SB4TrkSeg()
      : Sdst(0.0f), Fdst(0.0f), len(0.0f), rad(0.0f),
        left_side(NULL), Npts(0)
   {}
};
//-----END TrkSeg.h-----//

//-----START TrkSeg_Straight.h-----//
class SB4TrkSeg_Straight : public SB4TrkSeg
{
public:
   // MEMBERS
   SB4Line ls;
   
   // CONSTRUCTORS
   SB4TrkSeg_Straight(segment *seg, float trk_dist);
   
   // METHODS
   void  get_norm(SB4Line &ret_n, float trk_dist) const;
   float get_trk_dist(const SB4Vector &chk_pt) const;
};
//-----END TrkSeg_Straight.h-----//

//-----START TrkSeg_Turn.h-----//
class SB4TrkSeg_Turn : public SB4TrkSeg
{
public:
   // MEMBERS
   SB4Circle ls;
   float     beg_ang, end_ang;
   int       left;
   
   // CONSTRUCTORS
   SB4TrkSeg_Turn(segment *seg, float wid, float trk_dist);
   
   // METHODS
   void  get_norm(SB4Line &ret_n, float trk_dist) const;
   float get_trk_dist(const SB4Vector &chk_pt) const;
};
//-----END TrkSeg_Turn.h-----//

//-----START Path.h-----//
class SB4Path
{
public:
   // CONSTRUCTOR
   SB4Path(int num_segs, SB4TrkSeg **seg, float trk_wid, const char *name);
   // DESTRUCTOR
   virtual ~SB4Path();

   // Adjusts this Path, to produce the best lap times
   virtual void adjust(float       in_w,     /*IN : min dist from inside to path*/
                       float       out_w,    /*IN : min dist from outside to path*/
                       float       trk_wid,  /*IN : track width*/
                       int         num_segs, /*IN : number of track segments*/
                       SB4TrkSeg **seg       /*IN : track segments*/
                      );

   /* Locates the given point, relative to this Path.
      Returns a non zero value, if a result is found, otherwise zero.
    */
   virtual int locatePoint(      int       *j,         /*OUT: a start point from which path distance is measured (can be NULL)*/
                                 float     *path_dist, /*OUT: path distance to the point*/
                                 float     *to_side,   /*OUT: +ve if point is right of path, -ve if left of path*/
                                 float     *offset,    /*OUT: offset from racing line*/
                           const SB4Vector &chk_pt,    /*IN : the point to be located*/
                                 SB4Line   &trk_norm,  /*IN : track normal at the given point*/
                                 int        seg_ID     /*IN : ID of the track segment*/
                          ) const;

   // Returns the curvature of this Path at the specified point.
   virtual float get_curv(int   *j,         /*IN : a start point from which path distance is measured*/
                          float *path_dist  /*IN : path distance to the point*/
                         ) const;
protected:
   int       *Spt, Npts, Nwgts;
   SB4Vector *pt;
   float     *ang, *len, *curv_wgt;

   // PROTECTED CONSTRUCTOR
   SB4Path(SB4Path &p) 
      : Spt(p.Spt), Npts(p.Npts), pt(p.pt), ang(p.ang), len(p.len), //spd(p.spd),
        Nwgts(0), curv_wgt(0)
   {}

   // PROTECTED METHODS
   void  getPoint(SB4Vector &ret_pt, SB4Vector &ret_tan,
                  int j, float path_dist) const;
   void  fill_ang();
   void  fill_len();
   void  loadData(const char *name);
   void  adjust_pts(int num_segs, SB4TrkSeg **seg);
   void  adjust_pts(float in_w, float out_w, float trk_wid,
                    int num_segs, SB4TrkSeg **seg, float *dir);

   static float get_ang(const SB4Vector &p1, const SB4Vector &p2, 
                        const SB4Vector &p3);
   static float get_len(const SB4Vector &p1, float a1, float k1,
                        const SB4Vector &p2, float a2, float k2);
   static float P(float t, float A, float B, float C, float D);
   static float dP(float t, float A, float B, float C);
   static float dP_mag(float t, float A, float B, float C, float D, float E);
   static void  get_center(SB4Vector &ret_pt, const SB4Vector &p1, 
                           const SB4Vector &p2, float R);
   static void  get_mid_point(SB4Vector &ret_pt, const SB4Vector &p1, 
                              const SB4Vector &p2, float R);
   static float curvature(const SB4Vector &p1, const SB4Vector &p2, 
                          const SB4Vector &p3);
   static int   movePoint(const SB4Vector &p1, SB4Vector &p2, 
                          const SB4Vector &p3, float curv, float w);
   static float adjustPoint(const SB4Vector &p1, SB4Vector &p2,
                            SB4Vector &p2b, const SB4Vector &p3,
                            float min, float max, 
                            int num_segs, SB4TrkSeg **seg,
                            int sID, int last_sID);
   static int   findSeg(const SB4Vector &chk_pt, int seg_ID,
                        int num_segs, SB4TrkSeg **seg);
};
//-----END Path.h-----//

//-----START PassPath.h-----//
class SB4PassPath : public SB4Path 
{
public:
   // CONSTRUCTOR
   SB4PassPath(float       _trk_wid,  /*IN : track width*/
               int         _num_segs, /*IN : number of track segments*/
               SB4TrkSeg **_seg,      /*IN : track segments*/
               SB4Path    &_mainPath  /*IN : racing path*/
              );
   // DESTRUCTOR
   virtual ~SB4PassPath();

   // This type of path needs no adjustment
   virtual void adjust(float in_w, float out_w, float trk_wid,
                       int num_segs, SB4TrkSeg **seg)
   {}

   /* Locates the given point, relative to this Path.
      Returns a non zero value, if a result is found, otherwise zero.
    */
   virtual int locatePoint(      int       *j,         /*OUT: a start point from which path distance is measured (can be NULL)*/
                                 float     *path_dist, /*OUT: path distance to the point*/
                                 float     *to_side,   /*OUT: +ve if point is right of path, -ve if left of path*/
                                 float     *offset,    /*OUT: offset from racing line*/
                           const SB4Vector &chk_pt,    /*IN : the point to be located*/
                                 SB4Line   &trk_norm,  /*IN : track normal at the given point*/
                                 int        seg_ID     /*IN : ID of the track segment*/
                          ) const;

   // Returns the curvature of this Path at the specified point.
   virtual float get_curv(int   *j,         /*IN : a start point from which path distance is measured*/
                          float *path_dist  /*IN : path distance to the point*/
                         ) const;

   // Sets the offset, from the racing path.
   void set_offset(float offset);

protected:
   int         num_segs;
   float       trk_wid, wgt;
   SB4TrkSeg **seg;

   // PROTECTED METHODS
   void getPoint(SB4Vector &ret_pt, SB4Vector &ret_tan,
                 int j, float path_dist, const SB4Vector &p1, 
                 const SB4Vector &p2, float len2, float ang2, float k2,
                 const SB4Vector &p3, float ang3, float k3,
                 const SB4Vector &p4) const;
   void getNewPoint(SB4Vector &ret_pt, int i, int sID) const;
};
//-----END PassPath.h-----//

//-----START TrkData.h-----//
class SB4TrkData
{
public:
   // MEMBERS
   int   pit_side;
   float pit_entry, pit_spd, trk_len, trk_wid;

   // CONSTRUCTORS
   SB4TrkData()
      : name(NULL), mainPath(NULL), passPath(NULL), seg(NULL), num_segs(0),
        trk_len(0.0f), trk_wid(0.0f), pit_spd(0.0f), pit_entry(0.0f),
        pit_side(0)
   {}
   // DESTUCTOR
   ~SB4TrkData();
   
   // METHODS
   void  get_trk_pos_vel(SB4Line &ret_pv, float to_lft, 
                         float v, float vn, 
                         float to_end, int seg_ID) const;
   void  locate_car(int *pidx, float *path_dist, float *to_side,
                    const SB4Vector &pt, int seg_ID, float *offset=0) const;
   float get_tgt_spd(float *brk_dist, float v, float drag_accel, 
                     int pidx, float path_dist, float offset=0.0f) const;
   float get_curv(int pidx, float path_dist, float offset=0.0f) const;
   void  set_path(Track *trk);
   
private:
   // MEMBERS
   char        *name;
   SB4Path     *mainPath;
   SB4PassPath *passPath;
   SB4TrkSeg  **seg;
   int          num_segs;

   // PRIVATE METHODS
   void get_trk_data(Track *trk);
   void adjust_Path();
};
//-----END TrkData.h-----//

//-----START SBDriver.h-----//
#ifdef SB4_DLL
class SmoothB4
#else
class SmoothB4 : public Driver
#endif
{
public:
   // CONSTRUCTORS
   SmoothB4(int i=-1);
   
   // DESTRUCTOR
   ~SmoothB4();

   // METHODS
   con_vec drive(situation &s);
   
private:
   // PRIVATE MEMBERS
   con_vec   cp;
   float     start_fuel, fuel_per_lap, error_last, error_avg, offset,
             brk_v, brk_dist;
   int       passing;
   SB4Line   cV;
   SB4Vector A;

   // PRIVATE METHODS
   void  get_RARS_ctrl_params(situation &s);
   void  get_Ay(float v, float curv, float error);
   void  update_con_vec(float v, float tgt_v, float mass, 
                        float drag_accel);
   float anti_collision(situation &s, float curv, float *error,
                        float *free_lane);
};
//-----END SBDriver.h-----//


//-----START Vector.cpp-----//
// METHODS
void SB4Vector::add(const SB4Vector &v) 
{
   x += v.x;
   y += v.y;
}
   
void SB4Vector::sub(const SB4Vector &v) 
{
   x -= v.x;
   y -= v.y;
}

float SB4Vector::mag() const
{
   return (float)sqrt((double)((x*x) + (y*y)));
}

float SB4Vector::mag2() const
{
   return ((x*x) + (y*y));
}

float SB4Vector::dist2Point(const SB4Vector &chk_pt) const
{
   float xd = x - chk_pt.x;
   float yd = y - chk_pt.y;

   return (float)sqrt((double)((xd*xd) + (yd*yd)));
}

float SB4Vector::dist2Point2(const SB4Vector &chk_pt) const
{
   float xd = x - chk_pt.x;
   float yd = y - chk_pt.y;

   return ((xd*xd) + (yd*yd));
}

float SB4Vector::ang() const
{
   float retval;

   if ( (x == 0.0f)&&(y == 0.0f) )
      retval = MY_PI12;
   else {
      retval = (float)atan2((double)y, (double)x);

      if (retval < 0.0f)
         retval += MY_PI21;
   }
   return retval;
}

float SB4Vector::ang(const SB4Vector &v) const
{
   float retval = v.ang() - ang();

   if (retval > MY_PI)
      retval -= MY_PI21;
   else if (retval < -MY_PI)
      retval += MY_PI21;
   return retval;
}
   
void SB4Vector::scale(float factor) 
{
   x *= factor;
   y *= factor;
}
   
void SB4Vector::normalize()
{
   float m;

   m = mag();
   if (m != 0.0f)
      scale(1.0f / m);
}

float SB4Vector::dot(const SB4Vector &v) const
{
   return (x*v.x) + (y*v.y);
}

void SB4Vector::rotate(float angle)
{
   float m, a;

   m = mag();
   a = ang();
   x = m * (float)cos((double)(a + angle));
   y = m * (float)sin((double)(a + angle));
}
//-----END Vector.cpp-----//

//-----START Line.cpp-----//
// METHODS
void SB4Line::getPoint(SB4Vector &ret_pt, float dist) const
{
   ret_pt.x = pt.x + (dist*dir.x);
   ret_pt.y = pt.y + (dist*dir.y);
}

void SB4Line::movePoint(float dist)
{
   pt.x += dist * dir.x;
   pt.y += dist * dir.y;
}

float SB4Line::dist2Point(const SB4Vector &chk_pt) const
{
   // +ve return value means chk_pt is to right of this line
   // -ve return value means chk_pt is to left of this line
   return ((dir.y*(chk_pt.x - pt.x)) - (dir.x*(chk_pt.y - pt.y)));
}

int SB4Line::intersect(SB4Vector &ret_pt, const SB4Line &l) const
{
   float D = (l.dir.x*dir.y) - (l.dir.y*dir.x);

   if (D == 0.0f)
      return 0;
   
   D = ( (dir.x*(l.pt.y - pt.y)) - (dir.y*(l.pt.x - pt.x)) ) / D;
   l.getPoint(ret_pt, D);
   return ((D > 0.0f) ? 1 : -1);
}
//-----END Line.cpp-----//

//-----START Circle.cpp-----//
// METHODS
float SB4Circle::dist2Point(const SB4Vector &chk_pt) const
{
   // +ve return value means chk_pt is outside this circle
   // -ve return value means chk_pt is inside this circle
   return (c.dist2Point(chk_pt) - R);
}

void SB4Circle::getPoint(SB4Vector &ret_pt, float angle) const
{
   ret_pt.x = c.x + (R * (float)cos((double)angle));
   ret_pt.y = c.y + (R * (float)sin((double)angle));
}

float SB4Circle::ang(SB4Vector &pt) const
{
   float retval;

   pt.sub(c);
   retval = pt.ang();
   pt.add(c);
   return retval;
}
//-----END Circle.cpp-----//

//-----START TrkSeg.cpp-----//
// METHODS
void SB4TrkSeg::getPoint(SB4Vector &ret_pt, float trk_dist, float to_lft) const
{
   SB4Line n;

   get_norm(n, trk_dist);
   n.getPoint(ret_pt, to_lft);
}

float SB4TrkSeg::to_left(const SB4Vector &chk_pt) const
{
   if (rad >= 0.0f)
      return left_side->dist2Point(chk_pt);
   return -left_side->dist2Point(chk_pt);
}

int SB4TrkSeg::same_dir(const SB4TrkSeg &seg) const
{
   if (rad == 0.0f)
      return (seg.rad == 0.0f);
   else if (rad > 0.0f)
      return (seg.rad > 0.0f);
   return (seg.rad < 0.0f);
}
//-----END TrkSeg.cpp-----//

//-----START TrkSeg_Straight.cpp-----//
// CONSTRUCTOR
SB4TrkSeg_Straight::SB4TrkSeg_Straight(segment *seg, float trk_dist)
   : SB4TrkSeg(),
     ls((float)seg->beg_x, (float)seg->beg_y, (float)seg->beg_ang)
{
   Sdst = trk_dist;
   Fdst = trk_dist + (float)seg->length;
   len  = (float)seg->length;
   rad  = (float)seg->radius;

   left_side = &ls;

   if (len >= 900.0f)
      Npts = 18;
   else if (len >= 100.0f)
      Npts = (int)(len / 50.0f);
   else
      Npts = 1;
}

// METHODS
void SB4TrkSeg_Straight::get_norm(SB4Line &ret_n, float trk_dist) const
{
   if ( (trk_dist < (Sdst-2.5f))||(trk_dist > (Fdst+2.5f)) ) {
      ret_n.pt.scale(0.0f);
      ret_n.dir.scale(0.0f);
   }
   else {
      ls.getPoint(ret_n.pt, (trk_dist - Sdst));
      ret_n.dir.x =  ls.dir.y;
      ret_n.dir.y = -ls.dir.x;
   }
}

float SB4TrkSeg_Straight::get_trk_dist(const SB4Vector &chk_pt) const
{
   float N = (ls.dir.y*(chk_pt.y-ls.pt.y)) + (ls.dir.x*(chk_pt.x-ls.pt.x));

   return (N + Sdst);
}
//-----END TrkSeg_Straight.cpp-----//

//-----START TrkSeg_Turn.cpp-----//
// CONSTRUCTOR
SB4TrkSeg_Turn::SB4TrkSeg_Turn(segment *seg, float wid, float trk_dist)
   : SB4TrkSeg(),
     ls((float)seg->cen_x, (float)seg->cen_y, (float)fabs(seg->radius))
{
   beg_ang = (float)seg->beg_ang;
   end_ang = (float)seg->end_ang;
   if (seg->radius > 0.0f) {   // left turn
      len      = (float)seg->length * (( 0.5f*wid) + (float)seg->radius);
      rad      = (float)seg->radius;
      beg_ang -= MY_PI12;
      end_ang -= MY_PI12;
      left     = 1;
   }
   else {                      // right turn
      len      = (float)seg->length * ((-0.5f*wid) - (float)seg->radius);
      rad      = (float)seg->radius + wid;
      beg_ang += MY_PI12;
      end_ang += MY_PI12;
      left     = 0;
   }

   if (beg_ang < 0.0f)
      beg_ang += MY_PI21;
   else if (beg_ang > MY_PI21)
      beg_ang -= MY_PI21;
   if (end_ang < 0.0f)
      end_ang += MY_PI21;
   else if (end_ang > MY_PI21)
      end_ang -= MY_PI21;

   if ( (rad < 0.0f)&&(end_ang > beg_ang) )
      beg_ang += MY_PI21;
   else if ( (rad > 0.0f)&&(end_ang < beg_ang) )
      end_ang += MY_PI21;

   Sdst      = trk_dist;
   Fdst      = trk_dist + len;
   left_side = &ls;

   if (len >= 400.0f)
      Npts = (int)(len / 60.0f);
   else if (len >= 150.0f)
      Npts = (int)(len / 50.0f);
   else
      Npts = (int)(len / 40.0f);
   if (seg->length >= (0.35*2.0))
      Npts += (int)(seg->length / 0.35);
   else
      Npts++;

   Npts /= 2;
   Npts |= 1;
   if (Npts > 17)
      Npts = 17;
   len   = (float)seg->length;
}

// METHODS
void SB4TrkSeg_Turn::get_norm(SB4Line &ret_n, float trk_dist) const
{
   float ang;

   if ( (trk_dist < (Sdst-2.5f))||(trk_dist > (Fdst+2.5f)) ) {
      ret_n.pt.scale(0.0f);
      ret_n.dir.scale(0.0f);
      return;
   }

   ang  = (trk_dist - Sdst) / (Fdst - Sdst);
   ang *= end_ang - beg_ang;
   ls.getPoint(ret_n.pt, (ang + beg_ang));
   if (rad > 0.0f) {
      ret_n.dir.x = ret_n.pt.x - ls.c.x;
      ret_n.dir.y = ret_n.pt.y - ls.c.y;
   }
   else {
      ret_n.dir.x = ls.c.x - ret_n.pt.x;
      ret_n.dir.y = ls.c.y - ret_n.pt.y;
   }
   ret_n.dir.normalize();
}

float SB4TrkSeg_Turn::get_trk_dist(const SB4Vector &chk_pt) const
{
   float     dist;
   SB4Vector loc(chk_pt);

   loc.sub(ls.c);
   dist  = loc.ang() - beg_ang;
   if (dist > MY_PI21)
      dist -= MY_PI21;
   else if (dist < -MY_PI21)
      dist += MY_PI21;

   if ( (beg_ang < end_ang)&&(dist < -0.5f*(MY_PI21-len)) )
      dist += MY_PI21;
   else if ( (beg_ang > end_ang)&&(dist > 0.5f*(MY_PI21-len)) )
      dist -= MY_PI21;

   dist *= (Fdst - Sdst) / (end_ang - beg_ang);
   return (dist + Sdst);
}
//-----END TrkSeg_Turn.cpp-----//

//-----START Path.cpp-----//
// CONSTRUCTOR
SB4Path::SB4Path(int num_segs, SB4TrkSeg **seg, float trk_wid, 
                 const char *name)
{
   int   i, j, k;
   float trk_dist, delta;

   Spt = new int[num_segs];
   for (i = 0, Npts = 0; i < num_segs; i++) {
      Npts += seg[i]->Npts;
   }
   pt  = new SB4Vector[Npts];
   ang = new float[Npts];
   len = new float[Npts];
 
   Nwgts    = 0;
   curv_wgt = NULL;

   trk_wid *= 0.5f;
   for (i = j = 0; i < num_segs; i++) {
      delta    = (seg[i]->Fdst - seg[i]->Sdst) / (float)seg[i]->Npts;
      trk_dist = (0.5f * delta) + seg[i]->Sdst;
      Spt[i]   = j;
      for (k = 0; k < seg[i]->Npts; k++) {
         seg[i]->getPoint(pt[j], trk_dist, trk_wid);
         trk_dist += delta;
         j++;
      }
   }
   loadData(name);
}

// DESTRUCTOR
SB4Path::~SB4Path()
{
   if (Spt != NULL)
      delete [] Spt;
   if (pt != NULL)
      delete [] pt;
   if (ang != NULL)
      delete [] ang;
   if (len != NULL)
      delete [] len;
   if (curv_wgt != NULL)
      delete [] curv_wgt;
}

//METHODS
void SB4Path::adjust(float in_w, float out_w, float trk_wid,
                     int num_segs, SB4TrkSeg **seg)
{
   int    i;
   float *dir = new float[Npts];

   for (i = 0; i < 3; i++) {
      adjust_pts(in_w, out_w, trk_wid, num_segs, seg, dir);
   }

   delete [] dir;
   if (curv_wgt != NULL)
      delete [] curv_wgt;
}

int SB4Path::locatePoint(int *j, float *path_dist, float *to_side, 
                         float *offset, const SB4Vector &chk_pt, 
                         SB4Line &trk_norm, int seg_ID) const
{
   int k   = Spt[seg_ID],
       dir = 0;

   // find start point
   *j = (k + Npts - 1) % Npts;
   while (1) {
      if (dir <= 0) {
         *path_dist = trk_norm.dist2Point(pt[*j]);
         if (*path_dist < 0.0f) {
            k   = *j;
            *j  = (k+Npts-1) % Npts;
            dir = -1;
            continue;
         }
      }
      if (dir >= 0) {
         *path_dist = trk_norm.dist2Point(pt[k]);
         if (*path_dist > 0.0f) {
            *j  = k;
            k   = (*j+1) % Npts;
            dir = 1;
            continue;
         }
      }
      dir = 0;
      break;
   }

   // find path distance
   *path_dist  = (chk_pt.x - pt[*j].x) * (pt[k].x - pt[*j].x);
   *path_dist += (chk_pt.y - pt[*j].y) * (pt[k].y - pt[*j].y);
   *path_dist *= len[*j] / pt[k].dist2Point2(pt[*j]);
   if (*path_dist > len[*j])
      *path_dist = len[*j];
   else if (*path_dist < 0.0f)
      *path_dist = 0.0f;

   // find to_side
   getPoint(trk_norm.pt, trk_norm.dir, *j, *path_dist);
   *to_side = trk_norm.dist2Point(chk_pt);

   return 1;
}

float SB4Path::get_curv(int *j, float *path_dist) const
{
   int   i, k, l;
   float retval, w;

   if (*path_dist >= 0.0f) {
      while (*path_dist > len[*j]) {
         *path_dist -= len[*j];
         *j          = (*j+1) % Npts;
      }
   }
   else {
      do {
         *j          = (*j+Npts-1) % Npts;
         *path_dist += len[*j];
      } while (*path_dist < 0.0f);
   }
   i = (*j+Npts-1) % Npts;
   k = (*j+1) % Npts;
   l = (*j+2) % Npts;

   w       = *path_dist / len[*j];
   retval  = (1.0f-w) * curvature(pt[i], pt[*j], pt[k]);
   retval +=       w  * curvature(pt[*j], pt[k], pt[l]);  
   return retval;
}

// PROTECTED METHODS
void SB4Path::getPoint(SB4Vector &ret_pt, SB4Vector &ret_tan, 
                       int j, float path_dist) const
{
   int   k  = (j+1) % Npts;
   float k1 = curvature(pt[(j+Npts-1)%Npts], pt[j], pt[k]),
         k2 = curvature(pt[j], pt[k], pt[(k+1)%Npts]);

   path_dist /= len[j];
   k1 = (float)fabs((double)k1);
   k2 = (float)fabs((double)k2);
   if ( (len[j] < (0.5f*SB4_MIN_LEN))||
        (k1 < (0.4f*SB4_MIN_CURV))||(k2 < (0.4f*SB4_MIN_CURV)) ) {
      // linear interpolation between pt[j] and pt[k]
      ret_pt.x = pt[k].x - pt[j].x;
      ret_pt.y = pt[k].y - pt[j].y;
      ret_pt.scale(path_dist);
      ret_pt.add(pt[j]);

      // get the tangent vector
      ret_tan.x = (float)cos((double)ang[j]);
      ret_tan.y = (float)sin((double)ang[j]);
   }
   else {
      float A =  (2.0f*pt[j].x) - (2.0f*pt[k].x),
            B = (-3.0f*pt[j].x) + (3.0f*pt[k].x),
            C = (float)cos((double)ang[j]) / k1,
            D =  (2.0f*pt[j].y) - (2.0f*pt[k].y),
            E = (-3.0f*pt[j].y) + (3.0f*pt[k].y),
            F = (float)sin((double)ang[j]) / k1,
            t = (float)cos((double)ang[k]) / k2;
      // calculate spline parameters
      k1 *= 0.5f * k1 * pt[j].dist2Point2(pt[k]);
      k1  = (float)acos(1.0-(double)k1);
      C  *= k1;
      F  *= k1;
      k1  = 0.5f * k2 * k2 * pt[j].dist2Point2(pt[k]);
      k1  = (float)acos(1.0-(double)k1);
      t  *= k1;
      A  +=       C  + t;
      B  -= (2.0f*C) + t;
      t   = k1*((float)sin((double)ang[k]) / k2);
      D  +=       F  + t;
      E  -= (2.0f*F) + t;
         
      // use Hermite splines to interpolate between pt[j] and pt[k]
      ret_pt.x = P(path_dist, A, B, C, pt[j].x);
      ret_pt.y = P(path_dist, D, E, F, pt[j].y);

      // get the tangent vector
      ret_tan.x = dP(path_dist, (3.0f*A), (2.0f*B), C);
      ret_tan.y = dP(path_dist, (3.0f*D), (2.0f*E), F);
      ret_tan.normalize();
   }
}

void SB4Path::fill_ang()
{
   int i, j, k;

   for (i = Npts-1, j = 0; j < Npts; j++) {
      k = (j+1) % Npts;

      ang[j] = get_ang(pt[i], pt[j], pt[k]);

      i = j;
   }
}

void SB4Path::fill_len()
{
   int i, j, k, l;

   for (i = Npts-1, j = 0, k = 1; j < Npts; j++) {
      l = (k+1) % Npts;
      len[j] = get_len(pt[j], ang[j], curvature(pt[i], pt[j], pt[k]),
                       pt[k], ang[k], curvature(pt[j], pt[k], pt[l]));
      i = j;
      k = l;
   }
}

void SB4Path::loadData(const char *name)
{
   FILE  *in;
   int    i;
   char   buf[128];
   size_t name_len = strlen(name);
   
   if ( (0 < name_len)&&(name_len < 128)&&
        (in = fopen(SB4_DATA_FILE, "r")) ) {
      do {
         buf[name_len] = 0;
      } while ( (fgets(buf, 128, in))&&(strncmp(buf, name, name_len)) );

      Nwgts = 0;
      if ( (sscanf((buf+name_len), "%d", &Nwgts) > 0)&&(Nwgts > 0) ) {
         char *ptr;

         curv_wgt = new float[Nwgts];
         if (curv_wgt == NULL)
            Nwgts = 0;

         i = 0;
         while ( (i < Nwgts)&&(fgets(buf, 128, in)) ) {
            ptr = buf;
            do {
               curv_wgt[i] = 0.0f;
               if ( (sscanf(ptr, "%f", &(curv_wgt[i])) <= 0)||
                    (curv_wgt[i] <= 0.0f)||(curv_wgt[i] >= 1.0f) ) {
                  Nwgts = i;
                  break;
               }
               i++;
               ptr += strspn(ptr, " \t");
            } while ( (i < Nwgts)&&(*ptr != '\0')&&
                      ((ptr = strpbrk(ptr, " /t")) != NULL) );
         }
      }
      fclose(in);
   }
}

void SB4Path::adjust_pts(int num_segs, SB4TrkSeg **seg)
{
   int       i, j, k, rnds, seg_ID;
   float     w, max_change;
   SB4Vector tmp, tmp2;
      
   for (rnds = 0; rnds < 10; rnds++) {
      max_change = 0.0f;
      for (i = Npts-1, j = 0, seg_ID = 1; j < Npts; j++) {
         k = (j+1) % Npts;

         if ( (Spt[seg_ID] == j)&&(Spt[(seg_ID+1)%num_segs] == k) ) {
            if (seg[seg_ID]->rad != 0.0f) {
               tmp.x = 0.5f*(seg[seg_ID]->Sdst + seg[seg_ID]->Fdst);
               tmp.y = seg[seg_ID]->to_left(pt[j]);
               seg[seg_ID]->getPoint(tmp, tmp.x, tmp.y);

               w  = tmp.dist2Point(pt[i]);
               w /= tmp.dist2Point(pt[i]) + tmp.dist2Point(pt[k]);
               w *= len[i] + len[j];
            }
            else
               w  = len[i];
         }
         else {
            w = pt[j].dist2Point(pt[i]) / pt[j].dist2Point(pt[k]);
            if ( ((1.0f/1.1f) > w)||(w > 1.1f) )
               w = 0.5f * (len[i] + len[j]);
            else
               w = len[i];
         }

         if ( (j > 0)&&(w < len[i]) ) {
            getPoint(tmp, tmp2, i, w);
            ang[j]  = get_ang(pt[i], tmp, pt[k]);
            len[j]  = len[i] + len[j] - w;
            len[i]  = w;
            w       = pt[j].dist2Point(tmp);
            pt[j].x = tmp.x;
            pt[j].y = tmp.y;
            if (w > max_change)
               max_change = w;

            if (j == Spt[seg_ID]) {
               i = (seg_ID+num_segs-3) % num_segs;
               i = findSeg(pt[j], i, num_segs, seg);
               i = (i+1) % num_segs;
               while (j == Spt[i]) {
                  Spt[i] = k;
                  i      = (i+1) % num_segs;
               }
            }
         }
         else if ( (k > 0)&&(w > len[i]) ) {
            getPoint(tmp, tmp2, j, (w - len[i]));
            ang[j]  = get_ang(pt[i], tmp, pt[k]);
            len[j]  = len[i] + len[j] - w;
            len[i]  = w;
            w       = pt[j].dist2Point(tmp);
            pt[j].x = tmp.x;
            pt[j].y = tmp.y;
            if (w > max_change)
               max_change = w;

            if (k == Spt[seg_ID]) {
               i = (seg_ID+num_segs-3) % num_segs;
               i = findSeg(pt[j], i, num_segs, seg);
               while (k == Spt[i]) {
                  Spt[i] = j;
                  i      = (i+num_segs-1) % num_segs;
               }
            }
         }

         while (j == Spt[seg_ID]) {
            seg_ID = (seg_ID+1) % num_segs;
         }
         i = j;
      }
      if (max_change < SB4_MIN_CHANGE)
         break;
   }
}

void SB4Path::adjust_pts(float in_w, float out_w, float trk_wid,
                         int num_segs, SB4TrkSeg **seg, float *dir)
{
   int       i, j, k, rnd, sID, last_sID, min_rnd;
   float     ci, cj, ck, w, min, max, max_change, min_change;
   SB4Vector tmp, tmp2;

   min_rnd = 0;
   min_change = 100000.0f;
   for (rnd = 0; rnd < SB4_MAX_RNDS; rnd++) {
      max_change = 0.0f;
      sID        = 0;
      last_sID   = num_segs - 1;
      for (i = Npts-1, j = 0; j < Npts; j++) {
         k   = (j+1) % Npts;
         sID = findSeg(pt[j], last_sID, num_segs, seg);

         ci = curvature(pt[(i+Npts-1)%Npts], pt[i], pt[j]);
         ck = curvature(pt[j], pt[k], pt[(k+1)%Npts]);
         if ( (curv_wgt != NULL)&&(sID < Nwgts) )
            w = curv_wgt[sID];
         else if (seg[sID]->rad == 0.0f) {
            w = 0.53f;
            if ( (trk_wid < 300.0f)&&(seg[sID]->len < (3.5f*trk_wid))&&
                 (seg[(sID+num_segs-1)%num_segs]->same_dir(*(seg[(sID+1)%num_segs])))&&
                 (seg[(sID+num_segs-1)%num_segs]->len > 1.5f)&&
                 (seg[(sID+1)%num_segs]->len > 1.5f) )
               w *= trk_wid / 300.0f;
         }
         else {
            if ( (-700.0f > seg[sID]->rad)||(seg[sID]->rad > 700.0f) )
               w = 0.585f;
            else
               w = 0.535f;
         }
         if (fabs((double)ck) > fabs((double)ci))
            w = 1.0f - w;
         cj = (w*ci) + ((1.0f-w)*ck);
         
         ci = curvature(pt[i], pt[j], pt[k]);
         ck = (0.1f*ci) + (0.9f*cj);
         w  = pt[j].dist2Point(pt[i]) / pt[j].dist2Point(pt[k]);
         if (movePoint(pt[i], tmp, pt[k], cj, w) <= 0) {
            ck = 0.5f * (ci + cj);
            if (movePoint(pt[i], tmp, pt[k], ck, w) <= 0) {
               ck = (0.9f*ci) + (0.1f*cj);
               if (movePoint(pt[i], tmp, pt[k], ck, w) <= 0) {
                  last_sID = sID;
                  i        = j;
                  continue;
               }
            }
         }

         if (cj > (4.0f*SB4_MIN_CURV)) {        // left
            min  = in_w;
            max  = trk_wid - out_w;
         }
         else if (cj < (-4.0f*SB4_MIN_CURV)) {  // right
            min  = out_w;
            max  = trk_wid - in_w;
         }
         else {                                 // straight
            if (seg[sID]->rad == 0.0f) {
               min = out_w;
               max = trk_wid - out_w;
            }
            else {
               min = 0.7f*out_w;
               max = trk_wid - (0.7f*out_w);
            }
         }

         if (rnd < 200) {
            w = adjustPoint(pt[i], pt[j], tmp, pt[k], min, max,
                            num_segs, seg, sID, last_sID);
            if (!rnd)
               dir[j]  = w;
            else
               dir[j] += w;
         }
         else {
            tmp2.x = pt[j].x;
            tmp2.y = pt[j].y;
            w = adjustPoint(pt[i], tmp2, tmp, pt[k], min, max,
                            num_segs, seg, sID, last_sID);
            if ( ( (dir[j] < 0.0f)&&(w < 0.0f) )||
                 ( (dir[j] > 0.0f)&&(w > 0.0f) )||
                   (dir[j] == 0.0f) ) {
               pt[j].x = tmp2.x;
               pt[j].y = tmp2.y;
            }
            else
               w = 0.0f;
         }
         w = (float)fabs((double)w);
         if (w > max_change)
            max_change = w;

         last_sID = sID;
         i        = j;
      }

      if (max_change < min_change) {
         min_change = max_change;
         min_rnd    = rnd;
      }

      fill_ang();
      fill_len();
      adjust_pts(num_segs, seg);
      if (max_change < SB4_MIN_CHANGE)
         break;
   }
   fill_ang();
   fill_len();
}

// STATIC METHODS
float SB4Path::get_ang(const SB4Vector &p1, const SB4Vector &p2, 
                       const SB4Vector &p3)
{
   float     curv, a;
   SB4Vector c;

   curv = curvature(p1, p2, p3);
   if ( ((-0.4f*SB4_MIN_CURV) < curv)&&(curv < (0.4f*SB4_MIN_CURV)) ) {
      c.x = p3.x - p2.x;
      c.y = p3.y - p2.y;
      a   = c.ang();
   }
   else {
      get_center(c, p1, p3, (1.0f/curv));
      c.sub(p2);
      a = c.ang();
      if (curv > 0.0f)
         a -= MY_PI12;
      else
         a += MY_PI12;
   }
   return a;
}

float SB4Path::get_len(const SB4Vector &p1, float a1, float k1,
                       const SB4Vector &p2, float a2, float k2)
{
   float s = p1.dist2Point(p2);

   k1 = (float)fabs((double)k1);
   k2 = (float)fabs((double)k2);
   if ( (s > SB4_MIN_LEN)&&(k1 > SB4_MIN_CURV)&&(k2 > SB4_MIN_CURV) ) {
      float sA, sB, sC, sD, sE,
            A  =  (2.0f*p1.x) - (2.0f*p2.x),
            B  = (-3.0f*p1.x) + (3.0f*p2.x),
            C  = (float)cos((double)a1) / k1,
            D  =  (2.0f*p1.y) - (2.0f*p2.y),
            E  = (-3.0f*p1.y) + (3.0f*p2.y),
            F  = (float)sin((double)a1) / k1,
            t  = (float)cos((double)a2) / k2;
      s  *= s;
      k1 *= 0.5f * k1 * s;
      k1  = (float)acos(1.0 - (double)k1);
      C  *= k1;
      F  *= k1;
      k1  = 0.5f * k2 * k2 * s;
      k1  = (float)acos(1.0 - (double)k1);
      t  *= k1;
      A  +=       C  + t;
      B  -= (2.0f*C) + t;
      t   = k1*((float)sin((double)a2) / k2);
      D  +=       F  + t;
      E  -= (2.0f*F) + t;
         
      sA  =  9.0f * ((A*A) + (D*D));
      sB  = 12.0f * ((A*B) + (D*E));
      sC  =  4.0f * ((B*B) + (E*E));
      sC +=  6.0f * ((A*C) + (D*F));
      sD  =  4.0f * ((B*C) + (E*F));
      sE  =          (C*C) + (F*F);

      s = 0.0f;
      t = 0.5f * SB4_STEP;
      do {
         s += dP_mag(t, sA, sB, sC, sD, sE) * SB4_STEP;
         t += SB4_STEP;
      } while (t < (1.0f - (0.25f*SB4_STEP)));
   }
   return s;
}

float SB4Path::P(float t, float A, float B, float C, float D)
{
   D += (C*t) + (B*t*t) + (A*t*t*t);
   return D;
}

float SB4Path::dP(float t, float A, float B, float C)
{
   C += (B*t) + (A*t*t);
   return C;
}

float SB4Path::dP_mag(float t, float A, float B, float C, float D, float E)
{
   E += (D*t) + (C*t*t) + (B*t*t*t) + (A*t*t*t*t);
   return (float)sqrt((double)E);
}

void SB4Path::get_center(SB4Vector &ret_pt, const SB4Vector &p1, 
                         const SB4Vector &p2, float R)
{
   float ang, dist;

   ret_pt.x = 0.5f * (p2.x - p1.x);
   ret_pt.y = 0.5f * (p2.y - p1.y);
      
   dist = (float)sqrt((double)((R*R) - ret_pt.mag2()));
   ang  = ret_pt.ang();
   if (R > 0.0f)
      ang += MY_PI12;
   else
      ang -= MY_PI12;

   ret_pt.add(p1);
   ret_pt.x += dist * (float)cos((double)ang);
   ret_pt.y += dist * (float)sin((double)ang);
}

void SB4Path::get_mid_point(SB4Vector &ret_pt, const SB4Vector &p1, 
                            const SB4Vector &p2, float R)
{
   float ang,
         dist = (float)fabs((double)R);

   ret_pt.x = 0.5f * (p2.x - p1.x);
   ret_pt.y = 0.5f * (p2.y - p1.y);
      
   dist -= (float)sqrt((double)((R*R) - ret_pt.mag2()));
   ang   = ret_pt.ang();
   if (R > 0.0f)
      ang -= MY_PI12;
   else
      ang += MY_PI12;
         
   ret_pt.add(p1);
   ret_pt.x += dist * (float)cos((double)ang);
   ret_pt.y += dist * (float)sin((double)ang);
}

float SB4Path::curvature(const SB4Vector &p1, const SB4Vector &p2, 
                         const SB4Vector &p3)
{
   float num = (p3.x-p2.x)*(p1.y-p2.y) - (p1.x-p2.x)*(p3.y-p2.y),
         den = p2.dist2Point(p3)*p2.dist2Point(p1)*p1.dist2Point(p3);
   return ((2.0f*num)/den);
}

int SB4Path::movePoint(const SB4Vector &p1, SB4Vector &p2, 
                       const SB4Vector &p3, float curv, float w)
{
   float ang;
   float dst = p1.dist2Point(p3);
   int   pts = 0;

   p2.x = p1.x - p3.x;
   p2.y = p1.y - p3.y;
   ang  = p2.ang();
   if ( ( ((-0.1f*SB4_MIN_CURV) < curv)&&(curv < (0.1f*SB4_MIN_CURV)) )||
        (w == 0.0f) ) {
      p2.x = dst / (1.0f+w);
      p2.y = 0.0f;
      pts++;
   }
   else {
      float W  = 0.5f*(1.0f - (w*w)),
            C  = 0.5f*w*curv,
            qA = (C*C) + ((W*W)/(dst*dst)),
            qB = 1.0f - W,
            qC = 0.25f*dst*dst;

      qC *= -4.0f * qA;
      qC += qB * qB;
      if (qC >= 0.0f) {
         qA *= 2.0f;
         qC  = (float)sqrt((double)qC);
         qA  = (qB - qC) / qA;

         if (qA >= 0.0f) {
            W   /= dst;
            p2.x = (0.5f*dst) + (W*qA);
            p2.y = C*qA;
            pts++;
         }
      }
   }

   if (pts > 0) {
      p2.rotate(ang);
      p2.add(p3);
   }
   return pts;
}

float SB4Path::adjustPoint(const SB4Vector &p1, SB4Vector &p2,
                           SB4Vector &p2b, const SB4Vector &p3,
                           float min, float max,
                           int num_segs, SB4TrkSeg **seg,
                           int sID, int last_sID)
{
   int   i;
   float curv, w, 
         trk_dist = seg[sID]->get_trk_dist(p2),
         old_val  = seg[sID]->to_left(p2),
         to_lft   = seg[sID]->to_left(p2b);

   // make sure the new point is within limits
   if (to_lft < min)
      to_lft = min;
   else if (to_lft > max)
      to_lft = max;
   seg[sID]->getPoint(p2, trk_dist, to_lft);

   // make sure the area between p1 and p2 is within limits
   curv = curvature(p1, p2, p3);
   if ( (-SB4_MIN_CURV < curv)&&(curv < SB4_MIN_CURV) ) {
      p2b.x = 0.5f * (p2.x + p1.x);
      p2b.y = 0.5f * (p2.y + p1.y);
   }
   else {
      get_mid_point(p2b, p1, p2, (1.0f/curv));
      p2b.scale(0.5f);
      p2b.x += 0.5f * (0.5f*(p2.x + p1.x));
      p2b.y += 0.5f * (0.5f*(p2.y + p1.y));
   }
   i = findSeg(p2b, last_sID, num_segs, seg);
   w = seg[i]->to_left(p2b);
   if (w < min)
      w = min - w;   // increase to_lft
   else if (w > max)
      w = max - w;   // decrease to_lft
   else
      w = 0.0f;      // no change to to_lft
   if (w != 0.0f) {
      to_lft += w;
      if (to_lft < min)
         to_lft = min;
      else if (to_lft > max)
         to_lft = max;
      seg[sID]->getPoint(p2, trk_dist, to_lft);
      curv = curvature(p1, p2, p3);
   }
      
   // make sure the area between p2 and p3 is within limits
   if ( (-SB4_MIN_CURV < curv)&&(curv < SB4_MIN_CURV) ) {
      p2b.x = 0.5f * (p3.x + p2.x);
      p2b.y = 0.5f * (p3.y + p2.y);
   }
   else {
      get_mid_point(p2b, p2, p3, (1.0f/curv));
      p2b.scale(0.5f);
      p2b.x += 0.5f * (0.5f*(p3.x + p2.x));
      p2b.y += 0.5f * (0.5f*(p3.y + p2.y));
   }
   i = findSeg(p2b, sID, num_segs, seg);
   w = seg[i]->to_left(p2b);
   if (w < min)
      w = min - w;   // increase to_lft
   else if (w > max)
      w = max - w;   // decrease to_lft
   else
      w = 0.0f;      // no change to to_lft
   if (w != 0.0f) {
      to_lft += w;
      if (to_lft < min)
         to_lft = min;
      else if (to_lft > max)
         to_lft = max;
      seg[sID]->getPoint(p2, trk_dist, to_lft);
   }

   return (old_val - to_lft);
}

int SB4Path::findSeg(const SB4Vector &chk_pt, int seg_ID,
                     int num_segs, SB4TrkSeg **seg)
{
   int   i;
   float dist;

   for (i = 0; i < num_segs; i++) {
      dist = seg[seg_ID]->get_trk_dist(chk_pt);
      if ( (seg[seg_ID]->Sdst <= dist)&&(dist <= seg[seg_ID]->Fdst) )
         break;
      seg_ID = (seg_ID+1) % num_segs;
   }
   return seg_ID;
}
//-----END Path.cpp-----//

//-----START PassPath.cpp-----//
// CONSTRUCTOR
SB4PassPath::SB4PassPath(float       _trk_wid,  /*IN : track width*/
                         int         _num_segs, /*IN : number of track segments*/
                         SB4TrkSeg **_seg,      /*IN : track segments*/
                         SB4Path    &_mainPath  /*IN : racing path*/
                        )
   : SB4Path(_mainPath), trk_wid(_trk_wid), wgt(0.0f),
     num_segs(_num_segs), seg(_seg)
{
   ang = NULL;
}

// DESTRUCTOR
SB4PassPath::~SB4PassPath()
{
   seg = NULL;
   pt  = NULL;
   len = NULL;
   Spt = NULL;
}

// METHODS
int SB4PassPath::locatePoint(int *j, float *path_dist, float *to_side, 
                             float *offset, const SB4Vector &chk_pt, 
                             SB4Line &trk_norm, int seg_ID) const
{
   SB4Vector p1, p2, p3, p4;
   float     len2, ang2, k2, ang3, k3;
   int       k,
             dir = 0;

   // find start point
   *j = Spt[seg_ID];
   k  = (*j + 1) % Npts;
   while (1) {
      if (dir <= 0) {
         *path_dist = trk_norm.dist2Point(pt[*j]);
         if (*path_dist < 0.0f) {
            k   = *j;
            *j  = (k+Npts-1) % Npts;
            dir = -1;
            continue;
         }
      }
      if (dir >= 0) {
         *path_dist = trk_norm.dist2Point(pt[k]);
         if (*path_dist > 0.0f) {
            *j  = k;
            k   = (*j+1) % Npts;
            dir = 1;
            continue;
         }
      }
      dir = 0;
      break;
   }

   // generate the new points
   k      = (*j+Npts-1) % Npts;
   for (seg_ID = 0; (Spt[seg_ID] < k)&&(seg_ID < (num_segs-1)); seg_ID++) {
   }
   seg_ID = findSeg(pt[k], ((seg_ID+num_segs-1)%num_segs), num_segs, seg);
   getNewPoint(p1, k, seg_ID);
   seg_ID = findSeg(pt[*j], seg_ID, num_segs, seg);
   getNewPoint(p2, *j, seg_ID);
   k      = (k+2) % Npts;
   seg_ID = findSeg(pt[k], seg_ID, num_segs, seg);
   getNewPoint(p3, k, seg_ID);
   k      = (k+1) % Npts;
   seg_ID = findSeg(pt[k], seg_ID, num_segs, seg);
   getNewPoint(p4, k, seg_ID);

   ang2 = get_ang(p1, p2, p3);
   k2   = curvature(p1, p2, p3);
   ang3 = get_ang(p2, p3, p4);
   k3   = curvature(p2, p3, p4);
   len2 = get_len(p2, ang2, k2, p3, ang3, k3);

   // find path distance
   *path_dist  = (chk_pt.x - p2.x) * (p3.x - p2.x);
   *path_dist += (chk_pt.y - p2.y) * (p3.y - p2.y);
   *path_dist *= len2 / p3.dist2Point2(p2);
   if (*path_dist > len2)
      *path_dist = len2;
   else if (*path_dist < 0.0f)
      *path_dist = 0.0f;

   // find to_side
   getPoint(trk_norm.pt, trk_norm.dir, *j, *path_dist,
            p1, p2, len2, ang2, k2, p3, ang3, k3, p4);
   *to_side = trk_norm.dist2Point(chk_pt);
   *offset  = p2.dist2Point(pt[*j]);
   if (wgt < 0.0f)
      *offset *= -1.0f;

   return 1;
}

float SB4PassPath::get_curv(int *j, float *path_dist) const
{
   int       seg_ID, i;
   float     len2, ang2, k2, ang3, k3, retval;
   SB4Vector p1, p2, p3, p4;

   // find segment
   i = (*j+Npts-1) % Npts;
   for (seg_ID = 0; (Spt[seg_ID] < i)&&(seg_ID < (num_segs-1)); seg_ID++) {
   }

   if (*path_dist >= 0.0f) {
      // generate the new points
      seg_ID = findSeg(pt[i], ((seg_ID+num_segs-1)%num_segs), num_segs, seg);
      getNewPoint(p1, i, seg_ID);
      i      = (i+1) % Npts;
      seg_ID = findSeg(pt[i], seg_ID, num_segs, seg);
      getNewPoint(p2, i, seg_ID);
      i      = (i+1) % Npts;
      seg_ID = findSeg(pt[i], seg_ID, num_segs, seg);
      getNewPoint(p3, i, seg_ID);
      i      = (i+1) % Npts;
      seg_ID = findSeg(pt[i], seg_ID, num_segs, seg);
      getNewPoint(p4, i, seg_ID);

      ang2 = get_ang(p1, p2, p3);
      k2   = curvature(p1, p2, p3);
      while (1) {
         ang3 = get_ang(p2, p3, p4);
         k3   = curvature(p2, p3, p4);
         len2 = get_len(p2, ang2, k2, p3, ang3, k3);

         if (*path_dist <= len2)
            break;
         *path_dist -= len2;

         // shift points
         p1.x = p2.x; p1.y = p2.y;
         p2.x = p3.x; p2.y = p3.y;
         p3.x = p4.x; p3.y = p4.y;
         ang2 = ang3;
         k2   = k3;
         // generate a new point
         i      = (i+1) % Npts;
         *j     = (*j+1) % Npts;
         seg_ID = findSeg(pt[i], seg_ID, num_segs, seg);
         getNewPoint(p4, i, seg_ID);
      }
   }
   else {
      // generate the new points
      i      = (i+1) % Npts;
      seg_ID = findSeg(pt[i], seg_ID, num_segs, seg);
      getNewPoint(p4, i, seg_ID);
      i      = (i+Npts-1) % Npts;
      seg_ID = findSeg(pt[i], ((seg_ID+num_segs-1)%num_segs), num_segs, seg);
      getNewPoint(p3, i, seg_ID);
      i      = (i+Npts-1) % Npts;
      seg_ID = findSeg(pt[i], ((seg_ID+num_segs-1)%num_segs), num_segs, seg);
      getNewPoint(p2, i, seg_ID);
      i      = (i+Npts-1) % Npts;
      seg_ID = findSeg(pt[i], ((seg_ID+num_segs-1)%num_segs), num_segs, seg);
      getNewPoint(p1, i, seg_ID);

      ang2 = get_ang(p1, p2, p3);
      k2   = curvature(p1, p2, p3);
      ang3 = get_ang(p2, p3, p4);
      k3   = curvature(p2, p3, p4);
      len2 = get_len(p2, ang2, k2, p3, ang3, k3);
      do {
         *path_dist += len2;

         // shift points
         p4.x = p3.x; p4.y = p3.y;
         p3.x = p2.x; p3.y = p2.y;
         p2.x = p1.x; p2.y = p1.y;
         ang3 = ang2;
         k3   = k2;
         // generate a new point
         i      = (i+Npts-1) % Npts;
         *j     = (*j+Npts-1) % Npts;
         seg_ID = findSeg(pt[i], ((seg_ID+num_segs-1)%num_segs), num_segs, seg);
         getNewPoint(p1, i, seg_ID);

         ang2 = get_ang(p1, p2, p3);
         k2   = curvature(p1, p2, p3);
         len2 = get_len(p2, ang2, k2, p3, ang3, k3);
      } while (*path_dist < 0.0f);
   }

   len2    = *path_dist / len2;
   retval  = (1.0f-len2) * k2;
   retval +=       len2  * k3;  
   return retval;
}

void SB4PassPath::set_offset(float offset)
{
   wgt = offset / trk_wid;
   if (wgt > 1.0f)
      wgt =  1.0f;
   else if (wgt < -1.0f)
      wgt = -1.0f;
}

// PROTECTED METHODS
void SB4PassPath::getPoint(SB4Vector &ret_pt, SB4Vector &ret_tan,
                           int j, float path_dist, const SB4Vector &p1, 
                           const SB4Vector &p2, float len2, float ang2, float k2,
                           const SB4Vector &p3, float ang3, float k3,
                           const SB4Vector &p4) const
{
   path_dist /= len2;
   k2 = (float)fabs((double)k2);
   k3 = (float)fabs((double)k3);
   if ( (len2 <= SB4_MIN_LEN)||(k2 <= SB4_MIN_CURV)||(k3 <= SB4_MIN_CURV) ) {
      // linear interpolation between p2 and p3
      ret_pt.x = p3.x - p2.x;
      ret_pt.y = p3.y - p2.y;
      ret_pt.scale(path_dist);
      ret_pt.add(p2);

      // get the tangent vector
      ret_tan.x = (float)cos((double)ang2);
      ret_tan.y = (float)sin((double)ang2);
   }
   else {
      float A =  (2.0f*p2.x) - (2.0f*p3.x),
            B = (-3.0f*p2.x) + (3.0f*p3.x),
            C = (float)cos((double)ang2) / k2,
            D =  (2.0f*p2.y) - (2.0f*p3.y),
            E = (-3.0f*p2.y) + (3.0f*p3.y),
            F = (float)sin((double)ang2) / k2,
            t = (float)cos((double)ang3) / k3;
      // calculate spline parameters
      k2 *= 0.5f * k2 * p2.dist2Point2(p3);
      k2  = (float)acos(1.0-(double)k2);
      C  *= k2;
      F  *= k2;
      k2  = 0.5f * k3 * k3 * p2.dist2Point2(p3);
      k2  = (float)acos(1.0-(double)k2);
      t  *= k2;
      A  +=       C  + t;
      B  -= (2.0f*C) + t;
      t   = k2*((float)sin((double)ang3) / k3);
      D  +=       F  + t;
      E  -= (2.0f*F) + t;
         
      // use Hermite splines to interpolate between p2 and p3
      ret_pt.x = P(path_dist, A, B, C, p2.x);
      ret_pt.y = P(path_dist, D, E, F, p2.y);

      // get the tangent vector
      ret_tan.x = dP(path_dist, (3.0f*A), (2.0f*B), C);
      ret_tan.y = dP(path_dist, (3.0f*D), (2.0f*E), F);
      ret_tan.normalize();
    }
}

void SB4PassPath::getNewPoint(SB4Vector &ret_pt, int i, int sID) const
{
   float to_lft, wgt2;

   if (wgt > 0.0f) {    // move to the right
      to_lft = trk_wid;
      if (seg[sID]->rad < 0.0f)  // right turn
         to_lft -= SB4_IN_WID;
      else                       // left turn / straight
         to_lft -= SB4_OUT_WID;
      wgt2 =  wgt;
   }
   else {               // move to the left
      if (seg[sID]->rad > 0.0f)  // left turn
         to_lft = SB4_IN_WID;
      else                       // right turn / straight
         to_lft = SB4_OUT_WID;
      wgt2 = -wgt;
   }

   if (wgt2 > 0.0f) {
      to_lft *= wgt2;
      to_lft += (1.0f-wgt2) * seg[sID]->to_left(pt[i]);
   }

   wgt2 = seg[sID]->get_trk_dist(pt[i]);
   seg[sID]->getPoint(ret_pt, wgt2, to_lft);
}
//-----END PassPath.cpp-----//

//-----START TrkData.cpp-----//
// DESTRUCTOR
SB4TrkData::~SB4TrkData()
{
   if (mainPath != NULL) {
      delete mainPath;
      mainPath = NULL;
   }

   if (passPath != NULL) {
      delete passPath;
      passPath = NULL;
   }
   
   if (seg != NULL) {
      int i;

      for (i = 0; i < num_segs; i++) {
         delete seg[i];
      }
      delete [] seg;
      seg = NULL;
   }

   if (name) {
      free(name);
      name = NULL;
   }
}

// METHODS
void SB4TrkData::get_trk_pos_vel(SB4Line &ret_pv, float to_lft, 
                                 float v, float vn, 
                                 float to_end, int seg_ID) const
{
   float ang;

   // find track normal
   to_end /= seg[seg_ID]->len;
   ang     =       to_end  * seg[seg_ID]->Sdst;
   ang    += (1.0f-to_end) * seg[seg_ID]->Fdst;
   seg[seg_ID]->get_norm(ret_pv, ang);

   // current position
   ret_pv.movePoint(to_lft);

   // current velocity
   ang          = ret_pv.dir.ang();
   ret_pv.dir.x = -vn;
   ret_pv.dir.y = (float)sqrt((double)((v*v) - (vn*vn)));
   ret_pv.dir.rotate(ang);
}

void SB4TrkData::locate_car(int *pidx, float *path_dist, float *to_side, 
                            const SB4Vector &pt, int seg_ID, float *offset) const
{
   SB4Line trk_norm;
   SB4Path *cpath;

   seg[seg_ID]->get_norm(trk_norm, seg[seg_ID]->get_trk_dist(pt));
   
   if ( (offset == NULL)||(*offset == 0.0f) )
      cpath = mainPath;
   else {
      passPath->set_offset(*offset);
      cpath = passPath;
   }
   cpath->locatePoint(pidx, path_dist, to_side, offset, 
                      pt, trk_norm, seg_ID);
}

float SB4TrkData::get_tgt_spd(float *brk_dist, float v, float drag_accel,
                              int pidx, float path_dist, float offset) const
{
   int      cnt, brk;
   float    cur_spd, tgt_spd, brk_spd, nex_spd, curv, Ax, dist, step;
   SB4Path *cpath;

   if (offset == 0.0f)
      cpath = mainPath;
   else {
      passPath->set_offset(offset);
      cpath = passPath;
   }

   cnt     = brk = 0;
   dist    = -v * (float)delta_time;
   step    = (v*MAX_ACCEL*(float)(delta_time*delta_time))/SB4_MIN_LEN;
   Ax      = (float)fabs((double)cpath->get_curv(&pidx, &path_dist));
   cur_spd = tgt_spd = brk_spd = SB4_crn_spd(Ax);
   if (v > 190.0f) {
      step *= 0.0075f * v;
      v    += step;
   }
   while (1) {
      cnt++;
      path_dist += SB4_MIN_LEN;
      curv       = cpath->get_curv(&pidx, &path_dist);
      curv       = (float)fabs((double)curv);
      nex_spd    = SB4_crn_spd(curv);

      if (brk < 0) {
         if (curv < (1.0075f*Ax))
            brk =  1;
      }
      else if (brk > 0) {
         if (nex_spd < (0.8f*brk_spd))
            brk = -1;
      }
      else {
         if (curv > Ax)
            brk = -1;
         else
            brk =  1;
      }
      if (brk < 0) {
         dist   += (float)cnt * SB4_MIN_LEN;
         brk_spd = nex_spd;
         cnt     = 0;
      }

      if (curv > Ax)
         Ax = (0.3f*Ax) + (0.7f*curv);
      else
         Ax = (0.7f*Ax) + (0.3f*curv);
      Ax *= v*v;
      Ax *= -Ax;
      Ax += BRK2;
      if (Ax > 0.0f) {
         Ax = (float)sqrt((double)Ax);
         v  = SB4_vf(v, SB4_MIN_LEN, Ax, ((Ax/MAX_ACCEL)*drag_accel));
      }
      if (v < 5.0f)
         break;
      else if ((v+step) > nex_spd) {
         if (nex_spd > brk_spd)
            nex_spd = brk_spd;
         if (nex_spd < tgt_spd) {
            tgt_spd   = nex_spd;
            *brk_dist = dist;
         }
      }
      Ax = curv;
   }
   if (tgt_spd == cur_spd)
      *brk_dist = 0.0f; // no braking needed
   return tgt_spd;
}

float SB4TrkData::get_curv(int pidx, float path_dist, float offset) const
{
   if (offset == 0.0f)
      return mainPath->get_curv(&pidx, &path_dist);
   else {
      passPath->set_offset(offset);
      return passPath->get_curv(&pidx, &path_dist);
   }
}

void SB4TrkData::set_path(Track *trk)
{
   // remove old data
   this->~SB4TrkData();

   // get RARS track data
   get_trk_data(trk);

   // create the path
   mainPath = new SB4Path(num_segs, seg, trk_wid, name);
   adjust_Path();

   // create the passing path
   passPath = new SB4PassPath(trk_wid, num_segs, seg, *mainPath);
}

// PRIVATE METHODS
void SB4TrkData::get_trk_data(Track *trk)
{
   int i;

   // store track name
   i       = strlen(trk->m_sFileName);
   name    = (char *)malloc(i + 1);
   name[i] = 0;
   for (; i >= 0; i--) {
      name[i] = (char)tolower((int)trk->m_sFileName[i]);
   }

   trk_len   = 0.0f;
   trk_wid   = (float)trk->width;
   num_segs  = trk->NSEG;
   pit_side  = trk->m_iPitSide;
   pit_spd   = (float)trk->m_fPitLaneSpeed + 75.0f;
   pit_entry = (float)trk->m_fPitEntry;
   seg       = new SB4TrkSeg*[num_segs];
   for (i = 0; i < num_segs; i++) {
      if (trk->lftwall[i].radius == 0.0)   // straight
         seg[i] = new SB4TrkSeg_Straight((trk->lftwall + i), trk_len);
      else                                 // turn
         seg[i] = new SB4TrkSeg_Turn((trk->lftwall + i), trk_wid, trk_len);
      trk_len += seg[i]->Fdst - seg[i]->Sdst;
   }
}

void SB4TrkData::adjust_Path()
{
   if (mainPath != NULL)
      mainPath->adjust(SB4_IN_WID, SB4_OUT_WID, trk_wid, num_segs, seg);
}
//-----END TrkData.cpp-----//

//-----START SBDriver.cpp-----//
// GLOBALS
static int        trk_num = -1;
static SB4TrkData trk_data;

// CONSTRUCTORS
SmoothB4::SmoothB4(int i)
   : cV(), A(), start_fuel(0.0f), fuel_per_lap(0.0f), 
     error_last(0.0f), error_avg(0.0f), offset(0.0f), 
     brk_v(0.0f), brk_dist(0.0f), passing(0)
{
#ifndef SB4_DLL
   if ( (0 <= i)&&(i < SB4_CARS) )
      m_sName = SB4_NAME[i+1];
   else
      m_sName = SB4_NAME[0];
   m_sAuthor       = SB4_AUTHOR;
   m_sDescription  = SB4_DESCRIPTION;
   m_iNoseColor    = SB4_NOSE;
   m_iTailColor    = SB4_TAIL;
   m_sBitmapName2D = SB4_BITMAP2D;
   m_sModel3D      = SB4_MODEL3D;
#endif
}

// DESTRUCTOR
SmoothB4::~SmoothB4()
{
   if (trk_num >= 0) {
      // free memory used by the track and path data
      trk_data.~SB4TrkData();
      trk_num = -1;
   }
}

// METHODS
con_vec SmoothB4::drive(situation &s)
{
#ifndef SB4_DLL
   if ( (args.m_iCurrentTrack != trk_num)&&(currentTrack) ) {
      trk_num = args.m_iCurrentTrack;
      BRK2    = SB4_BRK[args.m_iSurface];
      SPD     = SB4_SPD[args.m_iSurface];
      switch (args.m_iSurface) {
#else
   if (!args)
      return cp;
   else if ( (args->m_iCurrentTrack != trk_num)&&(currentTrack) ) {
      trk_num = args->m_iCurrentTrack;
      BRK2    = SB4_BRK[args->m_iSurface];
      SPD     = SB4_SPD[args->m_iSurface];
      switch (args->m_iSurface) {
#endif
      case 2:
         MAX_ACCEL = (float)MYU_MAX2;
         break;
      case 1:
         MAX_ACCEL = (float)MYU_MAX1;
         break;
      default:
         MAX_ACCEL = (float)MYU_MAX0;
      }
      MAX_ACCEL *= (float)g;
      BRK2      *= MAX_ACCEL;
      BRK2      *= BRK2;
      MAX_PWR    = (float)(0.9999 * PM);
      trk_data.set_path(currentTrack);
   }
   else if (!currentTrack)
      return cp;

   if (s.starting) {
      s.side_vision    = 1;
      cp.request_pit   = 0;
      cp.repair_amount = 0;
      fuel_per_lap     = SB4_FC * (float)currentTrack->length;
      cp.fuel_amount   = SB4_fuel_needed(s.stage, s.laps_to_go, 
                                         fuel_per_lap, 0.0f);
      start_fuel       = 0.0f;
      error_last       = 0.0f;
      error_avg        = 0.0f;
      offset           = 0.0f;
      brk_dist         = 0.0f;
      passing          = 0;
   }
   get_RARS_ctrl_params(s);
   return cp;
}

// PRIVATE METHODS
void SmoothB4::get_RARS_ctrl_params(situation &s)
{
   int   pidx;
   float path_dist, to_side, curv, drag_accel, tgt_v, free_lane;

   if ( (s.laps_to_go > 1)&&(!s.starting) ) {
      // Handle fuel
      if (!cp.request_pit) {
         if (s.lap_flag) {
            if (start_fuel > (float)s.fuel) {
               fuel_per_lap *= 0.5f;
               fuel_per_lap += 0.5f*(start_fuel - (float)s.fuel);
            }
            start_fuel = (float)s.fuel;
         }
         cp.fuel_amount  = currentTrack->pit_entry - s.distance;
         cp.fuel_amount /= currentTrack->length;
         if (cp.fuel_amount <= 0.0)
            cp.fuel_amount += 1.0;
         cp.fuel_amount *= (double)fuel_per_lap;  // fuel needed to make it to pits
         if ( ((s.fuel-(double)fuel_per_lap) < cp.fuel_amount)||
              (s.damage > 22500) ) {
            cp.request_pit = 1;  // need to pit
            cp.fuel_amount = SB4_fuel_needed(s.stage, s.laps_to_go, 
                                             fuel_per_lap,
                                      (float)(s.fuel-cp.fuel_amount-1.0));
         }
      }
      else if (s.out_pits) {
         cp.request_pit   = 0;
         cp.repair_amount = 0;
         start_fuel       = 0.0f;
         offset           = 0.7f*(float)trk_data.pit_side*trk_data.trk_wid;
      }
      // Handle damage
      if ( (s.damage > 1000)&&(cp.request_pit) ){
         if (s.damage > 20000)
            cp.repair_amount = 12000;
         else {
            if (cp.fuel_amount > 95.0)
               cp.repair_amount = s.damage;
            else {
               cp.repair_amount = (int)((100.0-cp.fuel_amount) * 200.0);
               if ((int)s.damage > cp.repair_amount)
                  cp.repair_amount = (int)s.damage - cp.repair_amount;
               else
                  cp.repair_amount = 0; // don't repair
            }
         }
      }
   }
   else
      cp.request_pit = 0;  // no pit

   // get current track/path position and velocity
   trk_data.get_trk_pos_vel(cV, (float)s.to_lft, (float)s.v, (float)s.vn,
                            (float)s.to_end, s.seg_ID);
   free_lane = offset;
   trk_data.locate_car(&pidx, &path_dist, &to_side, cV.pt, 
                       s.seg_ID, &free_lane);

   curv = trk_data.get_curv(pidx, path_dist, offset);
   A.y  = (float)fabs((double)to_side);
   if ( (s.seg_ID > 0)||((s.cur_len-s.to_end) > (3.0f*s.v*delta_time)) ) {
      if ( (s.laps_done < 0)&&(A.y > 10.0f)&&(offset == 0.0f) ) {
         // car is off the racing line
         offset   = free_lane + (0.7f * to_side);
         brk_dist = 0.45f * trk_data.trk_wid;
         if (offset < -brk_dist)
            offset = -brk_dist;
         else if (offset > brk_dist)
            offset =  brk_dist;
         brk_dist  = 0.0f;
         error_avg = A.y;
      }
      else if ( (A.y < 4.0f)&&(offset != 0.0f)&&
                (fabs((double)free_lane) < 2.0) ) {
         offset   = 0.0f;  // car is back on the racing line
         brk_dist = 0.0f;  // find new braking point
         passing  = 0;     // end any passing attempts
      }
   }
   else if (A.y > trk_data.trk_wid)
      to_side = -0.7f * error_last;

   drag_accel = SB4_drag_accel(s.fuel, s.damage);
   A.y        = path_dist + (float)(s.v * delta_time);
   A.y        = trk_data.get_curv(pidx, A.y, offset);
   if ( ( (curv < 0.0f)&&(A.y > 0.0f) )||( (curv > 0.0f)&&(A.y < 0.0f) )||
        ( (curv > 0.0f)&&(A.y > 0.0f)&&(A.y > curv) )||
        ( (curv < 0.0f)&&(A.y < 0.0f)&&(A.y < curv) ) )
      curv = A.y;

   if ( (s.to_lft < 0.0)||(s.to_rgt < 0.0) ) {
      // off track
      tgt_v     = SB4_crn_spd((float)fabs((double)curv)) + 10.0f;
      free_lane = 0.0f;
      brk_dist  = 0.0f;
      A.y       = 0.0f;
      if (s.to_lft < 0.0)
         to_side -= 5.0f;
      else
         to_side += 5.0f;
   }
   else {
      // target velocity
      if (brk_dist < 2.0f) {
         brk_v = trk_data.get_tgt_spd(&brk_dist, (float)s.v, drag_accel,
                                      pidx, path_dist, offset);
      }
      else
         brk_dist -= (float)(s.v * delta_time);
      if (brk_dist < 2.0f)
         tgt_v = brk_v;
      else {
         A.x    = BRK2 / (MAX_ACCEL*MAX_ACCEL);
         tgt_v  = (brk_v*brk_v) - (float)(s.v*s.v);
         tgt_v *= (float)delta_time / (A.x*brk_dist);
         tgt_v += (float)s.v;
      }

      // adjust speed if car is not on the path
      if ( ( (curv >  SB4_MIN_CURV)&&(to_side >  10.0f) )||
           ( (curv < -SB4_MIN_CURV)&&(to_side < -10.0f) ) )
         A.x = (float)(10.0 / fabs((double)to_side));
      else if ( ( (curv >  SB4_MIN_CURV)&&(to_side < -10.0f) )||
                ( (curv < -SB4_MIN_CURV)&&(to_side >  10.0f) ) )
         A.x = (float)( 0.1 * fabs((double)to_side));
      else
         A.x = 1.0f;
      if (A.x < 1.0f) {
         if (A.x <= 0.5f)
            tgt_v *= 0.95f;
         else
            tgt_v *= 0.95f + ((0.05f/0.5f) * (A.x-0.5f));
      }
      else if (A.x > 1.0f) {
         if (A.x >= 2.0f)
            tgt_v *= 1.05f;
         else
            tgt_v *= 1.0f + (0.05f * (A.x-1.0f));
      }

      if ( (cp.request_pit)&&(trk_data.pit_entry > (float)s.distance) ) {
         // slow down, when entering the pits
         float dist2pits = trk_data.pit_entry - (float)s.distance,
               pit_brk   = SB4_brkdist((float)s.v, trk_data.pit_spd,
                                    MAX_ACCEL, drag_accel);

         if ( ((pit_brk > dist2pits)&&(tgt_v > trk_data.pit_spd))||
              ((dist2pits < 100.0f)&&(tgt_v < trk_data.pit_spd)) )
            tgt_v = trk_data.pit_spd;
      }
      if (s.laps_to_go <= 0) {
         // race is over
         if (tgt_v > (float)(s.v+0.5))
            tgt_v = (float)(s.v+0.5);
      }

      // avoid other cars
      cV.dir.normalize();
      A.x = anti_collision(s, curv, &to_side, &free_lane);
      if (A.x < tgt_v)
         tgt_v = -A.x;

      // normal acceleration needed to stay on track
      if ( (0.0 < s.to_lft)&&(s.to_lft < 4.5)&&
           (to_side < -0.5f)&&(s.vn > 0.0) ) {
         A.y  = (float)((s.vn*s.vn) / (-2.0*s.to_lft));
         A.y *= (float)((4.5 - s.to_lft) / 10.0);
      }
      else if ( (0.0 < s.to_rgt)&&(s.to_rgt < 4.5)&&
                (to_side > 0.5f)&&(s.vn < 0.0) ) {
         A.y  = (float)((s.vn*s.vn) / ( 2.0*s.to_rgt));
         A.y *= (float)((4.5 - s.to_rgt) / 10.0);
      }
      else
         A.y  = 0.0f;
      if (s.stage != RACING)
         A.y *= 0.25f;  // sacrifice some stability for a little extra speed
   }

   // find out how much normal acceleration is needed
   get_Ay((float)s.v, curv, to_side);

   // get new control parameters
   update_con_vec((float)s.v, tgt_v, (float)SB4_car_mass(s.fuel), drag_accel);

   drag_accel = A.mag() / MAX_ACCEL;
   if (drag_accel < 0.525f) {
      if (free_lane != 0.0f) {
         path_dist += 10.0f + (float)(4.0 * delta_time * s.v);
         curv       = trk_data.get_curv(pidx, path_dist, offset);
         if ( ( (curv >  1.0e-10f)&&(free_lane > 0.0f) )||
              ( (curv < -1.0e-10f)&&(free_lane < 0.0f) ) ) {
            if ( (drag_accel > 0.25f)||(s.v > 110.0) ) {
               if (offset == 0.0f)
                  passing = 0;
               return;
            }
            tgt_v = -20.0f;
         }
         else
            tgt_v =  -5.0f;
            
         if (offset != 0.0f) {
            curv = 0.0f;
            trk_data.locate_car(&pidx, &path_dist, &to_side, cV.pt, 
                                s.seg_ID, &curv);
            free_lane += to_side;
            if ( (free_lane > 0.0f)&&(offset < 0.0f) )
               free_lane = -0.5f;
            else if ( (free_lane < 0.0f)&&(offset > 0.0f) )
               free_lane =  0.5f;
         }

         tgt_v += trk_data.get_tgt_spd(&to_side, (float)s.v, drag_accel,
                                       pidx, path_dist, free_lane);
         if (tgt_v > (float)s.v) {
            offset   = free_lane;
            brk_dist = 0.0f;  // find new braking point
         }
         else if (offset == 0.0f)
            passing  = 0;
      }
      else if ( (!passing)&&(offset != 0.0f)&&
                (drag_accel < 0.5f)&&(error_avg < 0.75f) ) {
         // ease back onto the racing line
         if (offset > 2.0f)
            offset -= 2.0f;
         else if (offset < -2.0f)
            offset += 2.0f;
         else
            offset  = 0.0f;
         brk_dist = 0.0f;
      }
   }
   else if ( (offset == 0.0f)&&(free_lane != 0.0f)&&(passing) )
      passing = 0;
}

void SmoothB4::get_Ay(float v, float curv, float error)
{
   float Ay = (float)fabs((double)error);

   error_last += error;             // change in error
   error_last /= (float)delta_time; // error derivative
   if (Ay >= error_avg)
      error_avg = Ay;
   else {
      error_avg  *= 0.7f;
      error_avg  += 0.3f * Ay;      // running average
   }

   // calculate acceleration
   if (error_avg < 5.0f) {
      error_last *= 0.3f + (error_avg / 20.0f);
      Ay          = 0.04f * error_avg;
   }
   else if (error_avg < 40.0f) {
      error_last *= 0.55f + ((error_avg-5.0f) / 77.778f);
      Ay          = 1.0f / error_avg;
   }
   else
      Ay          = 1.0f / 40.0f;
   Ay  *= 0.005f * (error + error_last);
   Ay  += curv;
   Ay  *= v * v;
   A.y += Ay;

   error_last = -error;             // store until next iteration

   // make sure acceleration is within limits
   if (A.y > MAX_ACCEL)
      A.y =  MAX_ACCEL;
   else if (A.y < -MAX_ACCEL)
      A.y = -MAX_ACCEL;

   // calculate maximum A.x
   A.x = (MAX_ACCEL*MAX_ACCEL) - (A.y*A.y);
   if (A.x <= 0.0f)
      A.x = 0.0f;
   else
      A.x = (float)sqrt((double)A.x);
}

void SmoothB4::update_con_vec(float v, float tgt_v, float mass,
                              float drag_accel)
{
   SB4Vector W;
   int       traffic = 0;
   float     max_Ax  = A.x;

   if (tgt_v < 0.0f) {
      traffic = 1;
      tgt_v   = -tgt_v;
   }
   // calculate acceleration
   A.x = (tgt_v - v) / (float)delta_time;

   if (A.x >= 0.0f) {    // acceleration
      int   i;
      float Ay, pwr, pwr_H, Ax_H,
            pwr_L = 0.0f,
            Ax_L  = 0.0f;

      A.x += drag_accel * v * v;
      if (max_Ax < (0.015f*MAX_ACCEL))
         max_Ax = 0.015f * MAX_ACCEL;
      if (A.x > max_Ax)
         A.x = max_Ax;  // do not exceed tire limit

      Ay     = A.y;
      max_Ax = A.mag();
      if (max_Ax > MAX_ACCEL)
         A.y = (float)sqrt((double)((MAX_ACCEL*MAX_ACCEL) - (A.x*A.x)));

      for (i = 0; ; i++) {
         // find -W vector
         max_Ax = A.mag();
         W.x    = A.x;
         W.y    = A.y;
         W.scale(SB4_slip(max_Ax/(float)g) / max_Ax);
         W.x   += v;
         if (i == 40)
            break;      // stop search, ran out of attempts

         // calculate power needed
         pwr = mass * W.dot(A);
         if (pwr <= MAX_PWR) { // under target
            if ( ((pwr/MAX_PWR) >= 0.99f)||(!i)||(i == 39) )
               break;   // stop search, because we are within limits
            pwr_L = pwr;
            Ax_L  = A.x;
         }
         else {               // over target
            pwr_H = pwr;
            Ax_H  = A.x;
         }
         A.x  = Ax_L;
         A.x += ((MAX_PWR-pwr_L)*(Ax_H-Ax_L)) / (pwr_H-pwr_L);
      }

      if (Ay != A.y) {
         W.x    = A.x;
         W.y    = A.y = Ay;
         max_Ax = A.mag();
         if (max_Ax > MAX_ACCEL) {
            tgt_v  = MAX_ACCEL / max_Ax;
            max_Ax = MAX_ACCEL;
            A.scale(tgt_v);
         }
         else
            tgt_v = 1.0f;
         W.scale(tgt_v * (SB4_slip(max_Ax/(float)g) / max_Ax));
         W.x   += v;
      }
   }
   else {               // deceleration
      W.x = (float)fabs((double)error_last);
      if ( (traffic)||(W.x >= 25.0f) )
         max_Ax = 1.25f * MAX_ACCEL;
      else {
         if (W.x > 10.0f)
            W.x = 0.55f + ((0.7f/15.0f)*(W.x-10.0f));
         else if (W.x > 5.0f)
            W.x = 0.35f + ((0.2f/ 5.0f)*(W.x- 5.0f));
         else
            W.x = 0.35f;
         W.x *= MAX_ACCEL;
         if (max_Ax < W.x)
            max_Ax = W.x;
      }
      if (A.x < -max_Ax)
         A.x = -max_Ax; // do not exceed tire limit

      // find -W vector
      W.x    = A.x;
      W.y    = A.y;
      max_Ax = A.mag();
      if (max_Ax > MAX_ACCEL) {
         tgt_v  = MAX_ACCEL / max_Ax;
         max_Ax = MAX_ACCEL;
         A.scale(tgt_v);
      }
      else
         tgt_v = 1.0f;
      W.scale(tgt_v * (SB4_slip(max_Ax/(float)g) / max_Ax));
      W.x   += v;
   }
   // get control parameters
   cp.alpha = atan((double)(W.y / W.x));
   cp.vc    = W.mag();
}

float SmoothB4::anti_collision(situation &s, float curv, float *error,
                               float *free_lane)
{
   int       j, pass_idx;
   float     r_ang, min_dist, margin, acc, lx, rx, s_lft, s_rgt,
             to_side, err;
   SB4Vector rp, rv;

   err = (float)(2.5*CARWID);

   *free_lane = acc = 0.0f;
   s_lft = (float)s.to_lft - SB4_IN_WID;
   s_rgt = (float)s.to_rgt - SB4_IN_WID;
   for (j = 0, pass_idx = 0; j < NEARBY_CARS; j++) {
      if (s.nearby[j].who >= MAX_CARS)
         continue;
      else if (s.nearby[j].who == (passing - 1))
         pass_idx = j + 1;

      // other car's relative position
      rp.x  = (float)s.nearby[j].rel_x;
      rp.y  = (float)s.nearby[j].rel_y;
      r_ang = -rp.ang();
      rp.rotate((float)(-s.alpha));

      if ( (err != 0.0f)&&
           ((float)(-0.6*CARLEN) < rp.y)&&(rp.y < (float)(0.8*CARLEN)) ) {
         if ( (-err < rp.x)&&(rp.x < err) ) {
            *error *= 0.1f + (0.34f*(float)(fabs((double)rp.x) / CARWID));
            err     = 0.0f;
         }
      }

      // look for free space around the car
      if (s.nearby[j].to_rgt < 0.0f)
         to_side = (float)s.to_rgt;
      else if (s.nearby[j].to_lft < 0.0f)
         to_side = (float)(-s.to_lft);
      else
         to_side = (float)(s.to_rgt - s.nearby[j].to_rgt);
      if (rp.y < (float)(4.0*CARLEN)) {
         lx = (float)(-1.25*CARWID) - to_side;
         rx = (float)(-1.25*CARWID) + to_side;
         if ( (0.0f < lx)&&(lx < s_lft) ) 
            s_lft = lx;
         else if ( (0.0f < rx)&&(rx < s_rgt) ) 
            s_rgt = rx;
      }

      if ( (rp.y < -5.0f)||
           ((rp.y < 3.0f)&&(rp.mag() < (float)CARLEN))||
           ((rp.y < 5.0f)&&(s.nearby[j].to_lft < -1.0))||
           ((rp.y < 5.0f)&&(s.nearby[j].to_rgt < -1.0)) )
         continue;   // already passed this car

      // other car's relative velocity
      rv.x = (float)s.nearby[j].rel_xdot;
      rv.y = (float)s.nearby[j].rel_ydot;
      rv.rotate(r_ang);

      lx = (float)(s.vn - s.nearby[j].vn);
      if ( (!pass_idx)&&(rp.y > (float)(CARLEN-2.0)) ) {
         rx = (float)(fabs((double)rp.x) / CARWID);
         if ( (rx < 2.5f)&&
              ((s.nearby[j].to_lft < 0.0)||(s.nearby[j].to_rgt < 0.0)) )
            rx = (float)(5.0 * s.v * delta_time);
         else if (rx < 1.25f)
            rx = (float)(2.0 * s.v * delta_time) * (1.0f - (rx/5.0f));
         else
            rx = -200.0f;
         rx += (float)CARLEN;

         if (rp.y < rx) {
            if ( (s.nearby[j].to_lft < 0.0)||(s.nearby[j].to_rgt < 0.0)||
                 ((fabs(s.vn) < 5.0f)&&(fabs((double)lx) < 10.0)) )
               pass_idx = j + 1; // found a car to pass
         }
      }

      if ( (rp.y < (float)(2.0*CARLEN))&&
           ( (( lx > 2.5f)&&( rp.x > (float)CARLEN))||
             ((-lx > 2.5f)&&(-rp.x > (float)CARLEN))||
             ((fabs((double)rp.x) > CARWID)&&
              ((float)fabs((double)curv) < (0.25f*SB4_MIN_CURV))) ) )
         continue;   // almost passed this car
      else if (rv.x >= -2.0f)
         continue;   // not approaching this car

      if ( (s.nearby[j].to_lft < 0.0)||(s.nearby[j].to_rgt < 0.0)||
           (rv.x <= -20.0f) )
         margin = 1.0f + (20.0f/25.0f);
      else
         margin = 1.0f - (rv.x/25.0f);

      margin *= (float)CARLEN;
      if (s.nearby[j].braking)
         margin -= 0.1f * rv.x;
      rx = (float)fabs(cos((double)r_ang + s.alpha));
      if ( ( (s.cen_a >  5.0)&&(-rp.x > (float)CARWID) )||
           ( (s.cen_a < -5.0)&&( rp.x > (float)CARWID) ) )
         margin -= 4.0f * rx * rv.x;
      else if ( ( (s.cen_a >  10.0)&&( rp.x > (float)CARWID) )||
                ( (s.cen_a < -10.0)&&(-rp.x > (float)CARWID) ) )
         margin  = 0.5f * (margin + (float)CARLEN);

      rp.rotate(r_ang + (float)s.alpha);

      min_dist = (float)fabs((double)((rv.y * rp.x) / rv.mag()));
      if (min_dist >= margin)
         continue;   // not going to hit this car

      // since we are going to hit this car, reduce impact speed
      if (rp.x <= margin) {
         if (min_dist < CARLEN)
            margin = (float)CARLEN;
         else
            margin = min_dist;
         margin += rp.x;
         margin *= 0.5f;
      }
      min_dist *= min_dist;
      rp.y      = (float)sqrt((double)((rp.x*rp.x) - min_dist));
      rp.y     -= (float)sqrt((double)((margin*margin) - min_dist));
      if (rp.y <= 0.0f) {
         if (rp.x > (float)CARLEN)
            rp.y = MAX_ACCEL;
         else
            rp.y = 0.0f;
      }
      else {
         rp.y = ((rv.x*rv.x) - 4.0f) / rp.y;
         if (rp.y > MAX_ACCEL)
            rp.y = MAX_ACCEL;
      }

      if (rp.y > (0.8f*MAX_ACCEL)) {
         if (s.nearby[j].rel_y < CARLEN)
            rp.y *= 0.6f + (0.4f * (float)((s.nearby[j].rel_y+5.0)/(CARLEN+5.0)));
         rv.x += rp.y * (float)delta_time;
         rv.rotate(-r_ang);
         rv.x  = (float)((s.nearby[j].rel_ydot-(double)rv.y-s.v) / delta_time);
         if (rv.x < acc)
            acc = rv.x;
      }
   }

   if ( (pass_idx)&&(passing >= 0) ) {
      j = pass_idx - 1;
      if ( (offset == 0.0f)||
           ((s.nearby[j].rel_y < (2.0*CARLEN))&&
            (fabs((double)*error) < 10.0)) ) {
         if (s.nearby[j].to_rgt < 0.0f)
            to_side = (float)s.to_rgt;
         else if (s.nearby[j].to_lft < 0.0f)
            to_side = (float)(-s.to_lft);
         else
            to_side = (float)(s.to_rgt - s.nearby[j].to_rgt);
         lx = (float)(-1.25*CARWID) - to_side;
         rx = (float)(-1.25*CARWID) + to_side;
         if ( (lx <= 0.0f)&&(rx <= 0.0f) ) {
            if (curv > 1.0e-10f)       // move to the left
               *free_lane = (s.nearby[j].to_lft > CARLEN) ?  rx : 0.0f;
            else if (curv < -1.0e-10f) // move to the right
               *free_lane = (s.nearby[j].to_rgt > CARLEN) ? -lx : 0.0f;
            else if (s.nearby[j].to_lft > (1.2*s.nearby[j].to_rgt))
               *free_lane =  rx;       // move to the left
            else if ((1.2*s.nearby[j].to_lft) < s.nearby[j].to_rgt)
               *free_lane = -lx;       // move to the right
            else if ( (s.nearby[j].vn <= -10.0f)||
                      ( (s.nearby[j].vn < 10.0f)&&(rx > lx) ) )
               *free_lane =  rx;       // move to the left
            else
               *free_lane = -lx;       // move to the right
         }
         else if (to_side < 0.0f)
            *free_lane = (curv > 0.0f) ?  rx : 0.0f; // move to the left
         else
            *free_lane = (curv < 0.0f) ? -lx : 0.0f; // move to the right

         if ( ( (offset > 0.0f)&&(*free_lane < 0.0f) )||
              ( (offset < 0.0f)&&(*free_lane > 0.0f) ) )
            *free_lane = 0.0f;
         else if (*free_lane != 0.0f) {
            if ( (0.0f < *free_lane)&&(*free_lane < 3.0f) )
               *free_lane =  3.0f;
            else if ( (0.0f > *free_lane)&&(*free_lane > -3.0f) )
               *free_lane = -3.0f;

            // make sure there is room to make the pass
            if ( (*free_lane < 0.0f)&&(*free_lane < -s_lft) )
               *free_lane = (s_lft >= 5.0f) ? -s_lft : 0.0f;   // no space to pass
            else if ( (*free_lane > 0.0f)&&(*free_lane > s_rgt) )
               *free_lane = (s_rgt >= 5.0f) ?  s_rgt : 0.0f;   // no space to pass 
         }
      }
      if ( (offset != 0.0f)||(*free_lane != 0.0f) )
         passing = s.nearby[j].who + 1;
      else
         passing = 0;
   }
   else if (passing > 0)
      passing = 0;   // end passing, since no cars are around

   if ( (acc != 0.0f)&&(s.v > 20.0) ) {
      rv.y = (float)s.v + (acc * (float)delta_time);
      if (rv.y < 15.0f)
         return 15.0f;
      return rv.y;
   }
   return (SB4_MAX_SPD + 10.0f);
}
//-----END SBDriver.cpp-----//

// STATIC FUNCTIONS
//-----START spd_func.cpp-----//
float SB4_crn_spd(float curv)
{
   float retval;

   if (curv == 0.0f)
      return SB4_MAX_SPD;

   retval = SPD * (float)sqrt((double)(MAX_ACCEL / curv));
   if (retval > SB4_MAX_SPD)
      return SB4_MAX_SPD;
   return retval;
}

float SB4_brkdist(float vi, float vf, float max_a, float drag_accel)
{
   if (vf >= vi)
      return 0.0f;

   if (drag_accel > 0.0f) {
      vi *= vi * drag_accel;
      vi += max_a;
      vf *= vf * drag_accel;
      vf += max_a;
      return ( (float)log((double)(vf/vi)) / (-2.0f*drag_accel) );
   }
   else {
      vi *= vi;
      vf *= vf;
      if (max_a < 0.001f)
         max_a = 0.001f;
      return ( (vi - vf) / (2.0f*max_a) );
   }
}

float SB4_vf(float vi, float brkdist, float max_a, float drag_accel)
{
   if (drag_accel > 0.0f) {
      vi *= vi * drag_accel;
      vi += max_a;
      vi *= (float)exp(-2.0*(double)(drag_accel*brkdist));
      vi -= max_a;
      if (vi <= 0.0f)
         return 0.0f;
      return (float)sqrt((double)(vi / drag_accel));
   }
   else {
      vi *= vi;
      vi -= 2.0f*max_a*brkdist;
      if (vi <= 0.0f)
         return 0.0f;
      return (float)sqrt((double)vi);
   }
}

// inverse friction function
float SB4_slip(float u)
{
   double retval;

#ifndef SB4_DLL
   switch (args.m_iSurface) {
#else
   switch (args->m_iSurface) {
#endif
   case 2:     // surface == 2 - (asphalt, F1 tyres)
      if (u >= (float)(0.9999546*MYU_MAX2))
         retval = 19.99999691;
      else
         retval = -SLIPPING * log(1.0 - ((double)u/MYU_MAX2));
      break;
   case 1:     // (harder surface, ordinary cars)
      if (u >= (float)MYU_MAX1)
         retval = 20.0;
      else
         retval = -SLIPPING * log(1.0 - ((double)u/MYU_MAX1));
      break;
   default:    // (very loose surface)
      if (u >= (float)MYU_MAX0)
         retval = 20.0;
      else
         retval = ((SLIPPING*(double)u)/MYU_MAX0) / (1.0 - ((double)u/MYU_MAX0));
   }
   return (float)retval;
}

double SB4_fuel_needed(Stage s, int lapsleft, float fpl, float cur_fuel)
{
   float retval;

   if (fpl <= 0.0f)
      return MAX_FUEL;
   else if (s == PRACTICE)
      retval = fpl * 10.0f;
   else if ( (s == QUALIFYING)&&(lapsleft >= 5) )
      retval = fpl *  5.0f;
   else
      retval = fpl * (float)lapsleft;

   retval -= cur_fuel;
   if (retval < 0.0f)
      return 0.0f;
   else if (retval > (float)MAX_FUEL)
      return MAX_FUEL;
   return ((double)retval);
}
//-----END spd_func.cpp-----//

Driver *getSBv4r0Instance()
{
   static int i = -SB4_USE_DEFAULT_NAME;
   
   return new SmoothB4(i++);
}

#undef SB4_car_mass
#undef SB4_drag
#undef SB4_drag_accel
#undef SB4_CARS
#undef SB4_USE_DEFAULT_NAME
