/*
 * SmoothB8.cpp
 * Dennis Lew (dennis_L76@yahoo.ca)
 *
 * ROBOT     : SmoothB3
 * DATA FILE : None
 *
 * Uses the OOP structure, call 'getSBv3r0Instance()' to get a
 * new instance.  All instances of SmoothB3 in a race will share
 * the same path data.
 *    
 * SmoothB will try and create its own path around the track its
 * on.  This routine will create a good path for all RARS tracks.
 * The data file can still be used to provide better paths if
 * necessary.
 *
 * If you want to use any part of this code in your own robot, or
 * are curious about anything I've done here, just let me know.
 *
 * Tracks this robot has been tested on the following tracks (71):
 *   aalborg.trk    adelaide.trk   albert.trk     albrtprk.trk
 *   anew.trk       austria.trk    barcelon.trk   brands.trk
 *   brazil.trk     buenos.trk     clkwis.trk     cstlcomb.trk
 *   doningtn.trk   dreiw.trk      elev2.trk      estoril.trk
 *   figure8.trk    fiorano.trk    fourmile.trk   hock.trk
 *   hungary.trk    imola.trk      indy500.trk    indygp.trk
 *   jerez.trk      kevin.trk      kimla01.trk    kimla02.trk
 *   kimla03.trk    kimla04.trk    kimla05.trk    longrand.trk
 *   loudon.trk     magnycrs.trk   michigan.trk   midohio.trk
 *   mlwaukee.trk   monaco.trk     montreal.trk   monza-55.trk
 *   monza-76.trk   mosport.trk    nazareth.trk   nurnburg.trk
 *   oval2.trk      phoenix.trk    pocono.trk     ra.trk
 *   ramdu.trk      rars.trk       s1.trk         s2.trk
 *   sepang(1).trk  silver97.trk   silverst.trk   spa.trk
 *   speed2.trk     stef2.trk      suzuka.trk     t_eight.trx
 *   t_spa.trx      t_zero.trx     timin.trk      tremblnt.trk
 *   trouble.trk    v01.trk        v02.trk        v03.trk
 *   watglen.trk    wierd.trk      zandvort.trk
 *
 * <CHANGES>
 * v1.4: -fixed NULL pointer bug when trying to load data
 *       -improved collision detection
 *       -robot will not actively try to pass other cars (UNDER DEVELOPMENT)
 *       -robot should be able to reliably finish races now
 *
 * v1.5: -bugs introduced by earlier improvements have been fixed
 *       -robot runs better on tracks with no data
 *
 * v1.6: -complete rewrite of the anti-collision code (MUCH BETTER)
 *
 * v2.0: -complete rewrite of the path creation routine (MUCH BETTER, BUT NEEDS MORE DEVELOPMENT)
 *       -new data file format (SMALLER THAN OLD VERSION)
 *       -improved control routine
 *       -improved anti-collision code
 *       -functional passing code (FINALLY!)
 *
 * v2.1: -changes for RARS v0.9
 *
 * v2.2: -defaults to using a constant string for its name
 *       -better control parameters for surfaces 0 and 2
 *       -small adjustments to anti-collision code
 *
 * v2.3: -rewrite of the path creation routine (WORKS ON ALL TRACKS!)
 *       -this routine takes longer than the last one:
 *           Avg Time (P200MMX): v2.2 =  0.3s
 *                               v2.3 =  1.3s
 *           Max Time (P200MMX): v2.2 = 17.7s
 *           (kimla04.trk)       v2.3 = 20.5s
 *           longrand.trk Time : v2.2 = 1min (cannot complete lap)
 *                               v2.3 = 9min
 *       -in most cases, this routine will create a faster path
 *       -small adjustments to the control routine, for better stability
 *
 * v3.0: -added elliptical components to the path for better performance
 *       -the routine is not efficient, it adds about 0.7s/turn to the
 *        path creation routine (STILL NEEDS WORK)
 *       -made robot more aggresive in traffic
 *
 * v3.1: -various improvements to the anti-collision code 
 *        (back-ported to v2.3)
 *       -removed an unused member in the SBv3r0 class.
 *       -replaced fuel and damage handling code.
 *
 * v3.2: -new and improved anti-collision code
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "os.h"
#include "car.h"
#include "track.h"

// Set the following macro to the number of SmoothB
// robots that will be used.
#define SB3_CARS 2
// Set the following macro to 1 if the default name
// will be used, or 0 if not
#define SB3_USE_DEFAULT_NAME 1


// CONSTANTS
static const char   *SB3_AUTHOR           = "Dennis Lew (dennis_L76@yahoo.ca)";
static const char   *SB3_DESCRIPTION      = "v3.2: Apr 25, 2003";
static const char   *SB3_NAME[1+SB3_CARS] = {"SmoothB3",    // the default name
                                             "SBv3r2-1",    // names for the other robots that may be used
                                             "SBv3r2-2"};
static const int     SB3_NOSE             = oLIGHTGRAY;
static const int     SB3_TAIL             = oGREEN;
static const char   *SB3_BITMAP2D         = "car_gray_green";
static const char   *SB3_MODEL3D          = NULL;

static const char   *DATA_FILE     = "SBv3.dat";
static const double  MY_PI         =    3.14159265358979323846;
static const double  MY_PI21       =    2.0 * MY_PI;
static const double  MY_PI12       =    0.5 * MY_PI;
static const double  MY_PI14       =    0.25 * MY_PI;
static const double  SCALE_CORNSPD =    0.98;
static const double  SB3_MAX_SPD    =  400.0;
static const double  SB3_FC         =    0.0003;  // fuel / ft
static const double  SB3_MARGIN     =   (0.5 * CARLEN) + 0.5;
static const double  SB3_CLOSE_RATE =   20.0;
static const double  SB3_IN_WID     =    2.0;
static const double  SB3_OUT_WID    =    5.5;
static const double  SB3_PATH_ERR   =    0.5;
static const double  SB3_MIN_LEN    =   10.0;
static const int     SB3_NUM_RNDS   =  300;
static const int     SB3_SEARCH_LEN =    6;

// GLOBALS
static double        MAX_ACCEL;


// STATIC FUNCTIONS
//-----START spd_func.h-----//
#define SB3_car_mass(fuel) \
   (M + (fuel/g))
#define SB3_drag(damage) \
   (DRAG_CON * (1.0 + ((double)(damage << 1) / (double)MAX_DAMAGE)))
#define SB3_drag_accel(fuel, damage) \
   (SB3_drag(damage) / SB3_car_mass(fuel))

static double SB3_crn_spd(double radius);
static double SB3_brkdist(double vi, double vf, double drag_accel);
static double SB3_iu(double cur_u, int surfaceType);
static double SB3_fuel_needed(Stage s, int lapsleft, double fpl,
                             double cur_fuel);
//-----END spd_func.h-----//


// CLASSES
//-----START Primitive.h-----//
class SB3Primitive
{
public:
   virtual double dist2Point(const SB3Primitive &chk_pt) const                = 0;
   virtual int    intersect(SB3Primitive &ret_pt, const SB3Primitive &l) const = 0;
   
protected:
   // CONSTRUCTOR
   SB3Primitive() {}
};
//-----END Primitive.h-----//

//-----START Vector.h-----//
class SB3Line;
class SB3Vector : public SB3Primitive
{
public:
   // MEMBERS
   double x, y;

   // CONSTRUCTORS
   SB3Vector();
   SB3Vector(double vx, double vy);
   SB3Vector(double angle);
   SB3Vector(const SB3Vector &v);
   
   // METHODS
   void   add(const SB3Vector &v);
   void   sub(const SB3Vector &v);
   double mag() const;
   double mag2() const;
   double dist2Point(const SB3Vector &chk_pt) const;
   double ang() const;
   double ang(const SB3Vector &v) const;
   void   scale(double factor);
   void   normalize();
   double dot(const SB3Vector &v) const;
   void   rotate(double angle);
   int    intersect(SB3Vector &ret_pt, const SB3Line &l) const;

   double dist2Point(const SB3Primitive &chk_pt) const
      {return dist2Point((const SB3Vector&)chk_pt);}
   int    intersect(SB3Primitive &ret_pt, const SB3Primitive &l) const
      {return intersect((SB3Vector&)ret_pt, (const SB3Line&)l);}
};
//-----END Vector.h-----//

//-----START Line.h-----//
class SB3Line : public SB3Primitive 
{
public:
   // MEMBERS
   SB3Vector pt;
   SB3Vector dir;

   // CONSTRUCTORS
   SB3Line();
   SB3Line(double px, double py, double dx, double dy);
   SB3Line(double px, double py, double d_ang);
   SB3Line(const SB3Line &l);

   // METHODS
   int    intersect(SB3Vector &ret_pt, const SB3Line &l) const;
   void   getPoint(SB3Vector &ret_pt, double dist) const;
   void   movePoint(double dist);
   int    movePoint(const SB3Line &l);
   double dist2Point(const SB3Vector &chk_pt) const;

   double dist2Point(const SB3Primitive &chk_pt) const
      {return dist2Point((const SB3Vector&)chk_pt);}
   int    intersect(SB3Primitive &ret_pt, const SB3Primitive &l) const
      {return intersect((SB3Vector&)ret_pt, (const SB3Line&)l);}
};
//-----END Line.h-----//

//-----START Circle.h-----//
class SB3Circle : public SB3Primitive
{
public:
   static const int TAN_1, TAN_2, TAN_3, TAN_4;

   // MEMBERS
   SB3Vector c;
   double   R;

   // CONSTRUCTORS
   SB3Circle();
   SB3Circle(double rad);
   SB3Circle(double cx, double cy, double rad);
   SB3Circle(const SB3Circle &cc);

   // METHODS
   int    intersect(SB3Vector &ret_pt, const SB3Line &l) const;
   int    intersect(SB3Vector &ret_pt1, SB3Vector &ret_pt2, 
                    const SB3Line &l) const;
   int    intersect(SB3Vector &ret_pt1, SB3Vector &ret_pt2,
                    const SB3Circle &cc) const;
   double dist2Point(const SB3Vector &chk_pt) const;
   void   getPoint(SB3Vector &ret_pt, double angle) const;
   double ang(SB3Vector &pt) const;
   int    tangent(SB3Line &ret_l, const SB3Circle &c2, const int lineID) const;

   double dist2Point(const SB3Primitive &chk_pt) const
      {return dist2Point((const SB3Vector&)chk_pt);}
   int    intersect(SB3Primitive &ret_pt, const SB3Primitive &l) const
      {return intersect((SB3Vector&)ret_pt, (const SB3Line&)l);}
};
//-----END Circle.h-----//

//-----START Ellipse.h-----//
class SB3Ellipse : public SB3Primitive 
{
public:
   static const int TAN_1, TAN_2, TAN_3, TAN_4;

   // MEMBERS
   SB3Vector c;
   double   Rx, Ry, r_ang;

   // CONSTRUCTORS
   SB3Ellipse();
   SB3Ellipse(const SB3Circle &cc);
   SB3Ellipse(const SB3Ellipse &ce);

   // METHODS
   double R(double angle) const;
   void   getPoint(SB3Vector &ret_pt, double angle) const;
   void   getTangent(SB3Vector &ret_tan, double angle) const;
   double getCurvature(double angle) const;
   double getArcLength(double start_ang, double len_ang) const;
   double ang(double ArcLength, double start_ang, int dir) const;
   double ang(SB3Vector &pt) const;
   double dist2Point(const SB3Vector &chk_pt) const;
   int    intersect(SB3Vector &ret_pt, const SB3Line &l) const;
   int    intersect(SB3Vector &ret_pt1, SB3Vector &ret_pt2, 
                    const SB3Line &l) const;
   int    intersect(SB3Vector &ret_pt1, SB3Vector &ret_pt2,
                    SB3Vector &ret_pt3, SB3Vector &ret_pt4, 
                    const SB3Circle &cc) const;
   int    tangent(SB3Line &ret_l, const SB3Circle &c2, int lineID) const;
   int    tangent(SB3Line &ret_l, const SB3Ellipse &e2, int lineID) const;

   double dist2Point(const SB3Primitive &chk_pt) const
      {return dist2Point((const SB3Vector&)chk_pt);}
   int    intersect(SB3Primitive &ret_pt, const SB3Primitive &l) const
      {return intersect((SB3Vector&)ret_pt, (const SB3Line&)l);}
private:
   // PRIVATE METHODS
   double find_intersect(SB3Vector &ret_pt, const SB3Circle &cc,
                         double start, double end, double delta) const;
   double find_discon(double A, double B, double C, double D,
                      double start, double end, double delta,
                      const SB3Ellipse &e2, int i) const;
   int    find_tan(double A, double B, double C, double D,
                   double start, double end, double delta, SB3Line &ret_l,
                   const SB3Ellipse &e2, int lineID, int i) const;
};
//-----END Ellipse.h-----//

//-----START TrkSeg.h-----//
class SB3TrkSeg
{
public:
   // MEMBERS
   double       Sdst, Fdst;  // Start and End locations (measured on centre line of trk)
   double       len, rad;
   SB3Primitive *left_side;
   
   // METHODS
   virtual void   get_norm(SB3Line &ret_n, double trk_dist) const = 0;
   virtual double get_trk_dist(const SB3Vector &chk_pt) const     = 0;
   virtual int    path_ok(const SB3Line &path, double S, double F,
                          double min_w, double max_w) const      = 0;
   virtual int    path_ok(const SB3Circle &path, double S, double F,
                          double min_w, double max_w) const      = 0;
   virtual int    path_ok(const SB3Ellipse &path, double S, double F,
                          double min_w, double max_w) const      = 0;

   int    same_dir(const SB3TrkSeg &seg) const;
   double to_left(const SB3Vector &chk_pt) const;
   
protected:
   // CONSTRUCTORS
   SB3TrkSeg(int ID);
};
//-----END TrkSeg.h-----//

//-----START TrkSeg_Straight.h-----//
class SB3TrkSeg_Straight : public SB3TrkSeg
{
public:
   // MEMBERS
   SB3Line   ls;
   SB3Vector norm;
   
   // CONSTRUCTORS
   SB3TrkSeg_Straight(int ID, segment *seg, double trk_dist);
   
   // METHODS
   void   get_norm(SB3Line &ret_n, double trk_dist) const;
   double get_trk_dist(const SB3Vector &chk_pt) const;
   int    path_ok(const SB3Line &path, double S, double F,
                  double min_w, double max_w) const;
   int    path_ok(const SB3Circle &path, double S, double F,
                  double min_w, double max_w) const;
   int    path_ok(const SB3Ellipse &path, double S, double F,
                  double min_w, double max_w) const;
};
//-----END TrkSeg_Straight.h-----//

//-----START TrkSeg_Turn.h-----//
class SB3TrkSeg_Turn : public SB3TrkSeg
{
public:
   // MEMBERS
   SB3Circle ls;
   double   beg_ang, end_ang;
   int      left;
   
   // CONSTRUCTORS
   SB3TrkSeg_Turn(int ID, segment *seg, double wid, double trk_dist);
   
   // METHODS
   void   get_norm(SB3Line &ret_n, double trk_dist) const;
   double get_trk_dist(const SB3Vector &chk_pt) const;
   int    path_ok(const SB3Line &path, double S, double F,
                  double min_w, double max_w) const;
   int    path_ok(const SB3Circle &path, double S, double F,
                  double min_w, double max_w) const;
   int    path_ok(const SB3Ellipse &path, double S, double F,
                  double min_w, double max_w) const;
   int    like_straight() const;
};
//-----END TrkSeg_Turn.h-----//

//-----START PathComponent.h-----//
class SB3PathComponent 
{
public:
   static const short TURN, STRAIGHT, TRANS;

   // MEMBERS
   short         typeID, PathDir; // type of component
   double        len;
   double        Sdst, Fdst;  // Start and End locations (measured on centre line of trk)
   int           num_segs;    // number of track segments this component covers
   SB3TrkSeg    **seg;         // track segments (data is taken from left wall only)
   SB3Primitive  *comp;        // the actual path component

   SB3PathComponent **pass;    // alternate path route for passing
   double            pre_len, post_len, len_trk;
   
   // DESTRUCTORS
   ~SB3PathComponent();

   // METHODS
   virtual double get_spd(double trk_dist, double *max=0) const = 0;
   virtual void   get_tan(SB3Vector &ret_pt, SB3Line &ret_norm) const = 0;
   virtual double get_path_dist(const SB3Vector &pt) const = 0;
   virtual double get_dist2path(const SB3Vector &pt) const = 0;
   virtual void   getPoint2(SB3Vector &ret_pt, SB3Vector &ret_tan, 
                            double path_dist) const = 0;
   virtual double brk_dist(double dist, double v) const = 0;

   void   copy(const SB3PathComponent &pc);
   int    getPoint(SB3Vector &ret_pt, SB3Line &ret_norm, 
                   double trk_dist, int ret_tan=0) const;
   double to_side(SB3Vector &ret_pt, SB3Line &ret_norm, 
                  double trk_dist, int ret_tan=0) const;
   void   insert_TrkSeg(SB3TrkSeg *ts, int idx);
   void   insert_TrkSeg(const SB3PathComponent &pc, int add_to_tail);
   void   remove_TrkSeg(int idx);
   void   trim();
   int    find_trk_seg(const SB3Vector &pt, double trk_wid,
                       int start=0, int dir=1) const;
   int    find_trk_seg(double trk_dist, int start=0, int dir=1) const;
   int    is_after(int idx, double dst, const SB3PathComponent &prev,
                   int prev_idx, double prev_dst) const;
   int    is_after(double dst, const SB3PathComponent &prev,
                   double prev_dst) const;
   double dst_between(double dst, const SB3PathComponent &prev,
                      double prev_dst) const;
   int    valid(double trk_wid) const;
   void   join(SB3PathComponent &prev_comp, double trk_len);
   
protected:
   // MEMBERS
   int seg_length;

   // CONSTRUCTORS
   SB3PathComponent();

   virtual int  path_ok(int seg_idx, double S, double F, 
                        double min_w, double max_w) const = 0;
   virtual void move_to_outside(SB3Line &ret_norm) const = 0;

private:
   // METHODS
   double dst_between(int idx, double dst, int prev_idx, 
                      double prev_dst) const;
};
//-----END PathComponent.h-----//

//-----START Straight.h-----//
class SB3Straight : public SB3PathComponent 
{
public:
   // CONSTRUCTORS
   SB3Straight();
   
   // METHODS
   double get_spd(double trk_dist, double *max=0) const;
   void   get_tan(SB3Vector &ret_pt, SB3Line &ret_norm) const;
   double get_path_dist(const SB3Vector &pt) const;
   double get_dist2path(const SB3Vector &pt) const;
   void   getPoint2(SB3Vector &ret_pt, SB3Vector &ret_tan, 
                    double path_dist) const;
   double brk_dist(double dist, double v) const;

   void   copy(const SB3Straight &sc);
   void   adjust(const SB3Vector &Spt, const SB3Vector &Fpt);
   int    intersect(SB3Vector &ret_pt, const SB3Straight &s) const;

protected:
   int  path_ok(int seg_idx, double S, double F, 
                double min_w, double max_w) const;
   void move_to_outside(SB3Line &ret_norm) const {};

private:
   // MEMBERS
   SB3Line l;
};
//-----END Straight.h-----//

//-----START Turn.h-----//
class SB3Trans;
class SB3Turn : public SB3PathComponent 
{
public:
   static const int DATALEN,      // Lines per Turn in data file
                    BAD_ENTRY,    // Entry to turn needs adjustment
                    BAD_EXIT,     // Exit from turn needs adjustment
                    BAD_TURN,     // Turn is not completely on course
                    IGNORE_TURN;  // Turn is not needed in path
   
   // CONSTRUCTORS
   SB3Turn();
   SB3Turn(SB3TrkSeg **ts, int idx, int trk_segs,
          double min_w, double max_w, double trk_wid);
   
   // METHODS
   double get_spd(double trk_dist, double *max=0) const;
   void   get_tan(SB3Vector &ret_pt, SB3Line &ret_norm) const;
   double get_path_dist(const SB3Vector &pt) const;
   double get_dist2path(const SB3Vector &pt) const;
   void   getPoint2(SB3Vector &ret_pt, SB3Vector &ret_tan, 
                    double path_dist) const;
   double brk_dist(double dist, double v) const;

   void   init(double min_w, double max_w, double trk_wid);
   void   copy(const SB3Turn &tc, int copy_segs);
   void   copy(const SB3Turn &tc, int copy_segs, int trim_tail,
               int extra_segs);
   void   adjust(SB3Turn &pt, SB3Turn &nt, int adj_val,
                 double min_w, double max_w, double trk_wid);
   int    makeStraight(SB3Straight &ret, SB3Turn &pt, double trk_wid);
   void   set_spd_len(int data_loaded, int set_spd=1);
   int    makePass(SB3Straight *p_s, SB3Turn *p_t, SB3Turn &nt,
                   SB3Turn &last, double min_w, double max_w, 
                   double trk_wid);
   void   save(FILE *out, SB3TrkSeg **ts, double rotate) const;
   int    load(FILE *in, SB3TrkSeg **ts, int trk_segs, double rotate);
   int    makeTransitions(SB3Turn *pt, SB3Straight *ps, SB3Trans &Tentry,
                          SB3Trans &Texit, SB3Straight *ns, SB3Turn *nt,
                          double trk_wid);

protected:
   int  path_ok(int seg_idx, double S, double F, 
                double min_w, double max_w) const;
   void move_to_outside(SB3Line &ret_norm) const;

private:
   // MEMBERS
   SB3Circle       curv;
   double         apex_dist;  // Apex position
   SB3TrkSeg_Turn *ts_turn;
   double         plen, nlen, Sang;
   double         min_spd, max_spd;
   int            allow_pass;

   // METHODS
   int    get_ts_turn_idx() const;
   double get_apex(double dst, double min_w, double max_w) const;
   double get_apex() const;
   void   moveApex(double amount, double min_w, double max_w, 
                   double trk_wid);
   double adjRadius(double amount, double min_w, double max_w);
   double moveTurn(double amount, double min_w, double max_w);
   void   setLimits(double trk_wid);
   void   set(double apex, double min_w, double max_w, double trk_wid);
   double get_weight(const SB3Turn &ot) const;
   double get_err(SB3Turn &pt, int get_pt_err, double trk_wid);
   void   fix_bad_turn(double min_w, double max_w, double trk_wid);
   int    adjustTransitions(SB3Trans &Tentry, double str_ang1, double ang1,
                            double len_ang1, double ps_len,
                            SB3Trans &Texit, double str_ang2, double ang2,
                            double len_ang2, double ns_len,
                            double trk_wid, double apex_ang);
};
//-----END Turn.h-----//

//-----START Trans.h-----//
class SB3Trans : public SB3PathComponent 
{
public:
   static const int RENTRY, LEXIT, REXIT, LENTRY;

   // CONSTRUCTORS
   SB3Trans();

   // METHODS
   double get_spd(double trk_dist, double *max=0) const;
   void   get_tan(SB3Vector &ret_pt, SB3Line &ret_norm) const;
   double get_path_dist(const SB3Vector &pt) const;
   double get_dist2path(const SB3Vector &pt) const;
   void   getPoint2(SB3Vector &ret_pt, SB3Vector &ret_tan, 
                    double path_dist) const;
   double brk_dist(double dist, double v) const;

   void   copy(const SB3Trans &tc);
   void   adjust(const SB3Circle &trn, double str_ang,
                 double len_ang, double len_y,
                 double trk_wid, int type);
   void   set_spd_len(double turn_spd);
   int    makeStraight(SB3Straight &ret, SB3PathComponent &pt,
                       int prev, double trk_wid);
   double rating() const;

   static double get_len_x(double circleR, double len_ang, 
                           double len_y);
   static double get_len_y(double circleR, double len_ang,
                           double len_x);
protected:
   int  path_ok(int seg_idx, double S, double F, 
                double min_w, double max_w) const;
   void move_to_outside(SB3Line &ret_norm) const;

private:
   // MEMBERS
   SB3Ellipse curv;
   int       entry;
   double    init_spd, scale_spd, bias_spd, start_ang, avg_R;
};
//-----END Trans.h-----//

//-----START TrkData.h-----//
class SB3TrkData
{
public:
   static const int NORMAL, OFF, START;

   // MEMBERS
   double pit_entry, pit_spd;

   // CONSTRUCTORS
   SB3TrkData();
   
   // DESTUCTOR
   ~SB3TrkData();
   
   // METHODS
   int    can_pass(double trk_dist, int pc) const;
   double get_trk_dist(int seg_ID, double to_end) const;
   double get_trk_dist(const SB3Vector &pt, double to_lft, 
                       int start_seg_ID=0) const;
   void   get_trk_pos_vel(SB3Line &ret_pv, double to_lft, 
                          double v, double vn, 
                          double trk_dist, int seg_ID) const;
   int    get_PathComponent(double trk_dist, int pass_flag=0,
                            int start=0) const;
   double get_tgt_spd(double v, double drag_accel, 
                      double path_dist, double trk_dist,
                      int pass_flag=0, int pc=-1) const;
   int    get_nex_pos(SB3Line &cV, SB3Vector &nV, int type,
                      double trk_dist, int pass_flag=0, int pc=-1) const;
   void   locate_car(double *dist2path, double *path_dist, 
                     const SB3Vector &pt, int start_pc, int end_pc,
                     double trk_dist=0.0, int pass_flag=0) const;
   void   set_path(Track *trk);
   
private:
   // MEMBERS
   char             *name;
   SB3PathComponent **path;
   SB3TrkSeg        **seg;
   int               num_segs, num_comps;
   double            trk_len, trk_wid;

   // PRIVATE METHODS
   int  get_trk_data(Track *trk);
   int  init_Turns(SB3Turn ***p_crn, int *p_num_turns);
   void adjust_Turns(SB3Turn **crn, SB3Straight **str, 
                     int num_turns, int rnds, int level, int *adj_val);
   void insert_Straights(SB3Turn **crn, SB3Straight **str, 
                         int num_turns, int data_loaded);
   void makePassingPath(SB3Turn **crn, SB3Straight **str, int num_turns);
   void addTransitions(SB3Turn **crn, SB3Straight **str, int num_turns,
                       int data_loaded);
   void saveData(SB3Turn **crn, int num_turns) const;
   int  loadData(SB3Turn ***p_crn, int *p_num_turns);

   int  get_PassComponent(double trk_dist, int pc) const;
   int  get_NextComponent(int pc) const;
};
//-----END TrkData.h-----//

//-----START SBDriver.h-----//
#ifdef SB3_DLL
class SBv3r0
#else
class SBv3r0 : public Driver
#endif
{
public:
   // CONSTRUCTORS
   SBv3r0(int i=-1);
   
   // DESTRUCTOR
   ~SBv3r0();

   // METHODS
   con_vec drive(situation &s);
   
private:
   // PRIVATE MEMBERS
   int      init;
   con_vec  cp;
   int      pc, pass_flag, car2pass;
   double   start_fuel, fuel_per_lap;
   SB3Line   cV;
   SB3Vector nV;

   // PRIVATE METHODS
   void   get_RARS_ctrl_params(situation &s);
   void   update_con_vec(situation &s, int damp);
   void   locate_car(double *dist2path, double *path_dist,
                     const SB3Vector &pt, double to_lft, int seg_ID);
   double anti_collision(situation &s, double tgt_v);
   int    pass_clear(situation &s, double trk_dist);
};
//-----END SBDriver.h-----//


//-----START Vector.cpp-----//
// CONSTRUCTORS
SB3Vector::SB3Vector() 
{
   x = y = 0.0;
}

SB3Vector::SB3Vector(double vx, double vy) 
{
   x = vx;
   y = vy;
}

SB3Vector::SB3Vector(double angle)
{
   x = cos(angle);
   y = sin(angle);
}

SB3Vector::SB3Vector(const SB3Vector &v)
{
   x = v.x;
   y = v.y;
}

// METHODS
void SB3Vector::add(const SB3Vector &v) 
{
   x += v.x;
   y += v.y;
}
   
void SB3Vector::sub(const SB3Vector &v) 
{
   x -= v.x;
   y -= v.y;
}

double SB3Vector::mag() const
{
   return sqrt((x*x) + (y*y));
}

double SB3Vector::mag2() const
{
   return ((x*x) + (y*y));
}

double SB3Vector::dist2Point(const SB3Vector &chk_pt) const
{
   double xd = x - chk_pt.x;
   double yd = y - chk_pt.y;

   return sqrt((xd*xd) + (yd*yd));
}

double SB3Vector::ang() const
{
   double retval;

   if (x != 0.0)
      retval = atan(y / x);
   else {
      if (y >= 0.0)
         retval =  MY_PI12;
      else
         retval = -MY_PI12;
   }

   if (x < 0.0)                       // Quad 2, Quad 3
      retval += MY_PI;
   else if ( (x >= 0.0)&&(y < 0.0) )  // Quad 4
      retval += MY_PI21;

   return retval;
}

double SB3Vector::ang(const SB3Vector &v) const
{
   double retval = v.ang() - ang();

   if (retval > MY_PI)
      retval -= MY_PI21;
   else if (retval < -MY_PI)
      retval += MY_PI21;
   return retval;
}
   
void SB3Vector::scale(double factor) 
{
   x *= factor;
   y *= factor;
}
   
void SB3Vector::normalize()
{
   double m;

   m = mag();
   if (m != 0.0)
      scale(1.0 / m);
}

double SB3Vector::dot(const SB3Vector &v) const
{
   return (x*v.x) + (y*v.y);
}

void SB3Vector::rotate(double angle)
{
   double m, a;

   m = mag();
   a = ang();
   x = m * cos(a + angle);
   y = m * sin(a + angle);
}

int SB3Vector::intersect(SB3Vector &ret_pt, const SB3Line &l) const
{
   double   da;
   SB3Vector tv(*this);

   tv.sub(l.pt);
   da = fabs(tv.ang(l.dir));
   if (da > MY_PI12)
      da = MY_PI - da;
   
   if (da <= 0.017453292) {
      ret_pt.x = x;
      ret_pt.y = y;
      return 1;
   }
   return 0;
}
//-----END Vector.cpp-----//

//-----START Line.cpp-----//
// CONSTRUCTORS
SB3Line::SB3Line()
   : pt(), dir()
{
}

SB3Line::SB3Line(double px, double py, double dx, double dy)
   : pt(px, py), dir(dx, dy)
{
   dir.normalize();
}

SB3Line::SB3Line(double px, double py, double d_ang)
   : pt(px, py), dir(d_ang)
{
}

SB3Line::SB3Line(const SB3Line &l)
   : pt(l.pt), dir(l.dir)
{
}


// METHODS
int SB3Line::intersect(SB3Vector &ret_pt, const SB3Line &l) const
{
   double D = (l.dir.x*dir.y) - (l.dir.y*dir.x);

   if (D == 0.0)
      return 0;
   
   D = ( (dir.x*(l.pt.y - pt.y)) - (dir.y*(l.pt.x - pt.x)) ) / D;
   l.getPoint(ret_pt, D);
   return ((D > 0.0) ? 1 : -1);
}

void SB3Line::getPoint(SB3Vector &ret_pt, double dist) const
{
   ret_pt.x = pt.x + (dist*dir.x);
   ret_pt.y = pt.y + (dist*dir.y);
}

void SB3Line::movePoint(double dist)
{
   pt.x += dist * dir.x;
   pt.y += dist * dir.y;
}

int SB3Line::movePoint(const SB3Line &l)
{
   double N, D;

   N = (dir.x*(l.pt.y - pt.y)) - (dir.y*(l.pt.x - pt.x));
   D = (l.dir.x*dir.y) - (l.dir.y*dir.x);

   if (D == 0.0)
      return 0;

   l.getPoint(pt, (N/D));
   return 1;
}

double SB3Line::dist2Point(const SB3Vector &chk_pt) const
{
   double   N, D;
   SB3Vector norm(dir.ang() + MY_PI12);

   N = (dir.x*(chk_pt.y - pt.y)) - (dir.y*(chk_pt.x - pt.x));
   D = (norm.x*dir.y) - (norm.y*dir.x);

   // +ve return value means chk_pt is to right of this line
   // -ve return value means chk_pt is to left of this line
   return (N/D);
}
//-----END Line.cpp-----//

//-----START Circle.cpp-----//
const int SB3Circle::TAN_1 = 1,
          SB3Circle::TAN_2 = 2,
          SB3Circle::TAN_3 = 3,
          SB3Circle::TAN_4 = 4;

// CONSTRUCTORS
SB3Circle::SB3Circle()
   : c()
{
   R = 0.0;
}

SB3Circle::SB3Circle(double rad)
   : c()
{
   R = rad;
}

SB3Circle::SB3Circle(double cx, double cy, double rad)
   : c(cx, cy)
{
   R = rad;
}

SB3Circle::SB3Circle(const SB3Circle &cc)
   : c(cc.c)
{
   R = cc.R;
}

// METHODS
int SB3Circle::intersect(SB3Vector &ret_pt, const SB3Line &l) const
{
   double   E, F, G, det;
   double   d1, d2;
   SB3Vector dif(l.pt);
   
   dif.sub(c);

   E = l.dir.mag2();
   F = 2.0 * dif.dot(l.dir);
   G = dif.mag2() - (R * R);

   det = (F*F) - (4.0*E*G);
   if (  (det < 0.0)||( (E == 0.0)&&(F == 0.0) )  )
      return 0;

   if (E == 0.0)  // Only 1 solution
      l.getPoint(ret_pt, (-G / F));
   else {
      // 2 solutions, choose 1
      d1 = (-F + sqrt(det)) / (2.0*E);
      d2 = (-F - sqrt(det)) / (2.0*E);

      // find closest point
      if (fabs(d1) < fabs(d2))
         l.getPoint(ret_pt, d1);
      else
         l.getPoint(ret_pt, d2);
   }
   return 1;
}

int SB3Circle::intersect(SB3Vector &ret_pt1, SB3Vector &ret_pt2, 
                        const SB3Line &l) const
{
   double   E, F, G, det;
   double   d1, d2;
   SB3Vector dif(l.pt);

   dif.sub(c);

   E = l.dir.mag2();
   F = 2.0 * dif.dot(l.dir);
   G = dif.mag2() - (R * R);

   det = (F*F) - (4.0*E*G);
   if (  (det < 0.0)||( (E == 0.0)&&(F == 0.0) )  )
      return 0;

   if (E == 0.0) {
      // Only 1 solution
      l.getPoint(ret_pt1, (-G / F));
      ret_pt2.x = ret_pt1.x;
      ret_pt2.y = ret_pt1.y;
   }
   else {
      // 2 solutions
      d1 = (-F + sqrt(det)) / (2.0*E);
      d2 = (-F - sqrt(det)) / (2.0*E);

      // find closest point
      if (fabs(d1) < fabs(d2)) {
         l.getPoint(ret_pt1, d1);
         l.getPoint(ret_pt2, d2);
      }
      else {
         l.getPoint(ret_pt1, d2);
         l.getPoint(ret_pt2, d1);
      }
   }
   return 1;
}

int SB3Circle::intersect(SB3Vector &ret_pt1, SB3Vector &ret_pt2,
                        const SB3Circle &cc) const
{
   double N, D, dst;
   SB3Vector dc(cc.c);

   dc.sub(c);
   dst = dc.mag();
   if (R >= cc.R) {
      if ( ((dst+cc.R) < R)||((dst-cc.R) > R) )
         return 0;
   }
   else {
      if ( ((dst+R) < cc.R)||((dst-R) > cc.R) )
         return 0;
   }

   N = (dst*dst) + (R*R) - (cc.R*cc.R);
   D = 2.0 * dst * R;
   if (D == 0.0)
      return 0;
   N = acos(N/D);
   D = dc.ang();

   getPoint(ret_pt1, (D+N));
   getPoint(ret_pt2, (D-N));
   return 1;
}

double SB3Circle::dist2Point(const SB3Vector &chk_pt) const
{
   // +ve return value means chk_pt is outside this circle
   // -ve return value means chk_pt is inside this circle
   return (c.dist2Point(chk_pt) - R);
}

void SB3Circle::getPoint(SB3Vector &ret_pt, double angle) const
{
   ret_pt.x = c.x + (R * cos(angle));
   ret_pt.y = c.y + (R * sin(angle));
}

double SB3Circle::ang(SB3Vector &pt) const
{
   double retval;

   pt.sub(c);
   retval = pt.ang();
   pt.add(c);
   return retval;
}

int SB3Circle::tangent(SB3Line &ret_l, const SB3Circle &c2, 
                      const int lineID) const
{
   double N, D, ang;

   ret_l.pt.x = c.x; ret_l.pt.y = c.y;
   ret_l.pt.sub(c2.c);
   D = ret_l.pt.mag();

   if ( (lineID == SB3Circle::TAN_1)||(lineID == SB3Circle::TAN_2) ) {
      if ( (R < c2.R)&&((D+R) < c2.R) )
         return 0;
      else if ( (R > c2.R)&&((D+c2.R) < R) )
         return 0;
      D *= R + c2.R;
      if (D == 0.0)
         return 0;
      N  = (c2.R*c2.R) - (R*R);
      if (lineID == SB3Circle::TAN_1)
         ang = ret_l.pt.ang() - acos(N/D);
      else
         ang = ret_l.pt.ang() + acos(N/D);
      getPoint(ret_l.pt, ang);
      c2.getPoint(ret_l.dir, ang);
      ret_l.dir.sub(ret_l.pt);
      return 1;
   }
   else if ( (lineID == SB3Circle::TAN_3)||(lineID == SB3Circle::TAN_4) ) {
      if ( (D < (R+c2.R))||(D == 0.0) )
         return 0;
      else if (R != c2.R) {
         D  *= R - c2.R;
         N   = (c2.R*c2.R) - (R*R);
         ang = ret_l.pt.ang();
      }
      else {
         N   = 2.0 * R;
         ang = ret_l.pt.ang() - MY_PI;
      }

      if (lineID == SB3Circle::TAN_3)
         ang += acos(N/D);
      else
         ang -= acos(N/D);
      getPoint(ret_l.pt, ang);
      c2.getPoint(ret_l.dir, (ang+MY_PI));
      ret_l.dir.sub(ret_l.pt);
      return 1;
   }
   return 0;
}
//-----END Circle.cpp-----//

//-----START Ellipse.cpp-----//
#define dANG  (MY_PI * (5.0/180.0))
const int SB3Ellipse::TAN_1 = 1,
          SB3Ellipse::TAN_2 = 2,
          SB3Ellipse::TAN_3 = 3,
          SB3Ellipse::TAN_4 = 4;

// CONSTRUCTORS
SB3Ellipse::SB3Ellipse()
   : c()
{
   Rx = Ry = r_ang = 0.0;
}
SB3Ellipse::SB3Ellipse(const SB3Circle &cc)
   : c(cc.c)
{
   Rx    = cc.R;
   Ry    = cc.R;
   r_ang = 0.0;
}
SB3Ellipse::SB3Ellipse(const SB3Ellipse &ce)
   : c(ce.c)
{
   Rx    = ce.Rx;
   Rx    = ce.Rx;
   r_ang = ce.r_ang;
}

// METHODS
double SB3Ellipse::R(double angle) const
{
   double x = Rx,
          y = Ry;

   angle -= r_ang;
   x     *= sin(angle);
   y     *= cos(angle);
   return ((Rx*Ry) / sqrt((x*x) + (y*y)));
}

void SB3Ellipse::getPoint(SB3Vector &ret_pt, double angle) const
{
   double Rv = R(angle);
      
   ret_pt.x = (Rv * cos(angle)) + c.x;
   ret_pt.y = (Rv * sin(angle)) + c.y;
}

void SB3Ellipse::getTangent(SB3Vector &ret_tan, double angle) const
{
   double sn1 = sin(angle - r_ang),
          cs1 = cos(angle - r_ang),
          sn2 = sin(r_ang),
          cs2 = cos(r_ang),
          xx  = Rx*Rx,
          yy  = Ry*Ry;

   ret_tan.x = (xx*sn1*cs2) + (yy*cs1*sn2);
   ret_tan.y = (xx*sn1*sn2) - (yy*cs1*cs2);
}

double SB3Ellipse::getCurvature(double angle) const
{
   double num, den,
          fv  = Rx*Rx,
          dfv = Ry*Ry;

   angle -= r_ang;
   angle *= 2.0;               // 2.0*(angle - r_ang)
   den    = fv - dfv;          // (Rx*Rx - Ry*Ry)
   num    = den * cos(angle);  // 0.5*ddf(angle)
   fv    += dfv - num;
   fv    *= 0.5;               // f(angle)
   dfv    = den * sin(angle);  // df(angle)
   dfv   *= 0.25 * dfv;        // 0.25*df(angle)*df(angle)
   num   += fv;
   num   *= fv;
   num   -= dfv;
   den    = fv + (dfv / fv);
   den   *= sqrt(den);
   return (fabs(num) / (Rx*Ry*den));
}

double SB3Ellipse::getArcLength(double start_ang, double len_ang) const
{
   double sn,
          d   = (len_ang > 0.0) ? dANG : -dANG,
          len = 0.0,
          MR  = (Rx > Ry) ? Rx : Ry,
          A   = sqrt(fabs((Ry*Ry) - (Rx*Rx)) / (MR*MR));

   A         *= A;
   start_ang -= r_ang;
   if (MR != Rx)
      start_ang -= MY_PI12;

   for (; fabs(len_ang) > dANG; len_ang -= d) {
      sn   = sin(start_ang + len_ang - (0.5*d));
      len += sqrt(1.0 - (A*sn*sn)) * dANG;
   }
   sn   = sin(start_ang + (0.5*len_ang));
   len += sqrt(1.0 - (A*sn*sn)) * fabs(len_ang);
   return (MR * len);
}

double SB3Ellipse::ang(double ArcLength, double start_ang, int dir) const
{
   double sn,
          d  = (dir > 0) ? dANG : -dANG,
          MR = (Rx > Ry) ? Rx : Ry,
          A  = sqrt(fabs((Ry*Ry) - (Rx*Rx)) / (MR*MR));

   if (ArcLength <= 0.0)
      return start_ang;

   A         *= A;
   start_ang -= r_ang + (0.5*d);
   if (MR != Rx)
      start_ang -= MY_PI12;

   ArcLength /= MR;
   do {
      start_ang += d;
      sn         = sin(start_ang);
      sn         = sqrt(1.0 - (A*sn*sn)) * dANG;
      ArcLength -= sn;
   } while (ArcLength > 0.0);
   start_ang += (d * ArcLength) / sn;
   if (MR != Rx)
      start_ang += MY_PI12;
   return (start_ang + r_ang + (0.5*d));
}

double SB3Ellipse::ang(SB3Vector &pt) const
{
   double retval;

   pt.sub(c);
   retval = pt.ang();
   pt.add(c);
   return retval;
}

double SB3Ellipse::dist2Point(const SB3Vector &chk_pt) const
{
   SB3Vector cpy(chk_pt);

   cpy.sub(c);
   // +ve return value means chk_pt is outside this ellipse
   // -ve return value means chk_pt is inside this ellipse
   return (cpy.mag() - R(cpy.ang()));
}

int SB3Ellipse::intersect(SB3Vector &ret_pt, const SB3Line &l) const
{
   double d1, d2;
   double A = (Ry*Ry) + (Rx*Rx),
          B = (Ry*Ry) - (Rx*Rx),
          C = A,
          D = B * cos(2.0*r_ang),
          E = 0.0,
          F = Rx*Ry;

   A += D; A *= 0.5;
   B *= sin(2.0*r_ang);
   C -= D; C *= 0.5;
   D  = (-2.0*A*c.x) - (B*c.y);
   E  = (-2.0*C*c.y) - (B*c.x);
   F *= -F;
   F += (A*c.x*c.x) + ((B*c.x + C*c.y)*c.y);

   if (fabs(l.dir.x) < fabs(l.dir.y)) {
      d1 = A;
      A  = C;
      C  = d1;
      d1 = D;
      D  = E;
      E  = d1;
      d1 = l.dir.x / l.dir.y;
      d2 = (-d1*l.pt.y) + l.pt.x;
      ret_pt.x = l.pt.y;
      ret_pt.y = l.dir.y;
   }
   else {
      d1 = l.dir.y / l.dir.x;
      d2 = (-d1*l.pt.x) + l.pt.y;
      ret_pt.x = l.pt.x;
      ret_pt.y = l.dir.x;
   }
   A += (B + C*d1)*d1;
   C *= d2;
   B *= d2;
   B += ((2.0*C + E)*d1) + D;
   C  = ((C + E)*d2) + F;

   if (A == 0.0) {
      if (B == 0.0)
         return 0;
      d1 = -C/B;
      d2 = d1;
   }
   else {
      D = (B*B) - (4.0*A*C);
      if ( (D < 0.0)||(ret_pt.y == 0.0) )
         return 0;
      D  = sqrt(D);
      A  = 0.5 / A;
      d1 = (-B + D) * A;
      d2 = (-B - D) * A;
   }
   d1 -= ret_pt.x;
   d2 -= ret_pt.x;
   d1 /= ret_pt.y;
   d2 /= ret_pt.y;
      
   // find closest point
   if (fabs(d1) < fabs(d2))
      l.getPoint(ret_pt, d1);
   else
      l.getPoint(ret_pt, d2);
   return 1;
}

int SB3Ellipse::intersect(SB3Vector &ret_pt1, SB3Vector &ret_pt2, 
                         const SB3Line &l) const
{
   double d1, d2;
   double A = (Ry*Ry) + (Rx*Rx),
          B = (Ry*Ry) - (Rx*Rx),
          C = A,
          D = B * cos(2.0*r_ang),
          E = 0.0,
          F = Rx*Ry;

   A += D; A *= 0.5;
   B *= sin(2.0*r_ang);
   C -= D; C *= 0.5;
   D  = (-2.0*A*c.x) - (B*c.y);
   E  = (-2.0*C*c.y) - (B*c.x);
   F *= -F;
   F += (A*c.x*c.x) + ((B*c.x + C*c.y)*c.y);

   if (fabs(l.dir.x) < fabs(l.dir.y)) {
      d1 = A;
      A  = C;
      C  = d1;
      d1 = D;
      D  = E;
      E  = d1;
      d1 = l.dir.x / l.dir.y;
      d2 = (-d1*l.pt.y) + l.pt.x;
      ret_pt1.x = l.pt.y;
      ret_pt1.y = l.dir.y;
   }
   else {
      d1 = l.dir.y / l.dir.x;
      d2 = (-d1*l.pt.x) + l.pt.y;
      ret_pt1.x = l.pt.x;
      ret_pt1.y = l.dir.x;
   }
   A += (B + C*d1)*d1;
   C *= d2;
   B *= d2;
   B += ((2.0*C + E)*d1) + D;
   C  = ((C + E)*d2) + F;

   if (A == 0.0) {
      if (B == 0.0)
         return 0;
      d1 = -C/B;
      d2 = d1;
   }
   else {
      D = (B*B) - (4.0*A*C);
      if ( (D < 0.0)||(ret_pt1.y == 0.0) )
         return 0;
      D  = sqrt(D);
      A  = 0.5 / A;
      d1 = (-B + D) * A;
      d2 = (-B - D) * A;
   }
   d1 -= ret_pt1.x;
   d2 -= ret_pt1.x;
   d1 /= ret_pt1.y;
   d2 /= ret_pt1.y;

   // find closest point
   if (fabs(d1) < fabs(d2)) {
      l.getPoint(ret_pt1, d1);
      l.getPoint(ret_pt2, d2);
   }
   else {
      l.getPoint(ret_pt1, d2);
      l.getPoint(ret_pt2, d1);
   }
   return 1;
}

int SB3Ellipse::intersect(SB3Vector &ret_pt1, SB3Vector &ret_pt2,
                         SB3Vector &ret_pt3, SB3Vector &ret_pt4, 
                         const SB3Circle &cc) const
{
   double ang;
      
   if (dist2Point(cc.c) > cc.R)
      return 0;

   ang = find_intersect(ret_pt1, cc, 0.0, MY_PI21, 0.0872);
   if (ang > MY_PI21)
      return 0;
         
   ang = find_intersect(ret_pt2, cc, ang, MY_PI21, 0.0872);
   if (ang > MY_PI21)
      return 1;

   ang = find_intersect(ret_pt3, cc, ang, MY_PI21, 0.0872);
   if (ang > MY_PI21)
      return 2;

   ang = find_intersect(ret_pt4, cc, ang, MY_PI21, 0.0872);
   if (ang > MY_PI21)
      return 3;
   return 4;
}

int SB3Ellipse::tangent(SB3Line &ret_l, const SB3Circle &c2, int lineID) const
{
   SB3Ellipse e2(c2);
   return tangent(ret_l, e2, lineID);
}

int SB3Ellipse::tangent(SB3Line &ret_l, const SB3Ellipse &e2, int lineID) const
{
   int    i;
   double start, end, delta, d,
          d1 = Rx*Rx,
          d2 = Ry*Ry,
          A  = sin(r_ang),
          B  = cos(r_ang),
          C  = B,
          D  = A;
      
   A *= d1;
   C *= d1;
   B *= d2;
   D *= d2;

   d1 = r_ang;
   if (C == 0.0)
      d1 += MY_PI12;
   else
      d1 += atan(-D/C);
   if (d1 <= -MY_PI)
      d1 += MY_PI21;
   else if (d1 < 0.0)
      d1 += MY_PI;
   else if (d1 > MY_PI)
      d1 -= MY_PI;

   if (fabs(e2.r_ang) == MY_PI12)
      d2 = d1;
   else {
      d  = tan(e2.r_ang);
      d2 = atan((-B - (D*d)) / ((C*d) - A)) + r_ang;
      if (d2 <= -MY_PI)
         d2 += MY_PI21;
      else if (d2 < 0.0)
         d2 += MY_PI;
      else if (d2 > MY_PI)
         d2 -= MY_PI;

      if (d2 < d1) {
         d  = d1;
         d1 = d2;
         d2 = d;
      }
   }

   for (i = 0; i < 4; i++) {
      // 0 -> d1
      start = 0.0;
      end   = d1 - 0.008727;
      while (start < end) {
         delta = 0.2 * (end - start);
         if (delta > 0.087266)
            delta = 0.087266;
         d = find_discon(A, B, C, D, start, end, delta, e2, i);
         if (d > end) {
            lineID = find_tan(A, B, C, D,
                              start, end, (0.5*delta),
                              ret_l, e2, lineID, i);
            if (lineID == 0)
               return true;
            start = end + 0.017453;
            end   = d1 - 0.008727;
         }
         else
            end = d;
      }
      // d1 -> d2
      start = d1 + 0.008727;
      end   = d2 - 0.008727;
      while (start < end) {
         delta = 0.2 * (end - start);
         if (delta > 0.087266)
            delta = 0.087266;
         d = find_discon(A, B, C, D, start, end, delta, e2, i);
         if (d > end) {
            lineID = find_tan(A, B, C, D,
                              start, end, (0.5*delta),
                              ret_l, e2, lineID, i);
            if (lineID == 0)
               return true;
            start = end + 0.017453;
            end   = d2 - 0.008727;
         }
         else
            end = d;
      }
      // d2 -> 180
      start = d2 + 0.008727;
      end   = MY_PI;
      while (start < end) {
         delta = 0.2 * (end - start);
         if (delta > 0.087266)
            delta = 0.087266;
         d = find_discon(A, B, C, D, start, end, delta, e2, i);
         if (d > end) {
            lineID = find_tan(A, B, C, D,
                              start, end, (0.5*delta),
                              ret_l, e2, lineID, i);
            if (lineID == 0)
               return 1;
            start = end + 0.017453;
            end   = MY_PI;
         }
         else
            end = d;
      }
   }
   return 0;
}

// PRIVATE METHODS
double SB3Ellipse::find_intersect(SB3Vector &ret_pt, const SB3Circle &cc,
                                 double start, double end, double delta) const
{
   double dst,
          scale = 3.0 / (Rx + Ry + cc.R);
   int    sgn   = 0;
      
   while (start <= end) {
      cc.getPoint(ret_pt, start);
      dst = dist2Point(ret_pt) * scale;
      if (fabs(dst) <= 0.0001) {
         start += delta;
         if (start > end)
            start = end;
         break;
      }
      else if (sgn == 0) {
         if (dst > 0.0)
            sgn =  1;
         else
            sgn = -1;
      }
      else if ( ( (sgn > 0)&&(dst < 0.0) )||
                ( (sgn < 0)&&(dst > 0.0) ) ) {
         if (delta < 0.0175)
            cc.getPoint(ret_pt, (start - (0.5*delta)));
         else
            start = find_intersect(ret_pt, cc, (start-delta),
                                   start, (0.2*delta));
         break;
      }
         
      if (start < end) {
         start += delta;
         if (start > end)
            start = end;
      }
      else
         start += delta;
   }
   return start;
}

double SB3Ellipse::find_discon(double A, double B, double C, double D,
                              double start, double end, double delta,
                              const SB3Ellipse &e2, int i) const
{
   double ang, sn, cs, tn;
   double x1, x2;
   int    sgn = 0;
      
   while (start <= end) {
      sn  = sin(start - r_ang);
      cs  = cos(start - r_ang);
      tn  = ((A*sn) - (B*cs)) / ((C*sn) + (D*cs));
      sn  = sin(e2.r_ang);
      cs  = cos(e2.r_ang);
      ang = (e2.Ry*e2.Ry*(cs + (sn*tn))) / (e2.Rx*e2.Rx*((cs*tn) - sn));
      ang = atan(-ang) + e2.r_ang;
      sn  = start;
      
      switch (i) {
      case 1:
         ang += MY_PI;
         break;
      case 2:
         sn  += MY_PI;
         break;
      case 3:
         sn  += MY_PI;
         ang += MY_PI;
      }
      x1 = (R(sn)*cos(sn)) + c.x;
      x2 = (e2.R(ang)*cos(ang)) + e2.c.x;

      if (x1 == x2) {
         start -= 0.008727;
         break;
      }
      else if (sgn == 0) {
         if (x1 > x2)
            sgn =  1;
         else
            sgn = -1;
      }
      else if ( ( (sgn > 0)&&(x1 < x2) )||
                ( (sgn < 0)&&(x1 > x2) ) ) {
         if (delta < 0.0175)
            start -= delta;
         else
            start  = find_discon(A, B, C, D,
                                 (start-delta), start, (0.2*delta),
                                 e2, i);
         break;
      }

      if (start < end) {
         start += delta;
         if (start > end)
            start = end;
      }
      else
         start += delta;
   }
   return start;
}

int SB3Ellipse::find_tan(double A, double B, double C, double D,
                        double start, double end, double delta, SB3Line &ret_l,
                        const SB3Ellipse &e2, int lineID, int i) const
{
   double ang, sn, cs, tn;
   int    sgn = 0;
      
   while (start <= end) {
      sn  = sin(start - r_ang);
      cs  = cos(start - r_ang);
      tn  = ((A*sn) - (B*cs)) / ((C*sn) + (D*cs));
      sn  = sin(e2.r_ang);
      cs  = cos(e2.r_ang);
      ang = (e2.Ry*e2.Ry*(cs + (sn*tn))) / (e2.Rx*e2.Rx*((cs*tn) - sn));
      ang = atan(-ang) + e2.r_ang;
      sn  = start;
         
      switch (i) {
      case 1:
         ang += MY_PI;
         break;
      case 2:
         sn  += MY_PI;
         break;
      case 3:
         sn  += MY_PI;
         ang += MY_PI;
      }
      getPoint(ret_l.pt, sn);
      e2.getPoint(ret_l.dir, ang);
      ret_l.dir.sub(ret_l.pt);
      tn -= ret_l.dir.y / ret_l.dir.x;
         
      if (fabs(tn) < 0.008727) {
         if (lineID == 1)
            return 0;
         lineID--;
         sgn = 0;
      }
      else if (sgn == 0) {
         if (tn > 0.0)
            sgn =  1;
         else
            sgn = -1;
      }
      else if ( ( (sgn > 0)&&(tn < 0.0) )||
                ( (sgn < 0)&&(tn > 0.0) ) ) {
         if (lineID == 1)
            return 0;
         lineID--;
         sgn = -sgn;
      }

      if (start < end) {
         start += delta;
         if (start > end)
            start = end;
      }
      else
         start += delta;
   }
   return lineID;
}
#undef dANG
//-----END Ellipse.cpp-----//

//-----START TrkSeg.cpp-----//
// CONSTRUCTORS
SB3TrkSeg::SB3TrkSeg(int ID)
{
   Sdst  = Fdst = 0.0;
   len   = rad  = 0.0;
   left_side = 0;
}

// METHODS
int SB3TrkSeg::same_dir(const SB3TrkSeg &seg) const
{
   if (rad == 0.0)
      return (seg.rad == 0.0);
   else if (rad > 0.0)
      return (seg.rad > 0.0);
   return (seg.rad < 0.0);
}

double SB3TrkSeg::to_left(const SB3Vector &chk_pt) const
{
   if (rad >= 0.0)
      return left_side->dist2Point(chk_pt);
   return -left_side->dist2Point(chk_pt);
}
//-----END TrkSeg.cpp-----//

//-----START TrkSeg_Straight.cpp-----//
// CONSTRUCTORS
SB3TrkSeg_Straight::SB3TrkSeg_Straight(int ID, segment *seg, double trk_dist)
   : SB3TrkSeg(ID),
     ls(seg->beg_x, seg->beg_y, seg->beg_ang),
     norm(seg->beg_ang - MY_PI12)
{
   Sdst = trk_dist;
   Fdst = trk_dist + seg->length;
   len  = seg->length;
   rad  = seg->radius;

   left_side = &ls;
}

// METHODS
void SB3TrkSeg_Straight::get_norm(SB3Line &ret_n, double trk_dist) const
{
   if ( (trk_dist < (Sdst-2.5))||(trk_dist > (Fdst+2.5)) ) {
      ret_n.pt.scale(0.0);
      ret_n.dir.scale(0.0);
   }
   else {
      ls.getPoint(ret_n.pt, (trk_dist - Sdst));
      ret_n.dir.x = norm.x;
      ret_n.dir.y = norm.y;
   }
}

double SB3TrkSeg_Straight::get_trk_dist(const SB3Vector &chk_pt) const
{
   double N, D;

   N = (norm.x*(chk_pt.y - ls.pt.y)) - (norm.y*(chk_pt.x - ls.pt.x));
   D = (ls.dir.y*norm.x) - (ls.dir.x*norm.y);

   return ((N/D) + Sdst);
}

int SB3TrkSeg_Straight::path_ok(const SB3Line &path, double S, double F,
                               double min_w, double max_w) const
{
   double   trk_dist;
   SB3Vector pt;
   SB3Line   edge(ls);

   // check finish edge
   edge.movePoint(F - Sdst);
   edge.dir.x = norm.x;
   edge.dir.y = norm.y;
   if (edge.intersect(pt, path)) {
      trk_dist = pt.dist2Point(edge.pt);
      if ( (min_w > trk_dist)||(trk_dist > max_w) )
         return 0;
   }
   else
      return 0;

   // check start edge
   edge.pt.x = ls.pt.x + ((S-Sdst) * ls.dir.x);
   edge.pt.y = ls.pt.y + ((S-Sdst) * ls.dir.y);
   if (edge.intersect(pt, path)) {
      trk_dist = pt.dist2Point(edge.pt);
      if ( (min_w > trk_dist)||(trk_dist > max_w) )
         return 0;
   }
   else
      return 0;

   // check left edge
   edge.pt.x  = ls.pt.x + (min_w * norm.x);
   edge.pt.y  = ls.pt.y + (min_w * norm.y);
   edge.dir.x = ls.dir.x;
   edge.dir.y = ls.dir.y;
   if (edge.intersect(pt, path)) {
      trk_dist = get_trk_dist(pt);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }

   // check right edge
   edge.pt.x += (max_w-min_w) * norm.x;
   edge.pt.y += (max_w-min_w) * norm.y;
   if (edge.intersect(pt, path)) {
      trk_dist = get_trk_dist(pt);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }
   return 1;
}

int SB3TrkSeg_Straight::path_ok(const SB3Circle &path, double S, double F,
                               double min_w, double max_w) const
{
   double   trk_dist;
   SB3Vector pt1, pt2;
   SB3Line   edge(ls);

   // check left edge
   edge.pt.x += min_w * norm.x;
   edge.pt.y += min_w * norm.y;
   if (path.intersect(pt1, pt2, edge)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }

   // check right edge
   edge.pt.x += (max_w-min_w) * norm.x;
   edge.pt.y += (max_w-min_w) * norm.y;
   if (path.intersect(pt1, pt2, edge)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }
   return 1;
}

int SB3TrkSeg_Straight::path_ok(const SB3Ellipse &path, double S, double F,
                               double min_w, double max_w) const
{
   double   trk_dist;
   SB3Vector pt1, pt2;
   SB3Line   edge(ls);

   // check left edge
   edge.pt.x += min_w * norm.x;
   edge.pt.y += min_w * norm.y;
   if (path.intersect(pt1, pt2, edge)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }

   // check right edge
   edge.pt.x += (max_w-min_w) * norm.x;
   edge.pt.y += (max_w-min_w) * norm.y;
   if (path.intersect(pt1, pt2, edge)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }
   return 1;
}
//-----END TrkSeg_Straight.cpp-----//

//-----START TrkSeg_Turn.cpp-----//
// CONSTRUCTORS
SB3TrkSeg_Turn::SB3TrkSeg_Turn(int ID, segment *seg, double wid,
                             double trk_dist)
   : SB3TrkSeg(ID),
     ls(seg->cen_x, seg->cen_y, fabs(seg->radius))
{
   beg_ang = seg->beg_ang;
   end_ang = seg->end_ang;
   if (seg->radius > 0.0) {    // left turn
      len      = seg->length * (( 0.5*wid) + seg->radius);
      rad      = seg->radius;
      beg_ang -= MY_PI12;
      end_ang -= MY_PI12;
      left     = 1;
   }
   else {                     // right turn
      len      = seg->length * ((-0.5*wid) - seg->radius);
      rad      = seg->radius + wid;
      beg_ang += MY_PI12;
      end_ang += MY_PI12;
      left     = 0;
   }

   if (beg_ang < 0.0)
      beg_ang += MY_PI21;
   else if (beg_ang > MY_PI21)
      beg_ang -= MY_PI21;
   if (end_ang < 0.0)
      end_ang += MY_PI21;
   else if (end_ang > MY_PI21)
      end_ang -= MY_PI21;

   if ( (rad < 0.0)&&(end_ang > beg_ang) )
      beg_ang += MY_PI21;
   else if ( (rad > 0.0)&&(end_ang < beg_ang) )
      end_ang += MY_PI21;

   Sdst = trk_dist;
   Fdst = trk_dist + len;
   len  = seg->length;

   left_side = &ls;
}

// METHODS
void SB3TrkSeg_Turn::get_norm(SB3Line &ret_n, double trk_dist) const
{
   double ang;

   if ( (trk_dist < (Sdst-2.5))||(trk_dist > (Fdst+2.5)) ) {
      ret_n.pt.scale(0.0);
      ret_n.dir.scale(0.0);
      return;
   }

   ang  = (trk_dist - Sdst) / (Fdst - Sdst);
   ang *= end_ang - beg_ang;
   ls.getPoint(ret_n.pt, (ang + beg_ang));
   if (rad > 0.0) {
      ret_n.dir.x = ret_n.pt.x - ls.c.x;
      ret_n.dir.y = ret_n.pt.y - ls.c.y;
   }
   else {
      ret_n.dir.x = ls.c.x - ret_n.pt.x;
      ret_n.dir.y = ls.c.y - ret_n.pt.y;
   }
   ret_n.dir.normalize();
}

double SB3TrkSeg_Turn::get_trk_dist(const SB3Vector &chk_pt) const
{
   double   dist;
   SB3Vector loc(chk_pt);

   loc.sub(ls.c);
   dist  = loc.ang() - beg_ang;
   if (dist > (MY_PI21-0.0000001))
      dist -= MY_PI21-0.0000001;
   else if (dist < (-MY_PI21+0.0000001))
      dist += MY_PI21-0.0000001;
   if ( (beg_ang < end_ang)&&(dist < 0.0) )
      dist += MY_PI21;
   else if ( (beg_ang > end_ang)&&(dist > 0.0) )
      dist -= MY_PI21;
   dist *= (Fdst - Sdst) / (end_ang - beg_ang);
   return (dist + Sdst);
}

int SB3TrkSeg_Turn::path_ok(const SB3Line &path, double S, double F,
                           double min_w, double max_w) const
{
   double   trk_dist, ang;
   SB3Vector pt1, pt2;
   SB3Circle edge(ls);

   edge.R = fabs(rad);

   // check start and finish edges
   ang  = (S - Sdst) / (Fdst - Sdst);
   ang *= end_ang - beg_ang;
   edge.getPoint(pt1, (ang + beg_ang));
   ang  = (F - Sdst) / (Fdst - Sdst);
   ang *= end_ang - beg_ang;
   edge.getPoint(pt2, (ang + beg_ang));
   if (rad > 0.0) {  // left turn
      trk_dist = -path.dist2Point(pt1);
      if ( (min_w > trk_dist)||(trk_dist > max_w) )
         return 0;
      trk_dist = -path.dist2Point(pt2);
      if ( (min_w > trk_dist)||(trk_dist > max_w) )
         return 0;
   }
   else {
      trk_dist = path.dist2Point(pt1);
      if ( (min_w > trk_dist)||(trk_dist > max_w) )
         return 0;
      trk_dist = path.dist2Point(pt2);
      if ( (min_w > trk_dist)||(trk_dist > max_w) )
         return 0;
   }

   // check inner edge
   edge.R += min_w;
   if (edge.intersect(pt1, pt2, path)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }

   // check outer edge
   edge.R += max_w - min_w;
   if (edge.intersect(pt1, pt2, path)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      return 1;
   }
   return 0;
}

int SB3TrkSeg_Turn::path_ok(const SB3Circle &path, double S, double F,
                           double min_w, double max_w) const
{
   double   trk_dist;
   SB3Vector pt1, pt2;
   SB3Circle edge(ls);

   edge.R = fabs(rad);

   // check inner edge
   edge.R += min_w;
   if (edge.intersect(pt1, pt2, path)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }

   // check outer edge
   edge.R += max_w - min_w;
   if (edge.intersect(pt1, pt2, path)) {
      trk_dist = get_trk_dist(pt1);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
      trk_dist = get_trk_dist(pt2);
      if ( (trk_dist >= S)&&(trk_dist <= F) )
         return 0;
   }
   return 1;
}

int SB3TrkSeg_Turn::path_ok(const SB3Ellipse &path, double S, double F,
                           double min_w, double max_w) const
{
	int      n;
	double   trk_dist;
	SB3Vector pt1, pt2, pt3, pt4;
	SB3Circle edge(ls);

	edge.R = fabs(rad);

	// check inner edge
	edge.R += min_w;
	n       = path.intersect(pt1, pt2, pt3, pt4, edge);
	if (n > 0) {
		trk_dist = get_trk_dist(pt1);
		if ( (trk_dist >= S)&&(trk_dist <= F) )
			return 0;
		else if (n > 1) {
			trk_dist = get_trk_dist(pt2);
			if ( (trk_dist >= S)&&(trk_dist <= F) )
				return 0;
			else if (n > 2) {
				trk_dist = get_trk_dist(pt3);
				if ( (trk_dist >= S)&&(trk_dist <= F) )
					return 0;
				else if (n > 3) {
					trk_dist = get_trk_dist(pt4);
					if ( (trk_dist >= S)&&(trk_dist <= F) )
						return 0;
				}
			}
		}
	}

	// check outer edge
	edge.R += max_w - min_w;
	n       = path.intersect(pt1, pt2, pt3, pt4, edge);
	if (n > 0) {
		trk_dist = get_trk_dist(pt1);
		if ( (trk_dist >= S)&&(trk_dist <= F) )
			return 0;
		else if (n > 1) {
			trk_dist = get_trk_dist(pt2);
			if ( (trk_dist >= S)&&(trk_dist <= F) )
				return 0;
			else if (n > 2) {
				trk_dist = get_trk_dist(pt3);
				if ( (trk_dist >= S)&&(trk_dist <= F) )
					return 0;
				else if (n > 3) {
					trk_dist = get_trk_dist(pt4);
					if ( (trk_dist >= S)&&(trk_dist <= F) )
						return 0;
				}
			}
		}
	}
	return 1;
}

int SB3TrkSeg_Turn::like_straight() const
{
   return ( ((fabs(rad) > 1000.0)&&(len < 0.55))||
            ((fabs(rad) > 1100.0)&&(len < 0.91))||
            ((fabs(rad) > 1400.0)&&(len < 1.05)) );
}
//-----END TrkSeg_Turn.cpp-----//

//-----START PathComponent.cpp-----//
const short SB3PathComponent::STRAIGHT = 0,
            SB3PathComponent::TURN     = 1,
            SB3PathComponent::TRANS    = 2;

// CONSTRUCTORS
SB3PathComponent::SB3PathComponent()
{
   Sdst       = Fdst = len = 0.0;
   num_segs   = 0;
   seg        = (SB3TrkSeg **)malloc(sizeof(SB3TrkSeg *));
   seg[0]     = NULL;
   seg_length = 1;
   comp       = NULL;
   pass       = NULL;
   typeID     = 0;
   PathDir    = 0;

   pre_len = post_len = len_trk = 0.0;
}

// DESTRUCTORS
SB3PathComponent::~SB3PathComponent()
{
   if (seg != NULL)
      free(seg);

   if (pass != NULL) {
      delete pass[0];
      delete pass[1];
      delete [] pass;
   }
}

// METHODS
void SB3PathComponent::copy(const SB3PathComponent &pc)
{
   len      = pc.len;
   Sdst     = pc.Sdst;
   Fdst     = pc.Fdst;
   num_segs = pc.num_segs;
   pass     = pc.pass;
   pre_len  = pc.pre_len;
   post_len = pc.post_len;
   len_trk  = pc.len_trk;
   if ( (seg == NULL)||(num_segs > seg_length) ) {
      seg_length = num_segs + 2;
      seg        = (SB3TrkSeg **)realloc(seg, (seg_length*sizeof(SB3TrkSeg *)));
   }
   if (num_segs > 0)
      memcpy(seg, pc.seg, (num_segs*sizeof(SB3TrkSeg *)));
}

int SB3PathComponent::getPoint(SB3Vector &ret_pt, SB3Line &ret_norm,
                              double trk_dist, int ret_tan) const
{
   int i;

   for (i = 0; i < num_segs; i++) {
      if ( (trk_dist >= seg[i]->Sdst)&&
           (trk_dist <= seg[i]->Fdst) )
         break;
   }
   if (i >= num_segs) {
      for (i = 0; i < num_segs; i++) {
         if ( (trk_dist >= seg[i]->Sdst)&&
              (trk_dist < (seg[i]->Fdst+2.5)) )
            break;
      }
      if (i >= num_segs)
         return 0;
   }

   seg[i]->get_norm(ret_norm, trk_dist);
   move_to_outside(ret_norm);
   i = comp->intersect(ret_pt, ret_norm);
   if ( (i)&&(ret_tan) )
      get_tan(ret_pt, ret_norm);
   return i;
}

double SB3PathComponent::to_side(SB3Vector &ret_pt, SB3Line &ret_norm,
                                double trk_dist, int ret_tan) const
{
   int    i;
   double retval;

   for (i = 0; i < num_segs; i++) {
   if ( (trk_dist >= seg[i]->Sdst)&&
        (trk_dist <= seg[i]->Fdst) )
      break;
   }
   if (i >= num_segs)
      return 0.0;

   seg[i]->get_norm(ret_norm, trk_dist);
   move_to_outside(ret_norm);
   if (!comp->intersect(ret_pt, ret_norm))
      return 0.0;

   if (seg[i]->rad >= 0.0)
      retval =  seg[i]->left_side->dist2Point(ret_pt);
   else
      retval = -seg[i]->left_side->dist2Point(ret_pt);

   if (ret_tan)
      get_tan(ret_pt, ret_norm);
   return ((retval == 0.0) ? 0.001 : retval);
}

void SB3PathComponent::insert_TrkSeg(SB3TrkSeg *ts, int idx)
{
   int i;

   // check for duplicates
   for (i = 0; i < num_segs; i++) {
      if (seg[i] == ts)
         return;
   }

   if (idx < 0)
      idx = 0;
   else if (idx > num_segs)
      idx = num_segs;

   num_segs++;
   if (num_segs > seg_length) {
      seg_length = num_segs + 2;
      seg        = (SB3TrkSeg **)realloc(seg, (seg_length*sizeof(SB3TrkSeg *)));
   }
   for (i = num_segs-1; i > idx; i--) {
      seg[i] = seg[i-1];
   }
   seg[idx] = ts;
}

void SB3PathComponent::insert_TrkSeg(const SB3PathComponent &pc, 
                                    int add_to_tail)
{
   int s;

   if (add_to_tail) {
      for (s = 0; s < pc.num_segs; s++) {
         insert_TrkSeg(pc.seg[s], num_segs);
      }
   }
   else {
      for (s = pc.num_segs-1; s >= 0; s--) {
         insert_TrkSeg(pc.seg[s], 0);
      }
   }
}

void SB3PathComponent::remove_TrkSeg(int idx)
{
   int i;

   if ( (idx < 0)||(idx >= num_segs) )
      return;

   num_segs--;
   for (i = idx; i < num_segs; i++) {
      seg[i] = seg[i+1];
   }
   seg[i] = NULL;
}

void SB3PathComponent::trim()
{
   int s, f;

   if (num_segs > 0) {
      s = find_trk_seg(Sdst, 0, 1);
      f = find_trk_seg(Fdst, (num_segs-1), -1);
      if ( (s >= 0)&&(f >= 0)&&(f < s) ) {
         int t;
            
         t = s;
         s = f;
         f = t;
      }
      if (f >= 0)
         f = num_segs - 1 - f;
      
      // remove unused TrkSeg from component
      for (; s > 0; s--) {
         remove_TrkSeg(0);
      }
      for (; f > 0; f--) {
         remove_TrkSeg(num_segs-1);
      }
   }

   if (seg_length > num_segs) {
      // free unused memory
      seg_length = num_segs;
      seg        = (SB3TrkSeg **)realloc(seg, (seg_length*sizeof(SB3TrkSeg *)));
   }
}

int SB3PathComponent::find_trk_seg(const SB3Vector &pt, double trk_wid,
                                  int start, int dir) const
{
   int    i;
   double d, to_lft;

   if (dir < 0) {
      for (i = start; i >= 0; i--) {
         d      = seg[i]->get_trk_dist(pt);
         to_lft = seg[i]->to_left(pt);
         if ( (d >= seg[i]->Sdst)&&(d <= seg[i]->Fdst)&&
              (to_lft >= 0.0)&&(to_lft <= trk_wid) )
            return i;
      }
   }
   else {
      for (i = start; i < num_segs; i++) {
         d      = seg[i]->get_trk_dist(pt);
         to_lft = seg[i]->to_left(pt);
         if ( (d >= seg[i]->Sdst)&&(d <= seg[i]->Fdst)&&
              (to_lft >= 0.0)&&(to_lft <= trk_wid) )
            return i;
      }
   }
   return -1;
}

int SB3PathComponent::find_trk_seg(double trk_dist, int start, int dir) const
{
   int i;

   if ( (trk_dist > (seg[start]->Sdst-5.0))&&
   	  (trk_dist < (seg[start]->Fdst+5.0)) )
   	return start;
   if (dir < 0) {
      for (i = start-1; i > 0; i--) {
         if ( (trk_dist >= seg[i]->Sdst)&&(trk_dist <= seg[i]->Fdst) )
            return i;
      }
      if ( (trk_dist > (seg[0]->Sdst-5.0))&&
           (trk_dist < (seg[0]->Fdst+5.0)) )
         return 0;
   }
   else {
      for (i = start+1; i < (num_segs-1); i++) {
         if ( (trk_dist >= seg[i]->Sdst)&&(trk_dist <= seg[i]->Fdst) )
            return i;
      }
      if ( (trk_dist > (seg[num_segs-1]->Sdst-5.0))&&
           (trk_dist < (seg[num_segs-1]->Fdst+5.0)) )
         return (num_segs-1);
   }
   return -1;
}

int SB3PathComponent::is_after(int idx, double dst,
                              const SB3PathComponent &prev,
                              int prev_idx, double prev_dst) const
{
   int i;

   // check if both points are in same seg
   if (seg[idx] == prev.seg[prev_idx])
      return (prev_dst <= dst);
   // check if a seg in this component holds the previous point
   for (i = 0; i < num_segs; i++) {
      if (seg[i] == prev.seg[prev_idx])
         return (i < idx);
   }
   // check if a seg in the prev component holds the after point
   for (i = 0; i < prev.num_segs; i++) {
      if (prev.seg[i] == seg[idx])
         return (i > prev_idx);
   }
   /* There are no overlapping segs in the 2 components, therefore
      we must try to return the correct decision here. */
   if (seg[0]->Sdst == prev.seg[prev.num_segs-1]->Fdst)
      return true;   // this component is placed right after prev
   else if (prev.seg[0]->Sdst == seg[num_segs-1]->Fdst)
      return false;  // prev component is placed right after this
   return (prev_dst <= dst);
}

int SB3PathComponent::is_after(double dst, const SB3PathComponent &prev,
                              double prev_dst) const
{
   int idx, prev_idx;

   idx = find_trk_seg(dst, 0, 1);
   if (idx < 0)
      return 0;
   prev_idx = prev.find_trk_seg(prev_dst, 0, 1);
   if (prev_idx < 0)
      return 0;
   return is_after(idx, dst, prev, prev_idx, prev_dst);
}

double SB3PathComponent::dst_between(double dst, const SB3PathComponent &prev,
                                    double prev_dst) const
{
   int idx, prev_idx, i;

   idx = find_trk_seg(dst, 0, 1);
   if (idx < 0)
      return 0.0;
   prev_idx = prev.find_trk_seg(prev_dst, 0, 1);
   if (prev_idx < 0)
      return 0.0;

   // check if both points are in same seg
   if (seg[idx] == prev.seg[prev_idx]) {
      dst -= prev_dst;
      return ((dst == 0.0) ? 0.0001 : dst);
   }
   // check if a seg in this component holds the previous point
   for (i = 0; i < num_segs; i++) {
      if (seg[i] == prev.seg[prev_idx]) {
         return dst_between(idx, dst, i, prev_dst);
      }
   }
   // check if a seg in the prev component holds the after point
   for (i = 0; i < prev.num_segs; i++) {
      if (prev.seg[i] == seg[idx]) {
         return -prev.dst_between(prev_idx, prev_dst, i, dst);
      }
   }
   /* There are no overlapping segs in the 2 components, therefore
      we must try to return the correct value here. */
   if (seg[0]->Sdst == prev.seg[prev.num_segs-1]->Fdst) {
      // this component is placed right after prev
      dst  = dst_between(idx, dst, 0, seg[0]->Sdst);
      i    = prev.num_segs - 1;
      dst += prev.dst_between(i, prev.seg[i]->Fdst, prev_idx, prev_dst);
   }
   else if (prev.seg[0]->Sdst == seg[num_segs-1]->Fdst) {
      // prev component is placed right after this
      i    = num_segs - 1;
      dst  = -dst_between(i, seg[i]->Fdst, idx, dst);
      dst -= prev.dst_between(prev_idx, prev_dst, 0, prev.seg[0]->Sdst);
   }
   return ((dst == 0.0) ? 0.0001 : dst);
}

int SB3PathComponent::valid(double trk_wid) const
{
   int    i;
   double beg, end, s, f,
          in_w  = SB3_IN_WID - SB3_PATH_ERR,
          out_w = trk_wid - SB3_OUT_WID + SB3_PATH_ERR,
          max_w = trk_wid - in_w;
      
   if (is_after(Sdst, *this, Fdst)) {
      beg = Fdst;
      end = Sdst;
   }
   else {
      beg = Sdst;
      end = Fdst;
   }

   s = beg;
   f = end + 1.0;
   for (i = 0; (i < num_segs)&&(f != end); i++) {
      // start location
      if (s == beg) {
         if ( (beg < seg[i]->Sdst)||(beg > seg[i]->Fdst) )
            continue;
         beg += 1.0;  // found first segment, stop looking
      }
      else
         s = seg[i]->Sdst;
      // finish location
      if ( (end >= seg[i]->Sdst)&&(end <= seg[i]->Fdst) )
         f = end;     // found last segment
      else
         f = seg[i]->Fdst;

      // make sure this path is valid between start and finish locations
      if ( (seg[i]->rad == 0.0)||(typeID == STRAIGHT) ) {
         if (!path_ok(i, s, f, in_w, max_w))
            return 0;
      }
      else {
         if (!path_ok(i, s, f, in_w, out_w))
            return 0;
      }
   }
   return 1;
}

void SB3PathComponent::join(SB3PathComponent &prev_comp, double trk_len)
{
   if (prev_comp.Fdst != Sdst) {
      double d;
      int    i1 = prev_comp.find_trk_seg(Sdst, 0, 1),
             i2 = find_trk_seg(prev_comp.Fdst, 0, 1);

      if ( (i1 >= 0)&&(i2 >= 0) ) {
         d  = dst_between(Sdst, *this, prev_comp.Fdst);
         d *= 0.5;
         d += prev_comp.Fdst;
         if (d < 0.0)
            d += trk_len;
         else if (d > trk_len)
            d -= trk_len;
      }
      else if (i1 >= 0)
         d = prev_comp.seg[i1]->Sdst;
      else if (i2 >= 0)
         d = seg[i2]->Fdst;
      else
         d = seg[0]->Sdst;
      prev_comp.Fdst = Sdst = d;
   }
}

double SB3PathComponent::dst_between(int idx, double dst, int prev_idx, 
                                    double prev_dst) const
{
   if (prev_idx < idx) {
      prev_dst = seg[prev_idx++]->Fdst - prev_dst;
      for (; prev_idx < idx; prev_idx++) {
         prev_dst += seg[prev_idx]->Fdst - seg[prev_idx]->Sdst;
      }
      dst -= seg[prev_idx]->Sdst;
   }
   else {
      prev_dst = seg[prev_idx--]->Sdst - prev_dst;
      for (; prev_idx > idx; prev_idx--) {
         prev_dst -= seg[prev_idx]->Fdst - seg[prev_idx]->Sdst;
      }
      dst -= seg[prev_idx]->Fdst;
   }
   dst += prev_dst;
   return ((dst == 0.0) ? 0.0001 : dst);
}
//-----END PathComponent.cpp-----//

//-----START Straight.cpp-----//
// CONSTRUCTORS
SB3Straight::SB3Straight()
   : SB3PathComponent(), l()
{
   typeID  = SB3PathComponent::STRAIGHT;
   PathDir = 0;
   comp    = &l;
}

// METHODS
double SB3Straight::get_spd(double trk_dist, double *max) const
{
   if (max)
      *max = SB3_MAX_SPD;
   return SB3_MAX_SPD;
}

void SB3Straight::get_tan(SB3Vector &ret_pt, SB3Line &ret_norm) const
{
   ret_norm.pt.x  = ret_norm.dir.x; ret_norm.pt.y  = ret_norm.dir.y;
   ret_norm.dir.x = l.dir.x;        ret_norm.dir.y = l.dir.y;
   ret_norm.pt.rotate(MY_PI12);
}

double SB3Straight::get_path_dist(const SB3Vector &pt) const
{
   SB3Vector v(pt);

   v.sub(l.pt);
   return (fabs(v.dot(l.dir)) / len);
}

double SB3Straight::get_dist2path(const SB3Vector &pt) const
{
   return l.dist2Point(pt);
}

void SB3Straight::getPoint2(SB3Vector &ret_pt, SB3Vector &ret_tan, 
                           double path_dist) const
{
   l.getPoint(ret_pt, (path_dist/len));
   ret_tan.x = l.dir.x; ret_tan.y = l.dir.y;
}

double SB3Straight::brk_dist(double dist, double v) const
{
   return dist;
}

void SB3Straight::copy(const SB3Straight &sc)
{
   SB3PathComponent::copy(sc);

   l.pt.x  = sc.l.pt.x;
   l.pt.y  = sc.l.pt.y;
   l.dir.x = sc.l.dir.x;
   l.dir.y = sc.l.dir.y;
}

void SB3Straight::adjust(const SB3Vector &Spt, const SB3Vector &Fpt)
{
   l.pt.x  = Spt.x; l.pt.y  = Spt.y;
   l.dir.x = Fpt.x; l.dir.y = Fpt.y;
   l.dir.sub(Spt);

   len = l.dir.mag();
}

int SB3Straight::intersect(SB3Vector &ret_pt, const SB3Straight &s) const
{
   return l.intersect(ret_pt, s.l);
}

int SB3Straight::path_ok(int seg_idx, double S, double F, 
                        double min_w, double max_w) const
{
   if (len < 1.0)
      return 1;
   return seg[seg_idx]->path_ok(l, S, F, min_w, max_w);
}
//-----END Straight.cpp-----//

//-----START Turn.cpp-----//
#define APEX_BEG  0.0
#define APEX_MID  0.5
#define APEX_END  1.0
#define ADJ_ANG   (2.0*(MY_PI/180.0))
#define ADJ_RAD   0.02
const int SB3Turn::DATALEN     = 2,  // Lines per Turn in data file
          SB3Turn::BAD_ENTRY   = 1,  // Entry to turn needs adjustment
          SB3Turn::BAD_EXIT    = 2,  // Exit from turn needs adjustment
          SB3Turn::BAD_TURN    = 4,  // Turn is not completely on course
          SB3Turn::IGNORE_TURN = 8;  // Turn is not needed in path


// CONSTRUCTORS
SB3Turn::SB3Turn()
   : SB3PathComponent(), curv()
{
   typeID  = SB3PathComponent::TURN;
   PathDir = 1;
   comp    = &curv;
   ts_turn = NULL;
   plen    = 0.0;
   nlen    = 0.0;
   Sang    = 0.0;
   min_spd = 1.0;
   max_spd = 1.01;
   allow_pass = 1;
}

SB3Turn::SB3Turn(SB3TrkSeg **ts, int idx, int trk_segs,
               double min_w, double max_w, double trk_wid)
   : SB3PathComponent(), curv()
{
   int            i, j;
   SB3TrkSeg_Turn *pt, *nt;

   typeID  = SB3PathComponent::TURN;
   PathDir = 1;
   comp    = &curv;
   ts_turn = (SB3TrkSeg_Turn *)ts[idx];
   plen    = 0.0;
   nlen    = 0.0;
   Sang    = 0.0;
   min_spd = 1.0;
   max_spd = 1.01;
   allow_pass = 1;

   insert_TrkSeg(ts[idx], 0);
   for (i = 1, j = 0; i < trk_segs; i++) {
      j = (idx + trk_segs - i) % trk_segs;
      if (ts[j]->rad != 0.0)
         break;
      insert_TrkSeg(ts[j], 0);
      plen += ts[j]->len;
   }
   pt = (SB3TrkSeg_Turn *)ts[j];
   for (i = 1, j = 0; i < trk_segs; i++) {
      j = (idx + i) % trk_segs;
      if (ts[j]->rad != 0.0)
         break;
      insert_TrkSeg(ts[j], num_segs);
      nlen += ts[j]->len;
   }
   nt = (SB3TrkSeg_Turn *)ts[j];

   insert_TrkSeg(pt, 0);
   insert_TrkSeg(nt, num_segs);

   init(min_w, max_w, trk_wid);
}

// METHODS
double SB3Turn::get_spd(double trk_dist, double *max) const
{
   double pct;

   if (max) {
      pct = dst_between(trk_dist, *this, Sdst);
      if (pct > 0.0)
         pct /= len_trk;

      if (pct <= 0.0)
         *max = min_spd;
      else
         *max = (pct*max_spd) + ((1.0-pct)*min_spd);
   }
   return min_spd;
}

void SB3Turn::get_tan(SB3Vector &ret_pt, SB3Line &ret_norm) const
{
   ret_norm.pt.x  = ret_norm.dir.x; ret_norm.pt.y  = ret_norm.dir.y;
   ret_norm.dir.x = ret_pt.x;       ret_norm.dir.y = ret_pt.y;
   ret_norm.dir.sub(curv.c);
   if (fabs(ret_norm.pt.ang(ret_norm.dir)) > MY_PI12)
      ret_norm.dir.rotate(MY_PI + MY_PI12);
   else
      ret_norm.dir.rotate(MY_PI12);
   ret_norm.pt.rotate(MY_PI12);
}

double SB3Turn::get_path_dist(const SB3Vector &pt) const
{
   double   a;
   SB3Vector v(pt);

   v.sub(curv.c);
   a = v.ang();

   if ( (PathDir > 0)&&(a < Sang) )
      a += MY_PI21;
   else if ( (PathDir < 0)&&(a > Sang) )
      a -= MY_PI21;
   a = fabs(a - Sang);
   if (a > (MY_PI21 - 0.05))
      a -= MY_PI21;
   return (curv.R * a);
}

double SB3Turn::get_dist2path(const SB3Vector &pt) const
{
   if (PathDir > 0)
      return curv.dist2Point(pt);
   return -curv.dist2Point(pt);
}

void SB3Turn::getPoint2(SB3Vector &ret_pt, SB3Vector &ret_tan, 
                       double path_dist) const
{
   if (PathDir > 0) {
      curv.getPoint(ret_pt, (Sang + (path_dist/curv.R)));
      ret_tan.x = ret_pt.x - curv.c.x; 
      ret_tan.y = ret_pt.y - curv.c.y;
      ret_tan.rotate(MY_PI12);
   }
   else {
      curv.getPoint(ret_pt, (Sang - (path_dist/curv.R)));
      ret_tan.x = ret_pt.x - curv.c.x; 
      ret_tan.y = ret_pt.y - curv.c.y;
      ret_tan.rotate(-MY_PI12);
   }
}

double SB3Turn::brk_dist(double dist, double v) const
{
   v *= v;
   v /= curv.R * MAX_ACCEL;
   v *= v;
   v  = 1.0 - v;
   if (v < 0.5625)
      dist *= 0.9 * 0.75;
   else
      dist *= 0.9 * sqrt(v);
   return dist;
}

void SB3Turn::init(double min_w, double max_w, double trk_wid)
{
   int            i;
   double         apex, min_len;
   SB3TrkSeg_Turn *pt = NULL,
                 *nt = NULL;

   i = get_ts_turn_idx() - 1;
   while ( (i >= 0)&&(seg[i]->rad == 0.0) ) {
      i--;
   }
   if (i >= 0)
      pt = (SB3TrkSeg_Turn *)seg[i];
      
   i = get_ts_turn_idx() + 1;
   while ( (i < num_segs)&&(seg[i]->rad == 0.0) ) {
      i++;
   }
   if (i < num_segs) {
      nt = (SB3TrkSeg_Turn *)seg[i];
      if (pt == NULL)
         pt = nt;
   }
   else if (pt != NULL)
      nt = pt;

   min_len = 3.5 * trk_wid;
   apex    = APEX_MID;
   if ( (plen < min_len)&&(!pt->like_straight())&&
        (nlen < min_len)&&(!nt->like_straight()) ) {
      // no straights between the prev turn, this turn, and next turn
      if ( (ts_turn->same_dir(*pt))&&(ts_turn->same_dir(*nt)) ) {
         // In the middle of 3 or more corners in same direction
         if (ts_turn->like_straight())  // this turn can be considered straight
            min_w = max_w - 5.0;
         else if ( (plen == 0.0)&&(nlen == 0.0) )
            max_w = min_w + 1.0;
         else {
            if (plen == 0.0)
               apex = -1.0;
            else if (nlen == 0.0)
               apex =  2.0;
            if ( (plen < nlen)||(nlen == 0.0) )
               max_w = (0.5 + 0.5*(plen/min_len))*(max_w - min_w);
            else
               max_w = (0.5 + 0.5*(nlen/min_len))*(max_w - min_w);
            max_w += min_w;
         }
      }
      else if (ts_turn->same_dir(*pt)) {
         // 2 turns in same dir, followed by 1 in opposite dir
         if (ts_turn->like_straight())  // this turn can be considered straight
            apex = APEX_END;
         else if (plen == 0.0)
            apex = -1.0;
         else
            apex = get_apex(plen, min_w, max_w);
      }
      else if (ts_turn->same_dir(*nt)) {
         // 2 turns in same dir, preceeded by 1 in opposite dir
         if (ts_turn->like_straight())  // this turn can be considered straight
            apex = APEX_BEG;
         else if (nlen == 0.0)
            apex = 2.0;
         else
            apex = 1.0 - get_apex(nlen, min_w, max_w);
      }
      else {
         // prev and next turns in same dir, this turn in opposite dir
         if (ts_turn->like_straight())  // this turn can be considered straight
            max_w = min_w + 5.0;
      }
   }
   else if (ts_turn->like_straight()) {
      // this turn can be considered straight
      if (pt->same_dir(*nt)) {
         if (ts_turn->same_dir(*pt))
            min_w = max_w - 5.0;
         else
            max_w = min_w + 5.0;
      }
      else {
         if (ts_turn->same_dir(*pt))
            apex = APEX_END;
         else
            apex = APEX_BEG;
      }
   }
   else if ( (plen < min_len)&&(!pt->like_straight()) ) {
      // no straight between the prev turn, and this turn
      // there is a straight between this turn, and the next
      if (ts_turn->same_dir(*pt)) { // this turn in same dir as prev
         if (plen == 0.0)
            apex = -1.0;
         else
            apex = get_apex(plen, min_w, max_w);
      }
   }
   else if ( (nlen < min_len)&&(!nt->like_straight()) ) {
      // there is a straight between the prev turn, and this turn
      // no straight between this turn, and the next
      if (ts_turn->same_dir(*nt)) { // this turn in same dir as next
         if (nlen == 0.0)
            apex = 2.0;
         else
            apex = 1.0 - get_apex(nlen, min_w, max_w);
      }
   }
   set(apex, min_w, max_w, trk_wid);
}

void SB3Turn::copy(const SB3Turn &tc, int copy_segs)
{
   if (copy_segs)
      SB3PathComponent::copy(tc);

   PathDir   = tc.PathDir;
   curv.c.x  = tc.curv.c.x;
   curv.c.y  = tc.curv.c.y;
   curv.R    = tc.curv.R;
   ts_turn   = tc.ts_turn;
   apex_dist = tc.apex_dist;
   plen      = tc.plen;
   nlen      = tc.nlen;
}

void SB3Turn::copy(const SB3Turn &tc, int copy_segs, int trim_tail,
                  int extra_segs)
{
   int end;

   copy(tc, copy_segs);
   if (copy_segs) {
      end = get_ts_turn_idx();
      if (trim_tail) {
         end += extra_segs;
         if (end >= num_segs)
            end = num_segs - 1;
         if (is_after(Fdst, *this, seg[end]->Fdst))
            Fdst = seg[end]->Fdst;
         num_segs = end + 1;
      }
      else {
         end -= extra_segs;
         if (end < 0)
            end = 0;
         if (is_after(seg[end]->Sdst, *this, Sdst))
            Sdst = seg[end]->Sdst;
         while (end > 0) {
            remove_TrkSeg(0);
            end--;
         }
      }
   }
   else {
      if (trim_tail)
         Sdst = tc.Sdst;
      else
         Fdst = tc.Fdst;
   }
}

void SB3Turn::adjust(SB3Turn &pt, SB3Turn &nt, int adj_val,
                    double min_w, double max_w, double trk_wid)
{
   double p_dst, n_dst, weight;

   if (adj_val == 0)
      return;
   else if ((adj_val&BAD_TURN) != 0) {
      fix_bad_turn(min_w, max_w, trk_wid);
      return;
   }

   // find the error at the entry and exit of this turn
   p_dst = get_err(pt, 0, trk_wid);
   n_dst = nt.get_err(*this, 1, trk_wid);

   if (ts_turn->like_straight()) {
      // adjust these turns here
      if ( (p_dst < 0.0)&&(n_dst < 0.0) )
         moveTurn( 2.0, min_w, max_w);  // move turn closer to inside wall
      else if ( (p_dst > 2.0)&&(n_dst > 2.0) )
         moveTurn(-2.0, min_w, max_w);  // move turn closer to outside wall
      else if (p_dst < 0.0)
         moveApex(-ADJ_ANG, min_w, max_w, trk_wid);  // move apex closer to entry
      else if (n_dst < 0.0)
         moveApex( ADJ_ANG, min_w, max_w, trk_wid);  // move apex closer to exit
      return;
   }
   
   // determine how much adjustment this turn will allow
   if (adj_val == BAD_ENTRY)
      weight = get_weight(pt);
   else if (adj_val == BAD_EXIT)
      weight = get_weight(nt);
   else
      weight = 0.5 * (get_weight(pt) + get_weight(nt));

   if (adj_val == BAD_ENTRY) {
      if (p_dst < 0.0) {
         // move entry of turn closer to inside wall
         if ( (ts_turn->Sdst < apex_dist)&&(apex_dist <= ts_turn->Fdst) ) {
            // move apex to entry
            moveApex((-ADJ_ANG*weight), min_w, max_w, trk_wid);
         }
         else if (apex_dist > ts_turn->Fdst) {
            // reduce turn radius
            weight *= -ADJ_RAD * curv.R;
            min_w  -= 0.1;
            moveTurn(weight, min_w, max_w);
            adjRadius(weight, min_w, max_w);
         }
      }
      else if (p_dst > 2.0) {
         // move entry of turn closer to outside wall
         if ( (ts_turn->Sdst <= apex_dist)&&(apex_dist < ts_turn->Fdst) ) {
            // move apex to exit
            moveApex(( ADJ_ANG*weight), min_w, max_w, trk_wid);
         }
      }
   }
   else if (adj_val == BAD_EXIT) {
      if (n_dst < 0.0) {
         // move exit of turn closer to inside wall
         if ( (ts_turn->Sdst <= apex_dist)&&(apex_dist < ts_turn->Fdst) ) {
            // move apex to exit
            moveApex(( ADJ_ANG*weight), min_w, max_w, trk_wid);
         }
         else if (ts_turn->Sdst > apex_dist) {
            // reduce turn radius
            weight *= -ADJ_RAD * curv.R;
            min_w  -= 0.1;
            moveTurn(weight, min_w, max_w);
            adjRadius(weight, min_w, max_w);
         }
      }
      else if (n_dst > 2.0) {
         // move exit of turn closer to outside wall
         if ( (ts_turn->Sdst < apex_dist)&&(apex_dist <= ts_turn->Fdst) ) {
            // move apex to entry
            moveApex((-ADJ_ANG*weight), min_w, max_w, trk_wid);
         }
      }
   }
   else if (p_dst < 0.0) {
      // move entry of turn closer to inside wall
      if (n_dst > 2.0) {
         // move exit of turn closer to outside wall
         if (ts_turn->Sdst < apex_dist) {
            // move apex to entry
            moveApex((-ADJ_ANG*weight), min_w, max_w, trk_wid);
         }
      }
      else {
         // reduce turn radius
         weight *= -ADJ_RAD * curv.R;
         moveTurn(weight, min_w, max_w);
         adjRadius(weight, min_w, max_w);
      }
   }
   else if ( (p_dst > 2.0)&&(n_dst < 0.0)&&
             (apex_dist < ts_turn->Fdst) ) {
      // move entry of turn closer to outside wall
      // move exit of turn closer to inside wall
      // move apex to exit
      moveApex(( ADJ_ANG*weight), min_w, max_w, trk_wid);
   }
}

int SB3Turn::makeStraight(SB3Straight &ret, SB3Turn &pt, double trk_wid)
{
   int    i, j, k, k1, ts1, ts2;
   double d1, d2;
   SB3Line s;

   ts1 = get_ts_turn_idx();
   ts2 = pt.get_ts_turn_idx();

   // first check if straight is needed
   s.dir.x = pt.curv.c.x - curv.c.x;
   s.dir.y = pt.curv.c.y - curv.c.y;
   d1      = s.dir.mag();
      
   if (ts_turn->same_dir(*(pt.ts_turn))) {
      j   = SB3Circle::TAN_1;
      d1 -= fabs(curv.R - pt.curv.R);
   }
   else {
      j   = SB3Circle::TAN_3;
      d1 -= curv.R + pt.curv.R;
   }

   if (fabs(d1) < 2.5) {
      curv.getPoint(s.pt, s.dir.ang());
      k = find_trk_seg(s.pt, trk_wid, ts1, -1);
      if (k >= 0) {
         // do not need a straight here
         ret.adjust(s.pt, s.pt);
         ret.num_segs = 0;
         ret.insert_TrkSeg(seg[k], 0);

         // start and finish locations
         d2 = seg[k]->get_trk_dist(s.pt);
         ret.Sdst = pt.Fdst = d2;
         ret.Fdst = Sdst    = d2;
         return 1;
      }
   }

   for (i = 0; i < 2; i++) {
      j += i;
      if (!curv.tangent(s, pt.curv, j))
         continue;
      s.dir.add(s.pt);

      k = find_trk_seg(s.pt, trk_wid, ts1, -1);
      if (k < 0) {
         k = find_trk_seg(s.pt, trk_wid, ts1+1, 1);
         if (k < 0)
            continue;
      }
      k1 = k;
      d1 = seg[k]->get_trk_dist(s.pt);
      if ( (ts_turn->len > (0.7*MY_PI))&&
           (!is_after(ts1, ts_turn->Fdst, *this, k, d1)) )
         continue;

      k = pt.find_trk_seg(s.dir, trk_wid, ts2, 1);
      if (k < 0) {
         k = pt.find_trk_seg(s.dir, trk_wid, ts2-1, -1);
         if (k < 0)
            continue;
      }
      d2 = pt.seg[k]->get_trk_dist(s.dir);
      if ( (pt.ts_turn->len > (0.7*MY_PI))&&
           (!pt.is_after(k, d2, pt, ts2, pt.ts_turn->Sdst)) )
         continue;

      if (is_after(k1, d1, pt, k, d2)) {
         ts1 = k1;
         ts2 = k;
         break;
      }
   }

   if (i < 2) {
      ret.adjust(s.dir, s.pt);

      // insert TrkSeg into Straight component
      ret.num_segs = 0;
      for (k = pt.num_segs-1; k >= ts2; k--) {
         ret.insert_TrkSeg(pt.seg[k], 0);
      }
      for (k = 0; k <= ts1; k++) {
         ret.insert_TrkSeg(seg[k], ret.num_segs);
      }

      // start and finish locations
      ret.Sdst = pt.Fdst = d2;
      ret.Fdst = Sdst    = d1;
      return 1;
   }
   return 0;
}

void SB3Turn::set_spd_len(int data_loaded, int set_spd)
{
   SB3Vector p;
   SB3Line   n1, n2;

   // set the length of this component
   getPoint(p, n1, Sdst, 1);
   Sang = curv.ang(p);
   getPoint(p, n2, (Sdst+2.0), 1);
   if (n1.dir.ang(n2.dir) < 0.0)
      PathDir = -1;   // right turn
   else
      PathDir =  1;   // left turn
   getPoint(p, n2, Fdst, 1);
   len = n1.dir.ang(n2.dir);
   if ( (PathDir > 0)&&(len < 0.0) )
      len += MY_PI21;
   else if ( (PathDir < 0)&&(len > 0.0) )
      len -= MY_PI21;
   p.x = fabs(len);
   len = fabs(len) * curv.R;
   
   if (!set_spd)
      return;
   else if (set_spd >= 2) {
      if ( (set_spd == 2)&&(len < (1.5*SB3_MIN_LEN)) )
         min_spd = 1.05;
      else
         min_spd = 1.0;
      data_loaded = 1;
   }
   // set the speed for this turn
#ifndef SB3_DLL
   switch (args.m_iSurface) {
#else
   switch (args->m_iSurface) {
#endif
   case 0:
      min_spd *= 0.945;
      max_spd *= 0.945;
      if (min_spd > 1.0)
         min_spd = 1.0;
      if (max_spd > 1.05)
         max_spd = 1.05;
      break;
   case 2:
      min_spd *= 0.99;
      max_spd *= 0.985;
//      if (max_spd < min_spd)
//         max_spd = min_spd;
   };

#ifndef SB3_DLL
   if ( (data_loaded)||(args.m_iSurface == 0) ) {
#else
   if ( (data_loaded)||(args->m_iSurface == 0) ) {
#endif
      min_spd *= SCALE_CORNSPD * SB3_crn_spd(curv.R);
      if (set_spd >= 2) {
         max_spd  = 2.0 * (0.04*MAX_ACCEL) * brk_dist(len, min_spd);
         max_spd += min_spd * min_spd;
         max_spd  = sqrt(max_spd);
      }
      else
         max_spd *= SCALE_CORNSPD * SB3_crn_spd(curv.R);
   }
   else {
      if (p.x < 0.175) {
         min_spd = max_spd = SB3_MAX_SPD;
      }
      else {
         min_spd *= SCALE_CORNSPD * SB3_crn_spd(curv.R);
         max_spd *= SCALE_CORNSPD * SB3_crn_spd(curv.R);
         if ( (p.x < 0.37)&&(len < 150.0) ) {
            min_spd *= ((p.x-0.37)*(1.6-1.1)/(0.175-0.37)) + 1.1;
            max_spd *= ((p.x-0.37)*(1.6-1.1)/(0.175-0.37)) + 1.1;
         }
      }
   }
   if (min_spd > SB3_MAX_SPD)
      min_spd = SB3_MAX_SPD;
   if (max_spd > SB3_MAX_SPD)
      max_spd = SB3_MAX_SPD;
}

int SB3Turn::makePass(SB3Straight *p_s, SB3Turn *p_t, SB3Turn &nt,
                     SB3Turn &last, double min_w, double max_w, 
                     double trk_wid)
{
   int      retval;
   double   dist;
   SB3Vector p;
   SB3Line   n;

   if ( (!allow_pass)||(nt.ts_turn->like_straight()) )
      return 0;  // turn too fast for passing

   dist  = nt.curv.R - nt.curv.c.dist2Point(nt.ts_turn->ls.c);
   dist -= fabs(nt.ts_turn->rad);
   if (dist > max_w)
      return 0;  // apex point too far from inner edge

   // attempt to create a passing path
   p_t->copy(nt, 1);
   p_t->pass    = NULL;
   p_t->pre_len = p_t->post_len = 0.0;
   p_t->set(nt.get_apex(), min_w, max_w, trk_wid);
   dist      = Fdst;
   retval    = p_t->makeStraight(*p_s, *this, trk_wid);
   Fdst      = dist;
   p_t->Fdst = nt.dst_between(nt.Fdst, nt, nt.apex_dist);
   if (p_t->Fdst > 0.0)
      p_t->Fdst *= 0.25;
   else
      p_t->Fdst  = 0.0;
   p_t->Fdst += nt.apex_dist;

   p_t->min_spd = 1.03;
   p_t->max_spd = 1.04;
   p_t->set_spd_len(1);

   // check if path is valid
   retval = ( (retval)&&(p_s->len > 20.0)&&(p_t->len > 20.0)&&
              (p_s->valid(trk_wid))&&
              (is_after(p_s->Sdst, *this, Sdst))&&
              (p_t->is_after(p_t->Fdst, *p_t, p_t->Sdst)) );

   if (retval) {
      // make sure passing path converges with racing path
      dist   = nt.to_side(p, n, p_t->Fdst) - p_t->to_side(p, n, p_t->Fdst);
      retval = (fabs(dist) <= CARWID);
   }
   if (retval) {
      // make sure passing path isn't too close to racing path
      dist = nt.to_side(p, n, nt.Sdst);
      if (p_s->is_after(p_s->Fdst, nt, nt.Sdst))
         dist -= p_s->to_side(p, n, nt.Sdst);
      else
         dist -= p_t->to_side(p, n, nt.Sdst);
      if (nt.ts_turn->rad < 0.0)
         dist = -dist;
      retval = (dist >= CARLEN);
   }

   if ( (retval)&&(last.pass != NULL) ) {
      // make sure there is no overlap
      if (is_after(last.pass[1]->Fdst, *this, p_s->Sdst)) {
         // there is an overlap, so decide which one to keep
         if ((nt.ts_turn->len*nt.get_apex()) > (ts_turn->len*get_apex())) {
            // drop the last one
            delete last.pass[0];
            delete last.pass[1];
            delete [] last.pass;
            last.pass    = NULL;
            last.pre_len = 0.0;
         }
         else
            retval = 0;  // drop this one
      }
   }

   pre_len = nt.post_len = 0.0;
   if (retval) {
      // store path
      if (pass != NULL) {
         // remove these
         delete pass[0];
         delete pass[1];
      }
      else
         pass = new SB3PathComponent*[2];
      p_s->trim();
      p_t->trim();
      pass[0] = p_s;
      pass[1] = p_t;

      p_s->len_trk = p_s->dst_between(p_s->Fdst, *p_s, p_s->Sdst);
      p_t->len_trk = p_t->dst_between(p_t->Fdst, *p_t, p_t->Sdst);

      getPoint(p, n, p_s->Sdst);
      pre_len = get_path_dist(p);
      
      nt.getPoint(p, n, p_t->Fdst);
      nt.post_len = nt.len - nt.get_path_dist(p);
   }
   else if (pass != NULL) {
      // clean up components
      p_t->num_segs = 0;
      p_s->num_segs = 0;

      // remove these
      delete pass[0];
      delete pass[1];
      delete [] pass;
      pass = NULL;
   }
   return retval;
}

void SB3Turn::save(FILE *out, SB3TrkSeg **ts, double rotate) const
{
   int      i;
   SB3Vector cen(curv.c);

   cen.sub(ts_turn->ls.c);
   cen.rotate(rotate);

   // SAVE LINE 1
   i = 0;
   while (ts[i] != ts_turn) {
      i++;
   }
   fprintf(out, "%d ", i);
   i = 0;
   while (ts[i] != seg[0]) {
      i++;
   }
   fprintf(out, "%d %d ", i, num_segs);

   fprintf(out, "%10.3lf %10.3lf %10.3lf\n", Sdst, apex_dist, Fdst);

   // SAVE LINE 2
   fprintf(out, "%11.3lf %11.3lf %10.3lf ", cen.x, cen.y, curv.R);

   cen.x = min_spd / (SCALE_CORNSPD * SB3_crn_spd(curv.R));
   cen.y = max_spd / (SCALE_CORNSPD * SB3_crn_spd(curv.R));
   fprintf(out, "%5.2lf %5.2lf\n", cen.x, cen.y);
}

int SB3Turn::load(FILE *in, SB3TrkSeg **ts, int trk_segs, double rotate)
{
   const char *ws = " \t\n\r";
   int  i, j, k;
   char buf[128];

   // LOAD LINE 1
   if (!fgets(buf, 128, in))
      return 0;
   k       = atoi(strtok(buf, ws));
   if ( (k < 0)||(k >= trk_segs) )
      return 0;
   ts_turn = (SB3TrkSeg_Turn *)ts[k];
   i       = atoi(strtok(NULL, ws));
   j       = atoi(strtok(NULL, ws)) - 1;
   if ( (i < 0)||(i >= trk_segs)||(j < 0)||(j >= trk_segs) )
      return 0;
   for (; j >= 0; j--) {
      insert_TrkSeg(ts[(i+j)%trk_segs], 0);
   }

   plen = nlen = 0.0;
   for (i = 1, j = 0; i < trk_segs; i++) {
      j = (k + trk_segs - i) % trk_segs;
      if (ts[j]->rad != 0.0)
         break;
      plen += ts[j]->len;
   }
   for (i = 1, j = 0; i < trk_segs; i++) {
      j = (k + i) % trk_segs;
      if (ts[j]->rad != 0.0)
         break;
      nlen += ts[j]->len;
   }

   Sdst      = atof(strtok(NULL, ws));
   apex_dist = atof(strtok(NULL, ws));
   Fdst      = atof(strtok(NULL, ws));

   // LOAD LINE 2
   if (!fgets(buf, 128, in))
      return 0;
   curv.c.x = atof(strtok(buf, ws));
   curv.c.y = atof(strtok(NULL, ws));
   curv.R   = atof(strtok(NULL, ws));
   curv.c.rotate(rotate);
   curv.c.add(ts_turn->ls.c);

   min_spd = atof(strtok(NULL, ws));
   max_spd = atof(strtok(NULL, ws));

   allow_pass = (strtok(NULL, ws) == NULL);
   return 1;
}

int SB3Turn::makeTransitions(SB3Turn *pt, SB3Straight *ps, SB3Trans &Tentry,
                            SB3Trans &Texit, SB3Straight *ns, SB3Turn *nt,
                            double trk_wid)
{
   int      retval, i, rc;
   double   len_x1, str_ang1, len_ang1, ang1, len_ang,
            len_x2, str_ang2, len_ang2, ang2, apex_ang;
   SB3Line   n1, n2;
   SB3Trans  t1, t2;
   SB3Circle new_curv, org_curv(curv);
   double   new_Sdst, new_Fdst,
            org_Sdst = Sdst,
            org_Fdst = Fdst;

   if (curv.R > 2200.0)
      return 0;

   // get tangent angles at the ends of this turn
   getPoint(n2.pt, n1, Sdst, 1);
   len_ang1 = curv.ang(n2.pt);
   str_ang1 = n1.dir.ang();
   ang1     = 0.0;
   getPoint(n1.pt, n2, Fdst, 1);
   len_ang2 = curv.ang(n1.pt);

   if (pre_len > len)
      getPoint(n1.pt, n2, pass[0]->Sdst, 1);
   str_ang2 = n2.dir.ang();
   len_ang  = n1.dir.ang(n2.dir);
   if ( (ts_turn->rad > 0.0)&&(len_ang < 0.0) )
      len_ang += MY_PI21;
   else if ( (ts_turn->rad < 0.0)&&(len_ang > 0.0) )
      len_ang -= MY_PI21;
   len_ang = fabs(len_ang);
   ang2    = len_ang - (len / curv.R);

   n1.pt.y  = len / curv.R;
   n1.pt.x  = dst_between(apex_dist, *this, Sdst);
   n1.pt.x /= dst_between(Fdst, *this, Sdst);
   n1.pt.x *= n1.pt.y;
   if ( (n1.pt.y > 0.35)&&((n1.pt.y-n1.pt.x) < 0.14) ) {
      apex_ang = len_ang2;
      ang1    += n1.pt.y;
      i        = 0;
   }
   else {
      apex_ang = len_ang1;
      ang2    += n1.pt.y;
      if (n1.pt.x < 0.14)
         i  = 0;
      else {
         i  = (int)(n1.pt.y / 0.043633) + 1;
         i &= 0xFFFFFFFE;
         if (i > 30)
            i = 30;
         else if (i < 6)
            i = 6;
         n1.pt.x = n1.pt.y / (double)i;
         if (ts_turn->rad < 0.0)
            n1.pt.y = -n1.pt.x;
         else
            n1.pt.y =  n1.pt.x;
      }
   }

   // divide the straight components
   if (ps == NULL)
      len_x1 = 0.0;
   else if (pt == NULL)
      len_x1 = ps->len;
   else
      len_x1 = (pt->curv.R / (pt->curv.R+curv.R)) * ps->len;
      
   if (pre_len > len)
      len_x2 = pass[0]->len;
   else if (ns == NULL)
      len_x2 = 0.0;
   else if (nt == NULL)
      len_x2 = ns->len;
   else
      len_x2 = (nt->curv.R / (nt->curv.R+curv.R)) * ns->len;

   // try and find the best component(s) for this turn
   t1.SB3PathComponent::copy(*this);
   t2.SB3PathComponent::copy(*this);
   n2.pt.x  = 0.5;
   n1.dir.x = curv.ang(ts_turn->ls.c);
   n1.dir.y = sin(n1.dir.x);
   n1.dir.x = cos(n1.dir.x);
   for (retval = 0; i >= 0; i--) {
      if (ang1 >= (MY_PI12+0.043633))
         len_ang1 = MY_PI12;
      else if (ang1 > 0.043633)
         len_ang1 = ang1 - 0.043633;
      else
         len_ang1 = 0.0;
      if (ang2 >= (MY_PI12+0.043633))
         len_ang2 = MY_PI12;
      else if (ang2 > 0.043633)
         len_ang2 = ang2 - 0.043633;
      else
         len_ang2 = 0.0;

      rc  = adjustTransitions(t1, str_ang1, ang1, len_ang1, len_x1,
                              t2, str_ang2, ang2, len_ang2, len_x2,
                              trk_wid, apex_ang);
   	rc &= 0x03;
      if (rc == 0) {
         curv.c.x += n1.dir.x;
         curv.c.y += n1.dir.y;
         curv.R   -= 0.5;
         rc  = adjustTransitions(t1, str_ang1, ang1, len_ang1, len_x1,
                                 t2, str_ang2, ang2, len_ang2, len_x2,
                                 trk_wid, apex_ang);
         rc &= 0x03;
         if (rc == 0)
            rc = 0x04;
      }
      n2.pt.y = 0.0;
      if ((rc & 0x01) != 0)
         n2.pt.y += t1.rating();
      if ((rc & 0x02) != 0)
         n2.pt.y += t2.rating();
      if (n2.pt.y > n2.pt.x) {
         n2.pt.x      = n2.pt.y;
         n2.dir.x     = len_ang1;
         n2.dir.y     = len_ang2;
         retval       = rc;
         new_Sdst     = Sdst;
         new_Fdst     = Fdst;
         new_curv.c.x = curv.c.x;
         new_curv.c.y = curv.c.y;
         new_curv.R   = curv.R;
         if ((rc & 0x01) != 0)
            Tentry.copy(t1);
         if ((rc & 0x02) != 0)
            Texit.copy(t2);
      }
      if (rc != 0) {
         Sdst     = org_Sdst;
         Fdst     = org_Fdst;
         curv.c.x = org_curv.c.x;
         curv.c.y = org_curv.c.y;
         curv.R   = org_curv.R;
      }
      ang1     += n1.pt.x;
      ang2     -= n1.pt.x;
      apex_ang += n1.pt.y;
   }

   if (retval != 0) {
      // adjust this turn to accomodate the new component(s)
      if ((retval & 0x01) != 0)
         len_ang -= n2.dir.x;
      if ((retval & 0x02) != 0)
         len_ang -= n2.dir.y;
      Sdst     = new_Sdst;
      Fdst     = new_Fdst;
      curv.c.x = new_curv.c.x;
      curv.c.y = new_curv.c.y;
      curv.R   = new_curv.R;
      len      = len_ang * curv.R;
   }
   t1.pass = t2.pass = NULL;
   return retval;
}

// PRIVATE METHODS
int SB3Turn::get_ts_turn_idx() const
{
   int i;

   for (i = 0; i < num_segs; i++) {
      if (seg[i] == ts_turn)
         return i;
   }
   return 0;
}

double SB3Turn::get_apex(double dst, double min_w, double max_w) const
{
   double A, B, C, dA,
          dB = 0.0,
          dC = cos(ts_turn->len),
          b  = dst / tan(0.5*ts_turn->len),
          d  = dst / sin(0.5*ts_turn->len);
   int    i  = 0;

   d *= d;
   B  = 0.0;
   do {
      i++;
      B += 0.5 * dB;
      A  = sqrt((B*B) + (2.0*B*b) + d);
      C  = dC * B;
      dA = (B + b) / A;
      dB = (b + C + max_w - A - min_w) / (dA - dC);
   } while ( (i < 50)&&(fabs(dB) > 1.0) );

   if (fabs(dB) > 1.0)
      return 0.5;

   B = atan(dst / (b+B));
   if (B < 0.0)
      B += MY_PI;

   B /= ts_turn->len;
   if (B < 0.0)
      B = 0.0;
   else if ( (0.5 < B)&&(B <= 1.0) )
      B = 1.0 - B;
   return B;
}

double SB3Turn::get_apex() const
{
   double apex;

   apex = ts_turn->Fdst - ts_turn->Sdst;
   if (apex != 0.0)
      apex = (apex_dist-ts_turn->Sdst) / apex;
   return apex;
}

void SB3Turn::moveApex(double amount, double min_w, double max_w,
                      double trk_wid)
{
   double apex  = apex_dist - ts_turn->Sdst,
          old_R = curv.R;

   apex = get_apex() + (amount / ts_turn->len);
   if (apex > APEX_END)
      apex = APEX_END;
   else if (apex < APEX_BEG)
      apex = APEX_BEG;

   set(apex, min_w, max_w, trk_wid);
   if (old_R < curv.R) {
      apex   = old_R - curv.R;
      curv.R = old_R;
      moveTurn(apex, min_w, max_w);
   }
}

double SB3Turn::adjRadius(double amount, double min_w, double max_w)
{
   double Ad = curv.c.dist2Point(ts_turn->ls.c);

   if ((curv.R-Ad+amount) < (fabs(ts_turn->rad)+min_w))
      amount = fabs(ts_turn->rad) + min_w - curv.R + Ad;
   else if ((curv.R-Ad+amount) > (fabs(ts_turn->rad)+max_w))
      amount = fabs(ts_turn->rad) + max_w - curv.R + Ad;
   curv.R += amount;
   return amount;
}

double SB3Turn::moveTurn(double amount, double min_w, double max_w)
{
	double Ad;
	SB3Line l;

	l.pt.add(curv.c);
	l.dir.add(curv.c);
	l.dir.sub(ts_turn->ls.c);
	Ad = l.dir.mag();
	if (Ad == 0.0)
		return 0.0;
	else if ((curv.R-Ad-amount) < (fabs(ts_turn->rad)+min_w))
		amount = curv.R - Ad - fabs(ts_turn->rad) - min_w;
	else if ((curv.R-Ad-amount) > (fabs(ts_turn->rad)+max_w))
		amount = curv.R - Ad - fabs(ts_turn->rad) - max_w;
	if ((Ad+amount) < 0.5)
		amount = 0.5 - Ad;
	l.dir.normalize();
	l.getPoint(curv.c, amount);
	return amount;
}

void SB3Turn::setLimits(double trk_wid)
{
	int            i;
	double         dst;
	SB3Line         a;
	SB3Circle      *cir;
	SB3TrkSeg_Turn *trn;
		
	a.dir.x = ts_turn->ls.c.x - curv.c.x;
	a.dir.y = ts_turn->ls.c.y - curv.c.y;
	dst     = a.dir.mag();
	Sdst    = dst * sin(get_apex()*ts_turn->len);
	Fdst    = dst * sin((1.0 - get_apex())*ts_turn->len);

	if (Sdst <= plen) {
		if (Sdst <= 0.0)
			Sdst = ts_turn->Sdst;
		else {
			dst = Sdst;
			i   = get_ts_turn_idx() - 1;
			for (; (i >= 0)&&(dst > 0.0); i--) {
				dst -= seg[i]->Fdst - seg[i]->Sdst;
			}
			Sdst = seg[i+1]->Sdst - dst;
		}
	}
	else {
		if ( (ts_turn->like_straight())||(get_apex() < 0.01) )
			Sdst = ts_turn->Sdst;
		else {
			dst = Sdst;
			i   = get_ts_turn_idx() - 1;
			for (; (i >= 0)&&(dst > 0.0); i--) {
				if (seg[i]->rad == 0.0) {
					seg[i]->get_norm(a, seg[i]->Sdst);
					cir = &curv;
				}
				else {
					trn     = (SB3TrkSeg_Turn *)seg[i];
					a.dir.x = trn->ls.c.x - curv.c.x;
					a.dir.y = trn->ls.c.y - curv.c.y;
					cir     = &(trn->ls);
				}
			
				cir->getPoint(a.pt, a.dir.ang());
				Sdst = seg[i]->get_trk_dist(a.pt);
				if ( (Sdst > (seg[i]->Sdst-2.5))&&
				     (Sdst < (seg[i]->Fdst+2.5)) )
					break;
				cir->getPoint(a.pt, (a.dir.ang()+MY_PI));
				Sdst = seg[i]->get_trk_dist(a.pt);
				if ( (Sdst > (seg[i]->Sdst-2.5))&&
				     (Sdst < (seg[i]->Fdst+2.5)) )
					break;
            if (!seg[i]->path_ok(curv, seg[i]->Sdst, seg[i]->Fdst,
                                 0.0, trk_wid)) {
               Sdst = seg[i]->Fdst;
               break;
            }
				dst -= seg[i]->Fdst - seg[i]->Sdst;
			}
			if ( (dst <= 0.0)||(i == -1) )
				Sdst = seg[i+1]->Fdst;
         else if (!seg[i]->path_ok(curv, Sdst, seg[i]->Fdst,
                                   0.0, trk_wid)) {
            do {
               Sdst *= 0.5;
               Sdst += 0.5 * seg[i]->Fdst;
               if ((seg[i]->Fdst-Sdst) < 5.0)
                  break;
            } while (!seg[i]->path_ok(curv, Sdst, seg[i]->Fdst,
                                      0.0, trk_wid));
         }
		}
	}

	if (Fdst <= nlen) {
		if (Fdst <= 0.0)
			Fdst = ts_turn->Fdst;
		else {
			dst = Fdst;
			i   = get_ts_turn_idx() + 1;
			for (; (i < num_segs)&&(dst > 0.0); i++) {
				dst -= seg[i]->Fdst - seg[i]->Sdst;
			}
			Fdst = seg[i-1]->Fdst + dst;
		}
	}
	else {
		if ( (ts_turn->like_straight())||(get_apex() > 0.99) )
			Fdst = ts_turn->Fdst;
		else {
			dst = Fdst;
			i   = get_ts_turn_idx() + 1;
			for (; (i < num_segs)&&(dst > 0.0); i++) {
				if (seg[i]->rad == 0.0) {
					seg[i]->get_norm(a, seg[i]->Fdst);
					cir = &curv;
				}
				else {
					trn     = (SB3TrkSeg_Turn *)seg[i];
					a.dir.x = trn->ls.c.x - curv.c.x;
					a.dir.y = trn->ls.c.y - curv.c.y;
					cir     = &(trn->ls);
				}

				cir->getPoint(a.pt, (a.dir.ang()+MY_PI));
				Fdst = seg[i]->get_trk_dist(a.pt);
				if ( (Fdst > (seg[i]->Sdst-2.5))&&
				     (Fdst < (seg[i]->Fdst+2.5)) )
					break;
				cir->getPoint(a.pt, a.dir.ang());
				Fdst = seg[i]->get_trk_dist(a.pt);
				if ( (Fdst > (seg[i]->Sdst-2.5))&&
				     (Fdst < (seg[i]->Fdst+2.5)) )
					break;
            if (!seg[i]->path_ok(curv, seg[i]->Sdst, seg[i]->Fdst,
                                 0.0, trk_wid)) {
               Fdst = seg[i]->Sdst;
               break;
            }
				dst -= seg[i]->Fdst - seg[i]->Sdst;
			}
			if ( (dst <= 0.0)||(i == num_segs) )
				Fdst = seg[i-1]->Sdst;
         else if (!seg[i]->path_ok(curv, seg[i]->Sdst, Fdst,
                                   0.0, trk_wid)) {
            do {
               Fdst *= 0.5;
               Fdst += 0.5 * seg[i]->Sdst;
               if ((Fdst-seg[i]->Sdst) < 5.0)
                  break;
            } while (!seg[i]->path_ok(curv, seg[i]->Sdst, Fdst,
                                      0.0, trk_wid));
         }
		}
	}
}

void SB3Turn::set(double apex, double min_w, double max_w, double trk_wid)
{
	double Ad;
	SB3Line a;

	if (apex == APEX_BEG)
		apex_dist = ts_turn->Sdst;
	else if (apex == APEX_END)
		apex_dist = ts_turn->Fdst;
	else if ( (APEX_BEG < apex)&&(apex < APEX_END) )
		apex_dist = (apex*ts_turn->Fdst) + ((1.0-apex)*ts_turn->Sdst);
	else {
		if (apex < APEX_BEG)
			apex_dist = ts_turn->Sdst - 2.0;
		else
			apex_dist = ts_turn->Fdst + 2.0;
		apex   = get_apex();
		min_w -= 0.1;
	}
	ts_turn->get_norm(a, apex_dist);
	if (ts_turn->rad > 0.0)
		a.dir.rotate(MY_PI);

   if (apex < APEX_MID) {
		apex *= ts_turn->len;
		Ad    = (max_w - min_w) / (1.0 - cos(ts_turn->len - apex));
	}
	else {
		apex *= ts_turn->len;
		Ad    = (max_w - min_w) / (1.0 - cos(apex));
	}

	a.pt.x = ts_turn->ls.c.x;
	a.pt.y = ts_turn->ls.c.y;
	a.getPoint(curv.c, Ad);
	curv.R = Ad + fabs(ts_turn->rad) + min_w;
		
	setLimits(trk_wid);
}

double SB3Turn::get_weight(const SB3Turn &ot) const
{
   double weight;

   weight = curv.R / (curv.R + ot.curv.R);
   if ( (0.15 < weight)&&(weight < 0.85) ) {
      double t = fabs(ts_turn->rad),
             o = fabs(ot.ts_turn->rad);
                
      weight  = 0.7 * (t / (t + o));
      t       = curv.R    / t;
      o       = ot.curv.R / o;
      weight += 0.3 * (o / (t + o));
   }
   return weight;
}

double SB3Turn::get_err(SB3Turn &pt, int get_pt_err, double trk_wid)
{
   double err;
      
   if (!ts_turn->same_dir(*(pt.ts_turn))) {
      err  = curv.c.dist2Point(pt.curv.c);
      err -= curv.R + pt.curv.R;
   }
   else {
      double org_Sdst    = Sdst,
             org_Fdst    = Fdst,
             org_pt_Sdst = pt.Sdst,
             org_pt_Fdst = pt.Fdst;

      pt.setLimits(trk_wid);
      setLimits(trk_wid);
      if (pt.is_after(pt.Fdst, *this, Sdst)) {
         err  = pt.dst_between(pt.Fdst, *this, Sdst);
         err /= pt.curv.R + curv.R;
         if (get_pt_err)
            err = -pt.curv.R;
         else
            err = -curv.R;
      }
      else {
         SB3Vector p;
         SB3Line   n;
            
         if (ts_turn->rad < 0.0)
            err = to_side(p, n, Sdst) - pt.to_side(p, n, pt.Fdst);
         else
            err = pt.to_side(p, n, pt.Fdst) - to_side(p, n, Sdst);
         if (get_pt_err)
            err = -err;
      }
      Sdst    = org_Sdst;
      Fdst    = org_Fdst;
      pt.Sdst = org_pt_Sdst;
      pt.Fdst = org_pt_Fdst;
   }
   return err;
}

void SB3Turn::fix_bad_turn(double min_w, double max_w, double trk_wid)
{
   int      i, j, end;
   double   w, start, seg_len,
            amount = 0.0;
   SB3Vector p;
   SB3Line   n;
      
   // check apex
   p.add(ts_turn->ls.c);
   p.sub(curv.c);
   w = curv.R - p.mag() - fabs(ts_turn->rad);
   if (w < (min_w-SB3_PATH_ERR))
      moveTurn((w-min_w), min_w, max_w);
   else if (w > (max_w+SB3_PATH_ERR))
      moveTurn((w-max_w), min_w, max_w);

   // check entry
   w = to_side(p, n, Sdst);
   if (ts_turn->rad < 0.0)
      w = (ts_turn->ls.R + ts_turn->rad) - w;
   if (w < (min_w+SB3_PATH_ERR)) {
      // increase radius
      if (w < (min_w-1.0))
         amount = min_w - w;
      else 
         amount = 1.0;
   }
   else {
      // decrease radius
      if (w > (max_w+2.0))
         amount = max_w - w;
      else
         amount = -2.0;
   }

   // check the rest of the turn
   i   = find_trk_seg(Sdst, 0, 1);
   end = find_trk_seg(Fdst, (num_segs-1), -1) + 1;
   for (; i < end; i++) {
      if ( (seg[i]->Sdst <= Sdst)&&(Sdst <= seg[i]->Fdst) )
         start = Sdst;
      else
         start = seg[i]->Sdst;
      if ( (seg[i]->Sdst <= Fdst)&&(Fdst <= seg[i]->Fdst) )
         seg_len = Fdst - start;
      else
         seg_len = seg[i]->Fdst - start;
      for (j = 1; j <= 5; j++) {
         w = to_side(p, n, (start + (0.2*(double)j*seg_len)));
         if (ts_turn->rad < 0.0)
            w = (ts_turn->ls.R + ts_turn->rad) - w;
         if (w < (min_w+SB3_PATH_ERR)) {
            // increase radius
            if (w < (min_w-1.0))
               w = min_w - w;
            else 
               w = 1.0;
         }
         else {
            // decrease radius
            if (w > (max_w+2.0))
               w = max_w - w;
            else
               w = -2.0;
         }

         // update the amount of adjustment needed
         if ( (amount > 0.0)&&(w > 0.0) ) {
            if (w > amount)
               amount = w;  // use larger value
         }
         else if (w < amount)
            amount = w;  // use smaller value
      }
   }

   // fix the turn
   curv.R += amount;
   moveTurn(0.0, min_w, max_w);
   setLimits(trk_wid);
}

int SB3Turn::adjustTransitions(SB3Trans &Tentry, double str_ang1, double ang1,
                              double len_ang1, double ps_len,
                              SB3Trans &Texit, double str_ang2, double ang2,
                              double len_ang2, double ns_len,
                              double trk_wid, double apex_ang)
{
	int      retval, i, Stype, Ftype;
	double   len_y, len_x, d;
	SB3Circle org_curv(curv);
   double   cs       = cos(apex_ang),
	         sn       = sin(apex_ang),
	         org_Sdst = Sdst,
	         org_Fdst = Fdst;

   if (ts_turn->rad > 0.0)  {
		Stype = SB3Trans.LENTRY;
		Ftype = SB3Trans.LEXIT;
	}
	else {
		Stype = SB3Trans.RENTRY;
		Ftype = SB3Trans.REXIT;
	}

   if ( (len_ang1 > 0.0)&&(ang1 >= len_ang1) )
      len_x = 1.0 - cos(ang1);
   else
      len_x = 0.0001;
   if ( (len_ang2 > 0.0)&&(ang2 >= len_ang2) )
      len_y = 1.0 - cos(ang2);
   else
      len_y = 0.0001;
   if (len_x < len_y)
      d = trk_wid / len_x;
   else
      d = trk_wid / len_y;
   d /= 0.00395 * org_curv.R;
   i  = (int)d - 8;
   if (i > 30)
      i = 30;
   else if (i < 1)
      i = 1;

   retval = 0;
	for (; i > 0; i--) {
      d        = 0.00395 * org_curv.R * (double)(i+9);  // 0.15 -> 0.0395
		curv.R   = org_curv.R - d;
		curv.c.x = org_curv.c.x + (d*cs);
		curv.c.y = org_curv.c.y + (d*sn);

   	Tentry.Fdst = Sdst - 1.0;
	   Texit.Sdst  = Fdst + 1.0;

      if ( (len_ang1 > 0.0)&&(ang1 >= len_ang1) ) {
   		len_y  = org_curv.R - (d*cos(ang1));
	   	len_x  = ps_len + (d*sin(ang1));
         len_x -= SB3Trans::get_len_x(curv.R, len_ang1, len_y);
   		if (len_x > -15.0) {
	   		retval |= 0x04;
		   	Tentry.adjust(curv, str_ang1, len_ang1, len_y, trk_wid, Stype);
			   if (Tentry.valid(trk_wid)) {
   				retval |= 0x0010;
	   			Sdst    = Tentry.Fdst;
		   		len_y  -= curv.R;
			   	if (len_y < 5.0)
				   	continue;
   			}
	   		else
		   		Tentry.Fdst = Sdst - 1.0;
   		}
      }
      else {
         Tentry.Fdst = Sdst;
         Tentry.len  = 0.0;
      }
		
      if ( (len_ang2 > 0.0)&&(ang2 >= len_ang2) ) {
   		len_y  = org_curv.R - (d*cos(ang2));
	   	len_x  = ns_len + (d*sin(ang2));
         len_x -= SB3Trans::get_len_x(curv.R, len_ang2, len_y);
   		if (len_x > -15.0) {
	   		retval |= 0x08;
		   	Texit.adjust(curv, str_ang2, len_ang2, len_y, trk_wid, Ftype);
   			if (Texit.valid(trk_wid)) {
	   			retval |= 0x0040;
		   		Fdst    = Texit.Sdst;
			   	len_y  -= curv.R;
   				if (len_y < 5.0)
	   				continue;
		   	}
			   else
   				Texit.Sdst = Fdst + 1.0;
	   	}
      }
      else {
         Texit.Sdst = Fdst;
         Texit.len  = 0.0;
      }

		if (valid(trk_wid)) {
			retval |= 0x0020;
			if ( (Sdst == Tentry.Fdst)&&(Fdst == Texit.Sdst) ) {
				if (Tentry.len > SB3_MIN_LEN)
					retval |= 0x01;
            else if (len_ang1 > 0.0)
               continue;
            else 
					Sdst = org_Sdst;

				if (Texit.len > SB3_MIN_LEN)
					retval |= 0x02;
            else if (len_ang2 > 0.0)
               continue;
            else
					Fdst = org_Fdst;

            if ((retval&0x03) != 0)
				   break;
			}
		}
	}

	if (i == 0) {
		Sdst     = org_Sdst;
		Fdst     = org_Fdst;
		curv.c.x = org_curv.c.x;
		curv.c.y = org_curv.c.y;
		curv.R   = org_curv.R;
	}
	return retval;
}

int SB3Turn::path_ok(int seg_idx, double S, double F, 
                    double min_w, double max_w) const
{
   return seg[seg_idx]->path_ok(curv, S, F, min_w, max_w);
}

void SB3Turn::move_to_outside(SB3Line &ret_norm) const
{
   if (ts_turn->rad > 0.0) {
      // move point from inside wall to outside wall
      ret_norm.movePoint(1000.0);
   }
}
#undef APEX_BEG
#undef APEX_MID
#undef APEX_END
#undef ADJ_ANG
#undef ADJ_RAD
//-----END Turn.cpp-----//

//-----START Trans.cpp-----//
const int SB3Trans::RENTRY = 0, 
          SB3Trans::LEXIT  = 1,
          SB3Trans::REXIT  = 2,
          SB3Trans::LENTRY = 3;

// CONSTRUCTORS
SB3Trans::SB3Trans()
   : SB3PathComponent(), curv()
{
   typeID    = SB3PathComponent::TRANS;
   PathDir   = 1;
   comp      = &curv;
   entry     = 1;
   init_spd  = 0.0;
   scale_spd = 1.0;
   bias_spd  = 0.0;
   start_ang = 0.0;
   avg_R     = 0.0;
}

// METHODS
double SB3Trans::get_spd(double trk_dist, double *max) const
{
   if (max) {
      if (scale_spd >= 0.0) {
         SB3Line l;
         int    i = find_trk_seg(trk_dist);

         seg[i]->get_norm(l, trk_dist);
         trk_dist = 1.0 / curv.getCurvature(curv.ang(l.pt));
         *max     = (scale_spd * SB3_crn_spd(trk_dist)) + bias_spd;
      }
      else
         *max = bias_spd;
   }
   return init_spd;
}

void SB3Trans::get_tan(SB3Vector &ret_pt, SB3Line &ret_norm) const
{
   ret_norm.pt.x = ret_norm.dir.x; ret_norm.pt.y = ret_norm.dir.y;
   curv.getTangent(ret_norm.dir, curv.ang(ret_pt));
   if (PathDir > 0) {
      ret_norm.dir.x = -ret_norm.dir.x;
      ret_norm.dir.y = -ret_norm.dir.y;
   }
}

double SB3Trans::get_path_dist(const SB3Vector &pt) const
{
   double   a;
   SB3Vector v(pt);

   v.sub(curv.c);
   a = v.ang();

   if ( (PathDir > 0)&&(a < start_ang) )
      a += MY_PI21;
   else if ( (PathDir < 0)&&(a > start_ang) )
      a -= MY_PI21;
   a -= start_ang;
   return curv.getArcLength(start_ang, a);
}

double SB3Trans::get_dist2path(const SB3Vector &pt) const
{
   if (PathDir > 0)
      return curv.dist2Point(pt);
   return -curv.dist2Point(pt);
}

void SB3Trans::getPoint2(SB3Vector &ret_pt, SB3Vector &ret_tan, 
                        double path_dist) const
{
   path_dist = curv.ang(path_dist, start_ang, PathDir);
   curv.getPoint(ret_pt, path_dist);
   curv.getTangent(ret_tan, path_dist);
   if (PathDir > 0) {
      ret_tan.x = -ret_tan.x;
      ret_tan.y = -ret_tan.y;
   }
}

double SB3Trans::brk_dist(double dist, double v) const
{
   v *= v;
   v /= avg_R * MAX_ACCEL;
   if (v > 0.65)
      dist *= 0.65 / v;
   return (0.925*dist);
}

void SB3Trans::copy(const SB3Trans &tc)
{
   SB3PathComponent::copy(tc);

   PathDir    = tc.PathDir;
   curv.c.x   = tc.curv.c.x;
   curv.c.y   = tc.curv.c.y;
   curv.Rx    = tc.curv.Rx;
   curv.Ry    = tc.curv.Ry;
   curv.r_ang = tc.curv.r_ang;
   entry      = tc.entry;
   init_spd   = tc.init_spd;
   scale_spd  = tc.scale_spd;
   bias_spd   = tc.bias_spd;
   start_ang  = tc.start_ang;
   avg_R      = tc.avg_R;
}

void SB3Trans::adjust(const SB3Circle &trn, double str_ang,
                     double len_ang, double len_y,
                     double trk_wid, int type)
{
   int      i1, i2;
   SB3Vector p;
      
   curv.c.x = -get_len_x(trn.R, len_ang, len_y);
   curv.c.y = 0.0;

   Fdst     = trn.R * sin(len_ang);
   curv.Rx  = sqrt(1.0 - (curv.c.x/Fdst));
   curv.Rx *= len_y;
   curv.Ry  = len_y;

   PathDir = (2*(type % 2)) - 1;
   entry   = (type == SB3Trans::RENTRY)||(type == SB3Trans::LENTRY);

   Fdst   -= curv.c.x;
   Sdst    = atan((trn.R*cos(len_ang))/Fdst);
   len     = curv.getArcLength(Sdst, (MY_PI12 - Sdst));
   switch (type) {
   case RENTRY:
      Fdst = Sdst;
      Sdst = MY_PI12;
      break;
   case LEXIT:
      Fdst     = MY_PI12;
      str_ang -= MY_PI;
      break;
   case REXIT:
      Sdst     = -Sdst;
      Fdst     = -MY_PI12;
      str_ang -=  MY_PI;
      break;
   case LENTRY:
      Fdst = -Sdst;
      Sdst = -MY_PI12;
   }
   curv.r_ang = str_ang;
   curv.c.rotate(str_ang);
   curv.c.add(trn.c);

   curv.getPoint(p, (Sdst+str_ang));
   i1 = find_trk_seg(p, trk_wid, 0, 1);
   if (i1 >= 0)
      Sdst = seg[i1]->get_trk_dist(p);
   else {
      i1   = 0;
      Sdst = seg[0]->Sdst;
   }
   curv.getPoint(p, (Fdst+str_ang));
   i2 = find_trk_seg(p, trk_wid, (num_segs-1), -1);
   if (i2 >= i1)
      Fdst = seg[i2]->get_trk_dist(p);
   else
      Fdst = seg[num_segs-1]->Fdst;
}

void SB3Trans::set_spd_len(double turn_spd)
{
   double   len_ang;
   SB3Vector p1, p2;
   SB3Line   n;

   getPoint(p1, n, Sdst, 0);
   getPoint(p2, n, Fdst, 0);
   p1.sub(curv.c);
   p2.sub(curv.c);
   len_ang   = p1.dot(p2) / (p1.mag()*p2.mag());
   len_ang   = fabs(acos(len_ang));
   if (PathDir < 0)
      len_ang = -len_ang;
   start_ang = p1.ang();
   len       = curv.getArcLength(start_ang, len_ang);

   if ( (pass != NULL)&&(is_after(pass[0]->Sdst, *this, Fdst)) ) {
      getPoint(p2, n, pass[0]->Sdst, 0);
      p2.sub(curv.c);
      len_ang = p1.dot(p2) / (p1.mag()*p2.mag());
      len_ang = fabs(acos(len_ang));
      if (PathDir < 0)
         len_ang = -len_ang;
   }
   p1.x   = 1.0 / curv.getCurvature(start_ang);
   p1.y   = 1.0 / curv.getCurvature(start_ang + len_ang);
   avg_R  = p1.x;
   avg_R += 1.0 / curv.getCurvature(start_ang + (0.25*len_ang));
   avg_R += 1.0 / curv.getCurvature(start_ang + (0.5 *len_ang));
   avg_R += 1.0 / curv.getCurvature(start_ang + (0.75*len_ang));
   avg_R += p1.y;
   avg_R *= 0.2;

#ifndef SB3_DLL
   switch (args.m_iSurface) {
#else
   switch (args->m_iSurface) {
#endif
   case 0:
      scale_spd = ((turn_spd < 70.0) ? 0.945 : 0.95) * SCALE_CORNSPD;
      break;
   case 1:
      scale_spd = ((turn_spd < 70.0) ? 0.99 : 1.025) * SCALE_CORNSPD;
      avg_R    *= 0.65;
      break;
   case 2:
      scale_spd = ((turn_spd < 80.0) ? 1.1 : 1.2) * SCALE_CORNSPD;
      avg_R    *= 1.5;
   };

   if (entry) {
      p2.x = SB3_crn_spd(p1.x);
      p2.y = SB3_crn_spd(p1.y);
   }
   else {
      p2.y = SB3_crn_spd(p1.x);
      if ( (turn_spd < p2.y)||(turn_spd >= SB3_MAX_SPD) ) {
         init_spd = p2.y * scale_spd;
         if ( (init_spd < turn_spd)&&(turn_spd < SB3_MAX_SPD) ) {
            bias_spd = turn_spd - init_spd;
            init_spd = turn_spd;
         }
         else
            bias_spd = 0.0;
         return;
      }
      p2.x = SB3_crn_spd(p1.y);
   }

   if (p2.x <= p2.y)
      scale_spd = 0.0;
   else {
      scale_spd = (p2.x - turn_spd) / (p2.x - p2.y);
      if (scale_spd < 0.5)
         scale_spd = 0.5;
      else if (scale_spd > 1.0)
         scale_spd = 1.0;
      scale_spd *= (entry) ? 0.93 : 1.05;
   }
   bias_spd = turn_spd - (scale_spd*p2.y);

   if (entry)
      init_spd = (scale_spd*p2.x) + bias_spd;
   else
      init_spd = turn_spd;
}

int SB3Trans::makeStraight(SB3Straight &ret, SB3PathComponent &pt,
                          int prev, double trk_wid)
{
   int    i, k, k1, k2; 
   double d1, d2;
   SB3Line s, n;

   if (pt.typeID == SB3PathComponent.STRAIGHT)
      return 0;

   for (i = 1; i <= 4; i++) {
      if (pt.typeID == SB3PathComponent.TURN) {
         if (!curv.tangent(s, (const SB3Circle&)*(pt.comp), i))
            continue;
      }
      else if (pt.typeID == SB3PathComponent.TRANS) {
         if (!curv.tangent(s, (const SB3Ellipse&)*(pt.comp), i))
            continue;
      }
      else
         return 0;
      s.dir.add(s.pt);
         
      k1 = find_trk_seg(s.pt, trk_wid, 0, 1);
      if (k1 < 0) {
         k1 = ret.find_trk_seg(s.pt, trk_wid, 0, 1);
         if (k1 < 0)
            continue;
         insert_TrkSeg(ret, (!prev));
         k1 = find_trk_seg(s.pt, trk_wid, 0, 1);
      }
      d1 = seg[k1]->get_trk_dist(s.pt);
      if ( ( ( prev)&&(!is_after(Fdst, *this, d1)) )||
           ( (!prev)&&(!is_after(d1, *this, Sdst)) ) )
         continue;
         
      k2 = pt.find_trk_seg(s.dir, trk_wid, 0, 1);
      if (k2 < 0) {
         k2 = ret.find_trk_seg(s.dir, trk_wid, 0, 1);
         if (k2 < 0)
            continue;
         pt.insert_TrkSeg(ret, prev);
         k2 = pt.find_trk_seg(s.dir, trk_wid, 0, 1);
      }
      d2 = pt.seg[k2]->get_trk_dist(s.dir);
      if ( ( ( prev)&&(!pt.is_after(d2, pt, pt.Sdst)) )||
           ( (!prev)&&(!pt.is_after(pt.Fdst, pt, d2)) ) )
         continue;

      if ( (prev)&&(is_after(d1, pt, d2)) ) {
         ret.adjust(s.dir, s.pt);

         // insert TrkSeg into Straight component, if necessary
         k = pt.find_trk_seg(ret.seg[0]->Sdst, (pt.num_segs-1), -1);
         for (; k >= k2; k--) {
            ret.insert_TrkSeg(pt.seg[k], 0);
         }
         k = find_trk_seg(ret.seg[ret.num_segs-1]->Fdst, 0, 1);
         for (; (k >= 0)&&(k <= k1); k++) {
            ret.insert_TrkSeg(seg[k], ret.num_segs);
         }
         
         // start and finish locations
         ret.Sdst = pt.Fdst = d2;
         ret.Fdst = Sdst    = d1;
      }
      else if ( (!prev)&&(pt.is_after(d2, *this, d1)) ) {
         ret.adjust(s.pt, s.dir);

         // insert TrkSeg into Straight component, if necessary
         k = find_trk_seg(ret.seg[0]->Sdst, (num_segs-1), -1);
         for (; k >= k1; k--) {
            ret.insert_TrkSeg(seg[k], 0);
         }
         k = pt.find_trk_seg(ret.seg[ret.num_segs-1]->Fdst, 0, 1);
         for (; (k >= 0)&&(k <= k2); k++) {
            ret.insert_TrkSeg(pt.seg[k], ret.num_segs);
         }
            
         // start and finish locations
         ret.Sdst = Fdst    = d1;
         ret.Fdst = pt.Sdst = d2;
      }
      else
         continue;
      if (ret.valid(trk_wid)) {
         ret.trim();
         return 1;
      }
   }

   if (prev) {
      ret.Sdst = pt.Fdst;
      ret.Fdst = Sdst;
      pt.getPoint(s.pt, n, ret.Sdst, 0);
      getPoint(s.dir, n, ret.Fdst, 0);
   }
   else {
      ret.Sdst = Fdst;
      ret.Fdst = pt.Sdst;
      getPoint(s.pt, n, ret.Sdst, 0);
      pt.getPoint(s.dir, n, ret.Fdst, 0);
   }
   ret.adjust(s.pt, s.dir);
   return 0;
}

double SB3Trans::rating() const
{
   if (curv.Rx >= curv.Ry)
      return (curv.Rx / curv.Ry);
   return (curv.Ry / curv.Rx);
}

double SB3Trans::get_len_x(double circleR, double len_ang, 
                          double len_y)
{
   len_ang  = circleR * sin(len_ang);
   len_y   *= len_y;
   circleR *= circleR;
   return ((len_y - circleR) / len_ang);
}

double SB3Trans::get_len_y(double circleR, double len_ang,
                          double len_x)
{
   len_ang = len_x * sin(len_ang);
   return (sqrt((len_ang + circleR)*circleR));
}

int SB3Trans::path_ok(int seg_idx, double S, double F, 
                     double min_w, double max_w) const
{
   return seg[seg_idx]->path_ok(curv, S, F, min_w, max_w);
}

void SB3Trans::move_to_outside(SB3Line &ret_norm) const
{
   if (PathDir > 0) {
      // move point from inside wall to outside wall
      ret_norm.movePoint(1000.0);
   }
}
//-----END Trans.cpp-----//

//-----START TrkData.cpp-----//
const int SB3TrkData::NORMAL = 0,
          SB3TrkData::OFF    = 1,
          SB3TrkData::START  = 2;

// CONSTRUCTORS
SB3TrkData::SB3TrkData()
{
   name = NULL;
   path = NULL;
   seg  = NULL;

   num_segs = num_comps = 0;
   trk_len  = trk_wid   = 0.0;
   pit_spd  = pit_entry = 0.0;
}

// DESTRUCTOR
SB3TrkData::~SB3TrkData()
{
   int i;

   if (path) {
      for (i = 0; i < num_comps; i++) {
         delete path[i];
      }
      delete [] path;
      path = NULL;
   }
   
   if (seg) {
      for (i = 0; i < num_segs; i++) {
         delete seg[i];
      }
      delete [] seg;
      seg = NULL;
   }

   if (name) {
      free(name);
      name = NULL;
   }
}

// METHODS
int SB3TrkData::can_pass(double trk_dist, int pc) const 
{
   return ( (path[pc]->pass != NULL)&&(get_PassComponent(trk_dist, pc) < 1) );
}

double SB3TrkData::get_trk_dist(int seg_ID, double to_end) const
{
   double d;

   to_end /= seg[seg_ID]->len;
   
   d = ((to_end*seg[seg_ID]->Sdst) + ((1.0-to_end)*seg[seg_ID]->Fdst));
   if (d > trk_len)
      d -= trk_len;
   else if (d < 0.0)
      d += trk_len;
   return d;
}

double SB3TrkData::get_trk_dist(const SB3Vector &pt, double to_lft,
                               int start_seg_ID) const
{
   int i, j;

   for (j = 0; j < num_segs; j++) {
      i = (start_seg_ID + j) % num_segs;

      if (fabs(seg[i]->to_left(pt) - to_lft) < 5.0)
         return seg[i]->get_trk_dist(pt);
   }
   return -trk_len;
}

void SB3TrkData::get_trk_pos_vel(SB3Line &ret_pv, double to_lft, 
                                double v, double vn, 
                                double trk_dist, int seg_ID) const
{
   double ang;

   seg[seg_ID]->get_norm(ret_pv, trk_dist);

   // current position
   ret_pv.movePoint(to_lft);

   // current velocity
   ang          = ret_pv.dir.ang();
   ret_pv.dir.x = -vn;
   ret_pv.dir.y = sqrt((v*v) - (vn*vn));
   ret_pv.dir.rotate(ang);
}

int SB3TrkData::get_PathComponent(double trk_dist, int pass_flag,
                                 int start) const
{
   int              i, j, k;
   double           Fdst;
   SB3PathComponent *pc;

   if (trk_dist > trk_len)
      trk_dist -= trk_len;
   else if (trk_dist < 0.0)
      trk_dist += trk_len;

   if (start < 0)
      start += num_comps;

   for (j = 0; j < num_comps; j++) {
      i  = (start + j) % num_comps;
      pc = path[i];

      if ( (pass_flag)&&(pc->pass) )
         Fdst = pc->pass[0]->Sdst;
      else
         Fdst = pc->Fdst;
      if ( (pc->is_after(trk_dist, *pc, pc->Sdst))&&
           (pc->is_after(Fdst, *pc, trk_dist)) )
         return i;

      if ( (pass_flag)&&(pc->pass) ) {
         for (k = 0; k < 2; k++) {
            pc = path[i]->pass[k];
            if ( (pc->is_after(trk_dist, *pc, pc->Sdst))&&
                 (pc->is_after(pc->Fdst, *pc, trk_dist)) )
               return i;
         }
      }
   }
   return -1;
}

double SB3TrkData::get_tgt_spd(double v, double drag_accel, 
                              double path_dist, double trk_dist,
                              int pass_flag, int pc) const
{
   int    i, j;
   double nex_spd, tst_spd, brk, brk_dist, brk_end;
   double dist2comp, end_comp, max_dist, min_dist;

   if (pc < 0) {
      pc = get_PathComponent(trk_dist, pass_flag);
      if (pc < 0)
         return 0.0;
   }

   i         = pc;
   max_dist  = SB3_brkdist(v, 0.0, drag_accel);
   min_dist  = v * delta_time;
   dist2comp = brk_dist = 0.0;
   end_comp  = -path_dist;
   if ( (pass_flag)&&(path[i]->pass) )
      j = get_PassComponent(trk_dist, i);
   else
      j = -1;

   if (j >= 0) {  
      // on the passing path
      path[i]->pass[j]->get_spd(trk_dist, &nex_spd);
      end_comp += path[i]->pre_len + path[i]->pass[0]->len;
      if (j == 1) {
         end_comp += path[i]->pass[j]->len;
         dist2comp = -2.0 * v * delta_time;
         brk_dist  = dist2comp;
      }
      brk_end = end_comp;
   }
   else {
      // on the racing path
      path[i]->get_spd(trk_dist, &nex_spd);
      end_comp += path[i]->len;
      brk_end   = path[i]->brk_dist(end_comp, v);
   }
   j++;

   do {
      dist2comp += end_comp;
      brk_dist  += brk_end;
      if ( (pass_flag)&&(path[i]->pass)&&(j >= 0)&&(j <= 2) ) {
         if (j < 2) {
            // check speed on passing path
            if (!j)
               dist2comp += path[i]->pre_len - path[i]->len;
            end_comp = path[i]->pass[j]->len;
            brk_end  = path[i]->pass[j]->brk_dist(end_comp, v);
            tst_spd  = path[i]->pass[j]->get_spd(path[i]->pass[j]->Sdst);
            j++;
         }
         else {
            // check speed on component after passing path
            j        = i;
            i        = get_NextComponent(i);
            end_comp = path[i]->post_len;
            brk_end  = path[i]->brk_dist(end_comp, v);
            path[i]->get_spd(path[j]->pass[1]->Fdst, &tst_spd);
            j = 0;
         }
      }
      else {
         // check speed on next component
         i        = (i + 1) % num_comps;
         end_comp = path[i]->len;
         brk_end  = path[i]->brk_dist(end_comp, v);
         tst_spd  = path[i]->get_spd(path[i]->Sdst);
      }

      brk = SB3_brkdist(v, tst_spd, drag_accel);
      if (brk_dist > min_dist)
         brk *= dist2comp / brk_dist;
      if ( (brk >= dist2comp)&&(nex_spd > tst_spd) )
         nex_spd = tst_spd;
      else if ( (dist2comp <= min_dist)&&(tst_spd > nex_spd) )
         nex_spd = tst_spd;
   } while (dist2comp < max_dist);
   return nex_spd;
}

int SB3TrkData::get_nex_pos(SB3Line &cV, SB3Vector &nV, int type,
                           double trk_dist, int pass_flag, int pc) const
{
   int      i, retval;
   double   v, step, MAX_dA, to_side;
   SB3Vector t1, t2;

   if (pc < 0) {
      pc = get_PathComponent(trk_dist, pass_flag);
      if (pc < 0)
         return 0;
   }

   to_side = nV.y;
   v       = nV.x;
   t2.x    = cV.dir.mag();
   if (t2.x <= 35.0)
      step = 20.0 / (35.0 * delta_time);
   else if (t2.x < 73.0)
      step = 20.0 / (t2.x * delta_time);
   else
      step =  5.0;

   if ( (pass_flag)&&(path[pc]->pass) ) {
      i = get_PassComponent(trk_dist, pc);
      if (i < 0) {
         // check if car has entered the passing path
         t1.x   = path[pc]->Fdst;
         t1.y   = path[pc]->pass[0]->Sdst;
         retval = ( (path[pc]->is_after(trk_dist, *(path[pc]), t1.x))&&
                    (path[pc]->is_after(t1.y, *(path[pc]), trk_dist)) );
      }
      else if (i == 1) {
         // check if car has left the passing path
         t1.x  = path[pc]->pass[1]->dst_between(path[pc]->pass[1]->Fdst,
                                                *(path[pc]->pass[1]),
                                                trk_dist);
         retval = (t1.x > (v*delta_time));
      }
      else
         retval = 1;
   }
   else {
      i      = -1;
      retval = 0;
   }

   // get current point, and tangent vector
   if ( (i >= 0)&&(i < 2) ) {
      trk_dist = path[pc]->pass[i]->get_path_dist(cV.pt);
      MAX_dA   = path[pc]->pass[i]->comp->dist2Point(cV.pt);
      path[pc]->pass[i]->getPoint2(nV, t1, trk_dist);
      if (path[pc]->pass[i]->PathDir < 0)
         to_side = trk_wid - to_side;
      else if (path[pc]->pass[i]->typeID == SB3PathComponent::STRAIGHT)
         to_side = 0.2 * trk_wid;
   }
   else {
      trk_dist = path[pc]->get_path_dist(cV.pt);
      MAX_dA   = path[pc]->comp->dist2Point(cV.pt);
      path[pc]->getPoint2(nV, t1, trk_dist);
      if (path[pc]->PathDir < 0)
         to_side = trk_wid - to_side;
      else if (path[pc]->typeID == SB3PathComponent::STRAIGHT)
         to_side = 0.2 * trk_wid;
      if (path[pc]->typeID == SB3PathComponent::TRANS)
         step *= 0.9;
   }

   t2.y = fabs(MAX_dA) / (2.0*CARWID);
   if (t2.y >= 1.0)
      step *= 1.2*t2.y;

   // find new point
   trk_dist += step * t2.x * delta_time;
   do {
      if ( (pass_flag)&&(path[pc]->pass) ) {
         if (i < 0) {
            if (trk_dist <= path[pc]->pre_len)
               path[pc]->getPoint2(nV, t2, trk_dist);
            trk_dist -= path[pc]->pre_len;
            i         = 0;
         }
         else if (i == 0) {
            if (trk_dist <= path[pc]->pass[0]->len)
               path[pc]->pass[0]->getPoint2(nV, t2, trk_dist);
            trk_dist -= path[pc]->pass[0]->len;
            i         = 1;
         }
         else {
            if (trk_dist <= path[pc]->pass[1]->len)
               path[pc]->pass[1]->getPoint2(nV, t2, trk_dist);
            trk_dist -= path[pc]->pass[1]->len;
            if (trk_dist > 0.0) {
               // next component
               pc        = get_NextComponent(pc);
               trk_dist += path[pc]->len - path[pc]->post_len;
               i         = -1;
            }
         }
      }
      else {
         if (trk_dist <= path[pc]->len)
            path[pc]->getPoint2(nV, t2, trk_dist);
         trk_dist -= path[pc]->len;
         pc = (pc + 1) % num_comps;
      }
   } while (trk_dist > 0.0);

   // new velocity vector
   nV.sub(cV.pt);
   nV.normalize();
   nV.scale(v);
   
   // limit the change in direction
   if (type != SB3TrkData::OFF) {
      MAX_dA = fabs(MAX_dA) / trk_wid;
      if (MAX_dA < 0.1)
         MAX_dA = 0.1;
      else if (MAX_dA > 0.4)
         MAX_dA = 0.4;
      if (type == SB3TrkData::START)
         MAX_dA *= 0.6;
   }
   else
      MAX_dA = 0.4;

   v = cV.dir.mag();
#ifndef SB3_DLL
   switch (args.m_iSurface) {
#else
   switch (args->m_iSurface) {
#endif
   case 0:
      if (v > 142.0)
         v  = 1.0;
      else
         v *= 0.007;
      break;
   case 1:
      if (v > 166.0)
         v  = 1.0;
      else
         v *= 0.006;
      break;
   case 2:
      v *= 0.0025;
   };

   if (to_side <= 0.0)
      to_side = 0.0;
   else if (to_side >= trk_wid)
      to_side = v;
   else
      to_side = v * (to_side / trk_wid);
   t1.rotate(to_side * t1.ang(t2));
   v    = t1.ang(nV);
   step = v / step;
   if (step > MAX_dA)
      nV.rotate(MAX_dA - v);
   else if (step < -MAX_dA)
      nV.rotate(-MAX_dA - v);
   else if (type != SB3TrkData::OFF) {
      MAX_dA = 1.0 - MAX_dA;
      nV.rotate((step - v)*MAX_dA);
   }
   return retval;
}

void SB3TrkData::locate_car(double *dist2path, double *path_dist, 
                           const SB3Vector &pt, int start_pc, int end_pc,
                           double trk_dist, int pass_flag) const
{
   int i, j;

   if ( (pass_flag)&&(path[end_pc]->pass) )
      j = get_PassComponent(trk_dist, end_pc);
   else
      j = -1;

   if (path_dist) {
      *path_dist = 0.0;
      for (i = start_pc; i != end_pc; i = (i+1)%num_comps) {
         *path_dist += path[i]->len;
      }
   }
   else
      i = end_pc;

   if ( (j >= 0)&&(j < 2) ) {
      if (path_dist) {
         *path_dist += path[i]->pre_len;
         if (j)
            *path_dist += path[i]->pass[0]->len;
         *path_dist += path[i]->pass[j]->get_path_dist(pt);
      }
      if (dist2path)
         *dist2path  = path[i]->pass[j]->get_dist2path(pt);
   }
   else {
      if (path_dist)
         *path_dist += path[i]->get_path_dist(pt);
      if (dist2path)
         *dist2path  = path[i]->get_dist2path(pt);
   }
}

void SB3TrkData::set_path(Track *trk)
{
   int          i, num_turns, data_loaded;
   SB3Turn     **crn = NULL;
   SB3Straight **str = NULL;

   // remove old data
   this->~SB3TrkData();

   // get RARS track data
   num_turns = get_trk_data(trk);

   // Make some corners
   data_loaded = init_Turns(&crn, &num_turns);

   // Make some straights
   str = new SB3Straight*[num_turns];
   for (i = 0; i < num_turns; i++) {
      if (crn[i]->Sdst >= trk_len)
         crn[i]->Sdst -= trk_len;
      if (crn[i]->Sdst < 0.0)
         crn[i]->Sdst += trk_len;
      if (crn[i]->Fdst >= trk_len)
         crn[i]->Fdst -= trk_len;
      if (crn[i]->Fdst < 0.0)
         crn[i]->Fdst += trk_len;
      str[i] = new SB3Straight();  // initialize the straights
   }

   // Try and make everything fit
   adjust_Turns(crn, str, num_turns, 
                ((data_loaded)?(SB3_NUM_RNDS-1):0), 0, NULL);

   // Combine turns and straights into the path
   insert_Straights(crn, str, num_turns, data_loaded);

   // Create the passing path
   makePassingPath(crn, str, num_turns);

   // Save path now!
//   saveData(crn, num_turns);

   // Try and improve the path
   addTransitions(crn, str, num_turns, data_loaded);

   // free memory
   delete [] crn;
   delete [] str;
}


// PRIVATE METHODS
int SB3TrkData::get_trk_data(Track *trk)
{
   int i, num_turns;

   // store track name
   i       = strlen(trk->m_sFileName);
   name    = (char *)malloc(i + 1);
   name[i] = 0;
   for (; i >= 0; i--) {
      name[i] = (char)tolower((int)trk->m_sFileName[i]);
   }

   num_turns = 0;
   trk_len   = 0.0;
   trk_wid   = trk->width;
   num_segs  = trk->NSEG;
   pit_spd   = trk->m_fPitLaneSpeed + 75.0;
   pit_entry = trk->m_fPitEntry;
   seg       = new SB3TrkSeg*[num_segs];
   for (i = 0; i < num_segs; i++) {
      if (trk->lftwall[i].radius == 0.0)   // straight
         seg[i] = new SB3TrkSeg_Straight(i, (trk->lftwall + i), trk_len);
      else {                               // turn
         seg[i] = new SB3TrkSeg_Turn(i, (trk->lftwall + i), trk_wid, trk_len);
         num_turns++;
      }
      trk_len += seg[i]->Fdst - seg[i]->Sdst;
   }
   return num_turns;
}

int SB3TrkData::init_Turns(SB3Turn ***p_crn, int *p_num_turns)
{
   int      i, j;
   double   min_w, max_w;
   SB3Turn **crn;

   // check if there is data that can be loaded
   if (loadData(p_crn, p_num_turns))
      return 1;

   min_w = SB3_IN_WID;
   max_w = trk_wid - SB3_OUT_WID;
   crn   = *p_crn = new SB3Turn*[*p_num_turns];
   for (i = 0, j = 0; i < num_segs; i++) {
      if (seg[i]->rad == 0.0)   // straight
         continue;
      if (j == *p_num_turns)
         break;

      crn[j++] = new SB3Turn(seg, i, num_segs, min_w, max_w, trk_wid);
   }
   return 0;
}

void SB3TrkData::adjust_Turns(SB3Turn **crn, SB3Straight **str, 
                             int num_turns, int rnds, int level, 
                             int *adj_val)
{
   int        i, j, k, del, chk, stop;
   SB3Turn     ct, ot;
   SB3Straight ps;
   double     min_w = SB3_IN_WID, 
              max_w = trk_wid - SB3_OUT_WID;

   if (adj_val == NULL) {
      adj_val = new int[num_turns];
      for (i = 0; i < num_turns; i++) {
         adj_val[i] = SB3Turn::BAD_ENTRY|SB3Turn::BAD_EXIT;
      }
   }

   stop = ((num_turns-1) > SB3_SEARCH_LEN) ? SB3_SEARCH_LEN : (num_turns-1);
   for (; rnds < SB3_NUM_RNDS; rnds++) {
      // Find problems in current path
      chk = 0;
      for (j = 0; j < num_turns; j++) {
         if ( (crn[j] == NULL)||(adj_val[j]&SB3Turn::IGNORE_TURN) )
            continue;

         if (!crn[j]->valid(trk_wid))
            adj_val[j] |= SB3Turn::BAD_TURN;
         else {
            // Check entry to turn
            ct.copy(*(crn[j]), 1, 1, (num_segs-crn[j]->num_segs));
            adj_val[j] |= SB3Turn::BAD_ENTRY;
            for (k = 1, del = 0; (k < stop)||(k == 1); k++) {
               if (k > 1) {
                  // insert TrkSegs from the last turn into this one
                  while (crn[i] == NULL) {
                     i = (i + 1) % num_turns;
                  }
                  if (i != j)
                     ct.insert_TrkSeg(*(crn[i]), 0);
               }
               i = (j + num_turns - k) % num_turns;
               if ( (crn[i] == NULL)||(adj_val[i]&SB3Turn::IGNORE_TURN) )
                  continue;
               ot.copy(*(crn[i]), 1, 0, (num_segs-crn[i]->num_segs));
               if (k > 1)
                  ot.insert_TrkSeg(ct, 1);

               if ( (ct.makeStraight(ps, ot, trk_wid))&&
                    (crn[j]->is_after(crn[j]->Fdst, ct, ct.Sdst))&&
                    (ct.valid(trk_wid))&&(ot.valid(trk_wid))&&
                    (ps.valid(trk_wid)) ) {
                  adj_val[j] &= ~SB3Turn::BAD_ENTRY;
                  str[j]->copy(ps);
                  if ( (k == 1)||
                       (!(adj_val[i]&(SB3Turn::BAD_TURN|SB3Turn::BAD_ENTRY))) ) {
                     // this turn is good
                     del = k - 1;
                     crn[i]->copy(ot, 0, 0, 0);
                     crn[j]->copy(ct, 0, 1, 0);
                     if (k > 1) {
                        crn[i]->insert_TrkSeg(ot, 1);
                        crn[j]->insert_TrkSeg(ct, 0);
                     }
                  }
               }
               if (adj_val[j]&(SB3Turn::BAD_TURN|SB3Turn::BAD_EXIT))
                  break;  // do not try and remove any turns
            }
            // Remove unecessary turns
            for (; del > 0; del--) {
               // remove these
               i = (j + num_turns - del) % num_turns;
               if (crn[i] != NULL) {
                  delete crn[i];
                  crn[i] = NULL;
               }
               if (str[i] != NULL) {
                  delete str[i];
                  str[i] = NULL;
               }
            }

            // Check exit from turn
            ct.copy(*(crn[j]), 1, 0, (num_segs-crn[j]->num_segs));
            adj_val[j] |= SB3Turn::BAD_EXIT;
            for (i = 1, del = 0; (i < stop)||(i == 1); i++) {
               if (i > 1) {
                  // insert TrkSegs from the last turn into this one
                  while (crn[k] == NULL) {
                     k = (k + num_turns - 1) % num_turns;
                  }
                  if (k != j)
                     ct.insert_TrkSeg(*(crn[k]), 1);
               }
               k = (j + i) % num_turns;
               if ( (crn[k] == NULL)||(adj_val[k]&SB3Turn::IGNORE_TURN) )
                  continue;
               ot.copy(*(crn[k]), 1, 1, (num_segs-crn[k]->num_segs));
               if (i > 1)
                  ot.insert_TrkSeg(ct, 0);

               if ( (ot.makeStraight(ps, ct, trk_wid))&&
                    (ct.is_after(ct.Fdst, *(crn[j]), crn[j]->Sdst))&&
                    (ct.valid(trk_wid))&&(ot.valid(trk_wid))&&
                    (ps.valid(trk_wid)) ) {
                  adj_val[j] &= ~SB3Turn::BAD_EXIT;
                  str[k]->copy(ps);
                  if ( (i == 1)||
                     (!(adj_val[k]&(SB3Turn::BAD_TURN|SB3Turn::BAD_EXIT))) ) {
                     // this turn is good
                     del = i - 1;
                     crn[j]->copy(ct, 0, 0, 0);
                     crn[k]->copy(ot, 0, 1, 0);
                     if (i > 1) {
                        crn[j]->insert_TrkSeg(ct, 1);
                        crn[k]->insert_TrkSeg(ot, 0);
                     }
                  }
               }
               if (adj_val[j]&(SB3Turn::BAD_TURN|SB3Turn::BAD_ENTRY))
                  break;  // do not try and remove any turns
            }
            // Remove unecessary turns
            for (; del > 0; del--) {
               // remove this
               k = (j + del) % num_turns;
               if (crn[k] != NULL) {
                  delete crn[k];
                  crn[k] = NULL;
               }
               if (str[k] != NULL) {
                  delete str[k];
                  str[k] = NULL;
               }
            }
            
            adj_val[j] &= ~SB3Turn::BAD_TURN;  // remove this flag
         }     
         chk += adj_val[j];
      }
      if ( (chk == 0)||((rnds+1) == SB3_NUM_RNDS) )
         break;

      // Fix problems in the path
      chk = 0;
      for (j = 0; j < num_turns; j++) {
         if ( (crn[j] == NULL)||(adj_val[j] == 0)||
              (adj_val[j]&SB3Turn::IGNORE_TURN) )
            continue;

         i = j;  // prev turn index
         do {
            i = (i + num_turns - 1) % num_turns;
         } while ( (crn[i] == NULL)||(adj_val[i]&SB3Turn::IGNORE_TURN) );
         k = j;  // next turn index
         do {
            k = (k + 1) % num_turns;
         } while ( (crn[k] == NULL)||(adj_val[k]&SB3Turn::IGNORE_TURN) );

         if ( (!(adj_val[i]&SB3Turn::BAD_TURN))&&
              (!(adj_val[k]&SB3Turn::BAD_TURN)) ) {
            ct.copy(*(crn[i]), 1);
            ct.insert_TrkSeg(*(crn[j]), 1);
            ot.copy(*(crn[k]), 1);
            ot.insert_TrkSeg(*(crn[j]), 0);
            if (ct.is_after(ct.Fdst, ot, ot.Sdst)) {
               adj_val[j] |= SB3Turn::IGNORE_TURN;
               continue;  // this turn is no longer needed
            }
         }
         if ( (crn[(j+num_turns-1) % num_turns] == NULL)||
              (crn[(j+1) % num_turns] == NULL) )
            continue;  // stop adjusting this turn

         del = adj_val[j];  // save value
         if ( (adj_val[j]&SB3Turn::BAD_ENTRY)&&
              (adj_val[i]&SB3Turn::BAD_TURN) )
            adj_val[j] &= ~SB3Turn::BAD_ENTRY;
         if ( (adj_val[j]&SB3Turn::BAD_EXIT)&&
              (adj_val[k]&SB3Turn::BAD_TURN) )
            adj_val[j] &= ~SB3Turn::BAD_EXIT;

         // Adjust turn
         crn[j]->adjust(*(crn[i]), *(crn[k]), adj_val[j], 
                        min_w, max_w, trk_wid);
         adj_val[j] = del;  // restore value

         chk++;
      }
      if (chk == 0)
         break;
   }

   if ( (chk != 0)&&(level < 4)&&(rnds > 0) ) {
      level++;
      stop    = 1;
      ct.len  = (double)level / 9.0;
      ct.Sdst = (min_w * (1.0-ct.len)) + (max_w * ct.len);
      ct.Fdst = (max_w * (1.0-ct.len)) + (min_w * ct.len);

      for (j = 0; j < num_turns; j++) {
         if ( (crn[j] == NULL)||(adj_val[j] == 0)||
              (adj_val[j] == SB3Turn::BAD_ENTRY)||
              (adj_val[j] == SB3Turn::BAD_EXIT) )
            continue;

         // a problem in the path could not be fixed, try again
         stop       = 0;
         adj_val[j] = SB3Turn::BAD_ENTRY|SB3Turn::BAD_EXIT;
         crn[j]->init(ct.Sdst, ct.Fdst, trk_wid);
      }
   
      if (!stop)
         adjust_Turns(crn, str, num_turns, 0, level, adj_val);
      level--;
   }
   if (level == 0)
      delete [] adj_val;   // free memory
}

void SB3TrkData::insert_Straights(SB3Turn **crn, SB3Straight **str, 
                                 int num_turns, int data_loaded)
{
   int    i, j, k;
   double d, Sdst, Fdst;

   // Check for invalid straights
   for (j = 0; j < num_turns; j++) {
      if (str[j] == NULL)
         continue;

      if ( (str[j]->Sdst > 0.0)||(str[j]->Fdst > 0.0) ) {
         k = j;  // find the previous turn
         do {
            k = (k + num_turns - 1) % num_turns;
         } while (crn[k] == NULL);
         crn[j]->Sdst = str[j]->Fdst;
         crn[k]->Fdst = str[j]->Sdst;
      }

      if (str[j]->len < SB3_MIN_LEN) {
         delete str[j];
         str[j] = NULL;   // remove
      }
   }
      
   // Now put it all together
   num_comps = 0;
   for (j = 0; j < num_turns; j++) {
      if (crn[j] != NULL)
         num_comps++;
      if (str[j] != NULL)
         num_comps++;
   }
   for (i = 0, j = 0; i < num_comps; i++, j++) {
      // find a valid turn
      while (crn[j] == NULL) {
         j++;
      }
         
      if (str[j] != NULL) {
         Sdst = str[j]->Sdst;
         str[j]->trim();
         str[j]->len_trk = str[j]->dst_between(str[j]->Fdst, 
                                               *(str[j]), 
                                               str[j]->Sdst);
         i++;
      }
      else {
         k = j;  // find the previous turn
         do {
            k = (k + num_turns - 1) % num_turns;
         } while (crn[k] == NULL);

         // connect these 2 turns
         crn[j]->join(*(crn[k]), trk_len);
         Sdst = crn[j]->Sdst;
      }
      crn[j]->set_spd_len(data_loaded);

      k = j;  // find the next turn
      do {
         k = (k + 1) % num_turns;
      } while (crn[k] == NULL);
      d               = Sdst;
      Sdst            = crn[j]->Sdst;   // store real values
      Fdst            = crn[j]->Fdst;
      crn[j]->Sdst    = d;              // extend PathComponent
      crn[j]->Fdst    = crn[k]->Sdst;
      crn[j]->trim();
      crn[j]->Sdst    = Sdst;           // restore real values
      crn[j]->Fdst    = Fdst;
      crn[j]->len_trk = crn[j]->dst_between(crn[j]->Fdst, 
                                            *(crn[j]), 
                                            crn[j]->Sdst);
   }
}

void SB3TrkData::makePassingPath(SB3Turn **crn, SB3Straight **str, 
                                int num_turns)
{
   int         i, j, k, _can_pass;
   double      max_w;
   SB3Turn     *p_t;
   SB3Straight *p_s;

   if (num_comps <= 0)
      return;

   // find last turn
   i = num_turns;
   do {
      i--;
   } while (crn[i] == NULL);

   // create passing path
   max_w = 2.0 * SB3_OUT_WID;
   p_s   = NULL;
   p_t   = NULL;
   for (j = 0; j < num_turns; j++) {
      if (crn[j] == NULL)
         continue;

      // find next turn
      k = j;
      do {
         k = (k + 1) % num_turns;
      } while (crn[k] == NULL);

      if ( (p_s == NULL)&&(p_t == NULL) ) {
         p_s = new SB3Straight();
         p_t = new SB3Turn();
      }
      if ( (p_s == NULL)||(p_t == NULL) )
         break;

      _can_pass = crn[j]->makePass(p_s, p_t, *(crn[k]), *(crn[i]), 
                                   SB3_IN_WID, max_w, trk_wid);
      i = j;
      if (_can_pass) {
         // get new components next time through loop
         p_s = NULL;
         p_t = NULL;
      }
   }

   // remove unused components
   if (p_s != NULL)
      delete p_s;
   if (p_t != NULL)
      delete p_t;
}

void SB3TrkData::addTransitions(SB3Turn **crn, SB3Straight **str, 
                               int num_turns, int data_loaded)
{
   int              i, j, k, rc, num_trans, t, array_size;
   double           d;
   SB3Trans        **trns;
   SB3PathComponent *pc;

   // count the turns in the path
   for (j = 0, array_size = 0; j < num_turns; j++) {
      if (crn[j] != NULL)
         array_size++;
   }
      
   // create an array to hold the transition components
   array_size *= 2;
   trns        = new SB3Trans*[array_size];
   for (j = 0; j < array_size; j++) {
      trns[j] = NULL;
   }
   array_size -= 2;
      
   // find last turn
   i = 0;
   do {
      i = (i + num_turns - 1) % num_turns;
   } while (crn[i] == NULL);

   // create the transition components for the path
   for (j = 0, t = 0, num_trans = 0; j < num_turns; j++) {
      // find a valid turn
      if (crn[j] == NULL)
         continue;
         
      // find next turn
      k = j;
      do {
         k = (k + 1) % num_turns;
      } while (crn[k] == NULL);

      // create the components
      if (trns[t] == NULL)
         trns[t] = new SB3Trans();
      if (trns[t+1] == NULL)
         trns[t+1] = new SB3Trans();
         
      if (i > j) {
         rc = crn[j]->makeTransitions(crn[i], str[j], *(trns[t]),
                                      *(trns[t+1]), str[k], crn[k], trk_wid);
      }
      else if (k > j) {
         rc = crn[j]->makeTransitions(NULL, str[j], *(trns[t]),
                                      *(trns[t+1]), str[k], crn[k], trk_wid);
      }
      else {
         rc = crn[j]->makeTransitions(NULL, str[j], *(trns[t]),
                                      *(trns[t+1]), str[k], NULL, trk_wid);
      }

      // Re-adjust the straight and passing components
      if ((rc & 0x01) != 0) {
         if ( (t > 0)&&(trns[t-1] != NULL) )
            pc = trns[t-1];
         else
            pc = crn[i];
         if (str[j] != NULL)
            trns[t]->makeStraight(*(str[j]), *pc, 1, trk_wid);

         if ( (trns[t]->pass != NULL)&&
              (crn[j]->is_after(crn[j]->Sdst, *(crn[j]), trns[t]->pass[0]->Sdst)) ) {
            d = trns[t]->Fdst;
            trns[t]->makeStraight((SB3Straight&)*(trns[t]->pass[0]),
                                  *(trns[t]->pass[1]), 0, trk_wid);
            trns[t]->Fdst = d;
            crn[j]->pass  = NULL;
            if (rc == 0x03)
               trns[t+1]->pass = NULL;
         }
         else
            trns[t]->pass = NULL;
         num_trans++;
      }
      if ((rc & 0x02) != 0) {
         if ( (k < j)&&(trns[0] != NULL) )
            pc = trns[0];
         else
            pc = crn[k];
         if (str[k] != NULL)
            trns[t+1]->makeStraight(*(str[k]), *pc, 0, trk_wid);

         if ( (trns[t+1]->pass != NULL)&&
              (crn[j]->is_after(trns[t+1]->pass[0]->Sdst, *(crn[j]), crn[j]->Fdst)) ) {
            d = trns[t+1]->Fdst;
            trns[t+1]->makeStraight((SB3Straight&)*(trns[t+1]->pass[0]),
                                    *(trns[t+1]->pass[1]), 0, trk_wid);
            trns[t+1]->Fdst = d;
            crn[j]->pass    = NULL;
            if (rc == 0x03)
               trns[t]->pass = NULL;
         }
         else
            trns[t+1]->pass = NULL;
         num_trans++;
      }
      i = j;

      // deal with unused components
      if (t < array_size) {
         if ((rc & 0x01) == 0) {
            trns[t+2] = trns[t];
            trns[t]   = NULL;
         }
         if ((rc & 0x02) == 0) {
            trns[t+3] = trns[t+1];
            trns[t+1] = NULL;
         }
         t += 2;
      }
      else {
         if ((rc & 0x01) == 0) {
            trns[t]->pass = NULL;
            delete trns[t];
            trns[t]       = NULL;    // remove this
         }
         if ((rc & 0x02) == 0) {
            trns[t+1]->pass = NULL;
            delete trns[t+1];
            trns[t+1]       = NULL;  // remove this
         }
      }
   }

   // Check for invalid straights
   for (j = 0; j < num_turns; j++) {
      if (str[j] == NULL)
         continue;

      if ( (str[j]->len < SB3_MIN_LEN)||
           (!str[j]->is_after(str[j]->Fdst, *(str[j]), str[j]->Sdst)) ) {
         delete str[j];
         str[j] = NULL;   // remove
         num_comps--;
      }
   }

   // find last component
   if (trns[array_size+1] != NULL)
      pc = trns[array_size+1];
   else {
      i = num_turns;
      do {
         i--;
      } while (crn[i] == NULL);
      pc = crn[i];
   }

   // Now create the new and improved path
   num_comps += num_trans;
   path       = new SB3PathComponent*[num_comps];
   for (i = 0, j = 0, t = 0; j < num_turns; j++) {
      // find a valid turn
      if (crn[j] == NULL)
         continue;
         
      if (str[j] != NULL) {
         str[j]->pre_len = str[j]->post_len = 0.0;
         str[j]->join(*pc, trk_len);
         path[i++] = pc = str[j];  // put this straight into the path
      }
      if (trns[t] != NULL) {
         trns[t]->join(*pc, trk_len);
         d = 0.98 * crn[j]->get_spd(0.0);
         if (data_loaded)
            crn[j]->set_spd_len(1, 0);
         else
            crn[j]->set_spd_len(1, ((str[j] != NULL) ? 2 : 3));
         if (crn[j]->get_spd(0.0) < SB3_MAX_SPD)
            d = crn[j]->get_spd(0.0);
         trns[t]->set_spd_len(d);
         if (trns[t]->pass == NULL)
            trns[t]->trim();
         trns[t]->pre_len = trns[t]->post_len = 0.0;
         path[i++] = pc = trns[t]; // put this transition into the path
      }
      t++;
      crn[j]->join(*pc, trk_len);
      if ( (trns[t-1] == NULL)&&(trns[t] == NULL) )
         crn[j]->set_spd_len(1, 0);
      crn[j]->pre_len = crn[j]->post_len = 0.0;
      path[i++] = pc = crn[j];     // put this turn into the path
      if (trns[t] != NULL) {
         trns[t]->join(*pc, trk_len);
         if (trns[t-1] == NULL)
            crn[j]->set_spd_len(1, (data_loaded ? 0 : 2));
         crn[j]->get_spd(crn[j]->Fdst, &d);
         trns[t]->set_spd_len(d);
         if (trns[t]->pass == NULL)
            trns[t]->trim();
         trns[t]->pre_len = trns[t]->post_len = 0.0;
         path[i++] = pc = trns[t]; // put this transition into the path
      }
      t++;
   }
   delete [] trns;   // free memory

   // Now that the path is finished, make sure the passing path is good
   for (j = 0; j < num_comps; j++) {
      SB3Vector p;
      SB3Line   l;

      if (path[j]->pass == NULL)
         path[j]->pre_len = 0.0;
      else {
         path[j]->pass[0]->getPoint(p, l, path[j]->pass[0]->Sdst);
         path[j]->pre_len = path[j]->get_path_dist(p);
      
         k = get_NextComponent(j);
         path[j]->pass[1]->getPoint(p, l, path[j]->pass[1]->Fdst);
         path[k]->post_len = path[k]->len - path[k]->get_path_dist(p);
      }
   }
}

void SB3TrkData::saveData(SB3Turn **crn, int num_turns) const
{
   FILE   *out;
   int     i, nt;
   double  rotate;
   SB3Line *start;

   for (i = 0, nt = 0; i < num_turns; i++) {
      if (crn[i] != NULL)
         nt++;
   }

   out = fopen("SB3path_out.txt", "w");
   if (!out)
      return;

   fprintf(out, "v3\n%s\n%d\n", name, nt);

   start  = (SB3Line *)seg[0]->left_side;
   rotate = -start->dir.ang();
   for (i = 0; i < num_turns; i++) {
      if (crn[i] != NULL)
         crn[i]->save(out, seg, rotate);
   }
   fclose(out);
}

int SB3TrkData::loadData(SB3Turn ***p_crn, int *p_num_turns)
{
   const char *ws = " \t\n\r";
   int      i, len, found_trk;
   char     buf[128];
   double   rotate;
   FILE    *in;
   SB3Turn **crn;
   SB3Line  *start;

   in = fopen(DATA_FILE, "r");
   if (!in)
      goto SB3TD_LOAD_EXIT_0;

   len       = 0;
   found_trk = 0;

   if (!fgets(buf, 128, in))
      goto SB3TD_LOAD_EXIT_1;
   if (strncmp(buf, "v3", 2))
      goto SB3TD_LOAD_EXIT_1;

   while (!found_trk) {
      if (!fgets(buf, 128, in))
         goto SB3TD_LOAD_EXIT_1;
      found_trk = (!strncmp(buf, name, strlen(name)));

      if (!fgets(buf, 128, in))
         goto SB3TD_LOAD_EXIT_1;
      len = atoi(strtok(buf, ws));
      if (found_trk)
         break;
      len *= SB3Turn::DATALEN;
      for (i = 0; i < len; i++) {
         if (!fgets(buf, 128, in))
            goto SB3TD_LOAD_EXIT_1;
      }
   }

   start  = (SB3Line *)seg[0]->left_side;
   rotate = start->dir.ang();
   crn    = new SB3Turn*[len];
   for (i = 0; i < len; i++) {
      crn[i] = new SB3Turn();
      if (!crn[i]->load(in, seg, num_segs, rotate))
         goto SB3TD_LOAD_EXIT_2;
   }
   fclose(in);
   *p_crn       = crn;
   *p_num_turns = len;
   return 1;

SB3TD_LOAD_EXIT_2:
   do {
      delete crn[i--];
   } while (i >= 0);
   delete [] crn;
SB3TD_LOAD_EXIT_1:
   fclose(in);
SB3TD_LOAD_EXIT_0:
   return 0;
}

int SB3TrkData::get_PassComponent(double trk_dist, int pc) const
{
   int              i;
   SB3PathComponent *comp;

   if (trk_dist > trk_len)
      trk_dist -= trk_len;
   else if (trk_dist < 0.0)
      trk_dist += trk_len;

   for (i = 0; i < 2; i++) {
      comp = path[pc]->pass[i];
      if ( (comp->is_after(trk_dist, *comp, comp->Sdst))&&
           (comp->is_after(comp->Fdst, *comp, trk_dist)) )
         return i;
   }
   return -1;
}

int SB3TrkData::get_NextComponent(int pc) const
{
   int    i;
   double end_pass = path[pc]->pass[1]->Fdst;

   i = pc;
   do {
      i = (i + 1) % num_comps;
   } while (!path[i]->is_after(end_pass, *(path[i]), path[i]->Sdst));
   return i;
}
//-----END TrkData.cpp-----//

//-----START SB3Driver.cpp-----//
// GLOBALS
static int       trk_num = -1;
static SB3TrkData trk_data;

// CONSTRUCTORS
SBv3r0::SBv3r0(int i)
   : cV(), nV()
{
#ifndef SB3_DLL
   if ( (0 <= i)&&(i < SB3_CARS) )
      m_sName = SB3_NAME[i+1];
   else
      m_sName = SB3_NAME[0];

   m_sAuthor       = SB3_AUTHOR;
   m_sDescription  = SB3_DESCRIPTION;
   m_iNoseColor    = SB3_NOSE;
   m_iTailColor    = SB3_TAIL;
   m_sBitmapName2D = SB3_BITMAP2D;
   m_sModel3D      = SB3_MODEL3D;
#endif
   init         = 2;
   start_fuel   = 0.0;
   fuel_per_lap = 0.0;
   pc           = 0;
   pass_flag    = 0;
   car2pass     = MAX_CARS;
}

// DESTRUCTOR
SBv3r0::~SBv3r0()
{
   if (trk_num >= 0) {
      // free memory used by the track and path data
      trk_data.~SB3TrkData();
      trk_num = -1;
   }
}

// METHODS
con_vec SBv3r0::drive(situation &s)
{
#ifndef SB3_DLL
   if ( (args.m_iCurrentTrack != trk_num)&&(currentTrack) ) {
      trk_num = args.m_iCurrentTrack;
      switch (args.m_iSurface) {
#else
   if (!args)
      return cp;
   else if ( (args->m_iCurrentTrack != trk_num)&&(currentTrack) ) {
      trk_num = args->m_iCurrentTrack;
      switch (args->m_iSurface) {
#endif
      case 0:
         MAX_ACCEL = MYU_MAX0;
         break;
      case 1:
         MAX_ACCEL = MYU_MAX1;
         break;
      case 2:
         MAX_ACCEL = MYU_MAX2;
      }
      MAX_ACCEL *= g;
      trk_data.set_path(currentTrack);
   }
   else if (!currentTrack)
      return cp;

   if (s.starting) {
      init             = (init < 1) ? 1 : init;
      s.side_vision    = 1;
      cp.request_pit   = 0;
      cp.repair_amount = 0;
      fuel_per_lap     = SB3_FC * currentTrack->length;
      cp.fuel_amount   = SB3_fuel_needed(s.stage, s.laps_to_go, 
                                         fuel_per_lap, 0.0);
      start_fuel       = 0.0;
      if ( (s.stage == RACING)&&(s.position > 3) ) {
         // use passing path for opening lap
         pc        = -2;
         pass_flag =  1;
      }
   }

   if (init == 2) {
      init--;
      cp.vc    = SB3_MAX_SPD;
      cp.alpha = 0.0;
   }
   else
      get_RARS_ctrl_params(s);
   return cp;
}

// PRIVATE METHODS
void SBv3r0::get_RARS_ctrl_params(situation &s)
{
   int    type;
   double trk_dist, tgt_v, acc;
   double path_dist, drag_accel;

   if ( (s.laps_to_go > 1)&&(!s.starting) ) {
      // Handle fuel
      if (!cp.request_pit) {
         if (s.lap_flag) {
            if (start_fuel > s.fuel) {
               fuel_per_lap *= 0.5;
               fuel_per_lap += 0.5*(start_fuel - s.fuel);
            }
            start_fuel = s.fuel;
         }
         cp.fuel_amount  = currentTrack->pit_entry - s.distance;
         cp.fuel_amount /= currentTrack->length;
         if (cp.fuel_amount <= 0.0)
            cp.fuel_amount += 1.0;
         cp.fuel_amount *= fuel_per_lap;  // fuel needed to make it to pits
         if ( ((s.fuel-fuel_per_lap) < cp.fuel_amount)||(s.damage > 20000) ) {
            cp.request_pit = 1;  // need to pit
            cp.fuel_amount = SB3_fuel_needed(s.stage, (s.laps_to_go+1), 
                                             fuel_per_lap,
                                             (s.fuel-cp.fuel_amount));
         }
      }
      else if (s.out_pits) {
         cp.request_pit   = 0;
         cp.repair_amount = 0;
         start_fuel       = 0.0;
      }
      // Handle damage
      if ( (s.damage > 1000)&&(cp.request_pit) ){
         if (s.damage > 20000)
            cp.repair_amount = 12000;
         else {
            if (cp.fuel_amount > 95.0)
               cp.repair_amount = s.damage;
            else {
               cp.repair_amount = (int)((100.0-cp.fuel_amount) * 150.0);
               if ((int)s.damage > cp.repair_amount)
                  cp.repair_amount = (int)s.damage - cp.repair_amount;
               else
                  cp.repair_amount = 0; // don't repair
            }
         }
      }
   }
   else
      cp.request_pit = 0;  // no pit

   // get current track/path position and velocity
   trk_dist  = trk_data.get_trk_dist(s.seg_ID, s.to_end);
   pc        = trk_data.get_PathComponent(trk_dist, pass_flag, pc);
   trk_data.get_trk_pos_vel(cV, s.to_lft, s.v, s.vn, trk_dist, s.seg_ID);
   locate_car(NULL, &path_dist, cV.pt, trk_dist, -1);

   // target velocity
   drag_accel = SB3_drag_accel(s.fuel, s.damage);
   tgt_v      = trk_data.get_tgt_spd(s.v, drag_accel, path_dist, 
                                     trk_dist, pass_flag, pc);
   acc        = trk_data.pit_entry - s.distance;
   if ( (cp.request_pit)&&(acc > 0.0) ) {
      if ( ((SB3_brkdist(s.v, trk_data.pit_spd, drag_accel) > acc)&&
            (tgt_v > trk_data.pit_spd))||
           ((acc < 100.0)&&(tgt_v < trk_data.pit_spd)) )
         tgt_v = trk_data.pit_spd;
   }

   // avoid other cars
   car2pass = MAX_CARS;
   tgt_v    = anti_collision(s, tgt_v);
   if ( (!pass_flag)&&(car2pass < MAX_CARS) ) {
      // try to use the passing path
      type = pc;
      pc   = trk_data.get_PathComponent(trk_dist, 1, (pc-1));
      if ( (trk_data.can_pass(trk_dist, pc))&&(pass_clear(s, trk_dist)) ) {
         pass_flag = 1;
         locate_car(NULL, &path_dist, cV.pt, trk_dist, -1);

         // target velocity
         acc = trk_data.get_tgt_spd(s.v, drag_accel, path_dist, 
                                    trk_dist, pass_flag, pc);
         if (acc < (tgt_v-5.0)) {
            // stay on racing path
            pass_flag = 0;
            pc        = type;
         }
      }
      else
         pc = type;
   }

   // limit acceleration
   if (s.v > 100.0) {
      acc = (tgt_v - s.v) / delta_time;
      if (acc > (5.0*MAX_ACCEL))
         tgt_v = s.v + (5.0 * MAX_ACCEL * delta_time);
      else if (acc < (-5.0*MAX_ACCEL))
         tgt_v = s.v - (5.0 * MAX_ACCEL * delta_time);
   }
   else if (s.v > 0.0) {
      acc = tgt_v / s.v;
      if (acc > 1.25)
         tgt_v = 1.25 * s.v;
      else if (acc < 0.85)
         tgt_v = 0.85 * s.v;
   }

   // find target position
   if (init == 1) {
      if ( (pc != 1)&&(s.seg_ID != 1) )
         type = SB3TrkData::START;
      else
         init--;
   }
   else if ( ((s.to_lft < 0.0)||(s.to_rgt < 0.0))&&
             (!s.out_pits) )
      type = SB3TrkData::OFF;
   else
      type = SB3TrkData::NORMAL;
   nV.x      = tgt_v;
   nV.y      = s.to_lft;
   pass_flag = trk_data.get_nex_pos(cV, nV, type, trk_dist, pass_flag, pc);

   // get new control parameters
   update_con_vec(s, (s.to_lft >= SB3_IN_WID)&&(s.to_rgt >= SB3_IN_WID));
}

void SBv3r0::update_con_vec(situation &s, int damp)
{
   SB3Vector dV, L, W;
   double   acc;

   nV.rotate(-cV.dir.ang());
   acc      = cV.dir.mag();
   cV.dir.x = acc;
   cV.dir.y = 0.0;

   dV.add(nV);
   dV.sub(cV.dir);

   L.add(dV);
   L.normalize();
   acc = (dV.mag() / delta_time) / g;
#ifndef SB3_DLL
   L.scale(-SB3_iu(acc, args.m_iSurface));
#else
   L.scale(-SB3_iu(acc, args->m_iSurface));
#endif

   W.add(L);
   W.sub(cV.dir);

   cp.alpha = W.ang() - MY_PI;
   cp.vc    = W.mag();
   if ( (damp)&&(!pass_flag) ) {
#ifndef SB3_DLL
      switch (args.m_iSurface) {
#else
      switch (args->m_iSurface) {
#endif
      case 0:
         cp.alpha *= 0.9;
         cp.alpha += 0.1*s.alpha;
         break;
      case 1:
         cp.alpha *= 0.85;
         cp.alpha += 0.15*s.alpha;
         break;
      case 2:
         cp.alpha *= 0.8;
         cp.alpha += 0.2*s.alpha;
      };
   }
   else {
      cp.alpha *= 0.925;
      cp.alpha += 0.075*s.alpha;
   }
   if ( (cp.vc > s.vc)&&(s.to_lft > 0.0)&&(s.to_rgt > 0.0) ) {
      cp.vc *= 0.75;
      cp.vc += 0.25*s.vc;
   }
}

void SBv3r0::locate_car(double *dist2path, double *path_dist, 
                         const SB3Vector &pt, double to_lft, int seg_ID)
{
   int    car_pc;
   double trk_dist;

   if (seg_ID >= 0) {
      trk_dist = trk_data.get_trk_dist(pt, to_lft, seg_ID);
      car_pc   = trk_data.get_PathComponent(trk_dist, pass_flag, pc);
   }
   else {
      trk_dist = to_lft;
      car_pc   = pc;
   }
   trk_data.locate_car(dist2path, path_dist, pt, pc, car_pc,
                       trk_dist, pass_flag);
}

double SBv3r0::anti_collision(situation &s, double tgt_v)
{
   int      j, left;
   double   min_v, v, t, rx, ry, lx, ly, max_x, min_x;
   SB3Vector pt;
   SB3Line   oc, tc;

   tc.pt.y  = 0.5*CARLEN;
   min_v    = tgt_v;
   for (j = 0; j < NEARBY_CARS; j++) {
      if (s.nearby[j].who >= MAX_CARS)
         continue;

      // other car's relative position
      oc.pt.x  = s.nearby[j].rel_x;
      oc.pt.y  = s.nearby[j].rel_y;
      oc.pt.rotate(-s.alpha);
      if (oc.pt.y < 1.0)
         continue;
      else
         left = (oc.pt.x < 0.0);

      // other car's relative velocity
      oc.dir.x = s.nearby[j].rel_xdot;
      oc.dir.y = s.nearby[j].rel_ydot;
      v        = oc.dir.mag();
      lx       = fabs(oc.pt.x) / oc.pt.mag();
      rx       = fabs(s.nearby[j].rel_xdot) / v;
      ry       = ((0.875/0.2) * lx) + 1.0 - (0.875/0.2);
      ly       = ((20.0/0.2) * lx) + 25.0 - (20.0/0.2);
      if ( ( (v < ly)&&(rx < ry)&&(lx > 0.8) )||
           (oc.pt.dot(oc.dir) > 0.0) )
         continue;
      oc.dir.rotate(-s.alpha);
      oc.dir.scale(1.0/v);
      oc.movePoint(0.5*CARLEN);
      rx = oc.pt.x + (0.5*CARWID*oc.dir.y);  // right side
      ry = oc.pt.y - (0.5*CARWID*oc.dir.x);
      lx = oc.pt.x - (0.5*CARWID*oc.dir.y);  // left side
      ly = oc.pt.y + (0.5*CARWID*oc.dir.x);

      if ( (oc.pt.y > CARLEN)&&
           ( (s.nearby[j].to_lft < 0.0)||(s.nearby[j].to_rgt < 0.0) ) )
         v *= 1.2;
      else if ( (v < 15.0)&&(oc.pt.mag() < (2.5*SB3_MARGIN)) )
         v *= ((0.035/SB3_MARGIN) * oc.pt.mag()) + 0.9125;

      // look for frontal impact
      if (oc.pt.y > tc.pt.y) {
         t     = oc.dir.x * delta_time;
         max_x = 0.5;
         if (s.v < 111.1111)
            max_x *= 0.009*s.v;
         max_x *= (v*v) / (SB3_CLOSE_RATE*SB3_CLOSE_RATE);
         max_x += 0.5;
         max_x *= SB3_MARGIN;
         min_x  = t - max_x;
         max_x += t;
         if (min_x > (-0.5*CARWID))
            min_x = -0.5*CARWID;
         if (max_x < ( 0.5*CARWID))
            max_x =  0.5*CARWID;

         tc.dir.x = 1.0;  // front
         tc.dir.y = 0.0;
         oc.pt.x  = rx;
         oc.pt.y  = ry;
         if ( (tc.intersect(pt, oc) > 0)&&(min_x < pt.x)&&(pt.x < max_x) )
            goto SB3_COLLISION_DETECTED;  // will hit front of this car
         oc.pt.x  = lx;
         oc.pt.y  = ly;
         if ( (tc.intersect(pt, oc) > 0)&&(min_x < pt.x)&&(pt.x < max_x) )
            goto SB3_COLLISION_DETECTED;  // will hit front of this car
      }
      
      // look for side impact
      if ( (oc.pt.y < (3.5*CARLEN))&&
           ( ((left)&&(s.cen_a > 5.0))||((!left)&&(s.cen_a < -5.0)) ) ) {
         max_x  = s.vn * s.vn;
         max_x /= 0.16 * SB3_CLOSE_RATE * SB3_CLOSE_RATE;
         if (max_x > 3.5)
            max_x = 3.5;
         min_x  = -0.5 * max_x;
         min_x -=  1.0;
         max_x +=  1.0;
         min_x *= SB3_MARGIN;
         max_x *= SB3_MARGIN;
      }
      else {
         min_x = -SB3_MARGIN;
         max_x  = SB3_MARGIN;
      }
      if (left) {
         tc.pt.x  = -0.5*CARWID; // left front corner
         tc.dir.x = 0.0;         // left side
         tc.dir.y = 1.0;
         oc.pt.x  = rx;
         oc.pt.y  = ry;
         if ( (tc.intersect(pt, oc) > 0)&&(min_x < pt.y)&&(pt.y < max_x) )
            goto SB3_COLLISION_DETECTED;  // will hit left side of this car
         oc.pt.x  = lx;
         oc.pt.y  = ly;
         if ( (tc.intersect(pt, oc) <= 0)||
              (min_x >= pt.y)||(pt.y >= max_x) )
            continue;                    // no collision detected
      }
      else {
         tc.pt.x  =  0.5*CARWID; // right front corner
         tc.dir.x = 0.0;         // right side
         tc.dir.y = 1.0;
         oc.pt.x  = rx;
         oc.pt.y  = ry;
         if ( (tc.intersect(pt, oc) > 0)&&(min_x < pt.y)&&(pt.y < max_x) )
            goto SB3_COLLISION_DETECTED;  // will hit right side of this car
         oc.pt.x  = lx;
         oc.pt.y  = ly;
         if ( (tc.intersect(pt, oc) <= 0)||
              (min_x >= pt.y)||(pt.y >= max_x) )
            continue;                    // no collision detected
      }
SB3_COLLISION_DETECTED:
      if ( (car2pass >= MAX_CARS)&&
           (fabs(oc.pt.x) < CARLEN)&&((ly+ry) < (3.5*CARLEN)) )
         car2pass = j;  // found a car to pass

      // slow down to avoid collision
      t = pt.dist2Point(oc.pt) / v;
      v = -v;
      if ( (oc.pt.y < CARLEN)&&(tc.dir.x == 0.0) ) {
         if (s.nearby[j].braking)
            v -= 0.6 * MAX_ACCEL * delta_time;
         else
            v += 0.2 * MAX_ACCEL * delta_time;
      }
      else if ( (tc.dir.y == 0.0)&&(s.nearby[j].braking) )
         v -= 0.3 * MAX_ACCEL * delta_time;
      else if (!s.nearby[j].braking)
         v += 0.4 * MAX_ACCEL * delta_time;

      if (t > (1.5*delta_time))
         v *= delta_time / (t - (0.5*delta_time));
      v += s.v;
      if (v < min_v) {
         if (v <= 2.0)
            return 2.0;
         min_v = v;
      }
   }
   return min_v;
}

int SBv3r0::pass_clear(situation &s, double trk_dist)
{
   int    j;
   double dist_to_path;

   locate_car(&dist_to_path, NULL, cV.pt, trk_dist, -1);
   if (dist_to_path > CARWID) {
      dist_to_path += CARWID;
      // pass on the left side of track
      for (j = 0; j < NEARBY_CARS; j++) {
         if ( (s.nearby[j].who < MAX_CARS)&&
              (s.nearby[j].rel_y > (-0.5*CARLEN))&&
              (s.nearby[j].rel_y < ( 4.0*CARLEN))&&
              (s.nearby[j].rel_x > -dist_to_path) ) {
            if (s.nearby[j].rel_y > (CARLEN+1.0)) {
               if (s.nearby[j].rel_x < -CARWID)
                  return 0;
            }
            else if (s.nearby[j].rel_x < (-0.25*CARWID))
               return 0;
         }
      }
   }
   else if (dist_to_path < -CARWID) {
      dist_to_path -= CARWID;
      // pass on the right side of track
      for (j = 0; j < NEARBY_CARS; j++) {
         if ( (s.nearby[j].who < MAX_CARS)&&
              (s.nearby[j].rel_y > (-0.5*CARLEN))&&
              (s.nearby[j].rel_y < ( 4.0*CARLEN))&&
              (s.nearby[j].rel_x < -dist_to_path) ) {
            if (s.nearby[j].rel_y > (CARLEN+1.0)) {
               if (s.nearby[j].rel_x > CARWID)
                  return 0;
            }
            else if (s.nearby[j].rel_x > (0.25*CARWID))
               return 0;
         }
      }
   }
   return 1;   // passing path is clear
}
//-----END SBDriver.cpp-----//


// STATIC FUNCTIONS
//-----START spd_func.cpp-----//
double SB3_crn_spd(double radius)
{
   double retval;

   if (radius == 0.0)
      return SB3_MAX_SPD;

   retval = sqrt(radius * MAX_ACCEL);
   if (retval > SB3_MAX_SPD)
      retval = SB3_MAX_SPD;

   return retval;
}

double SB3_brkdist(double vi, double vf, double drag_accel)
{
   double scale;

   if (vf >= vi)
      return 0.0;

#ifndef SB3_DLL
   switch (args.m_iSurface) {
#else
   switch (args->m_iSurface) {
#endif
   case 0:
      scale = 1.0 + ((vi - vf)*0.0020);
      break;
   case 1:
      scale = 1.0 + ((vi - vf)*0.0018);
      break;
   case 2:
      scale = 1.0 + ((vi - vf)*0.0015);
   };

   vi *= vi * drag_accel;
   vi += MAX_ACCEL;
   vf *= vf * drag_accel;
   vf += MAX_ACCEL;
   return (scale * log(vf/vi) / (-2.0*drag_accel));
}

// inverse friction function
double SB3_iu(double cur_u, int surfaceType)
{
   double retval;

   switch (surfaceType) {
   case 2:
      if (cur_u >= MYU_MAX2)
         retval = 20.0;
      else
         retval = -SLIPPING * log(1.0 - (cur_u/MYU_MAX2));
      break;
   case 1:
      if (cur_u >= MYU_MAX1)
         retval = 20.0;
      else
         retval = -SLIPPING * log(1.0 - (cur_u/MYU_MAX1));
      break;
   default:
      if (cur_u >= MYU_MAX0)
         retval = 20.0;
      else
         retval = ((SLIPPING*cur_u)/MYU_MAX0) / (1.0 - (cur_u/MYU_MAX0));
   }
   return retval;
}

double SB3_fuel_needed(Stage s, int lapsleft, double fpl, double cur_fuel)
{
   double retval;

   if (fpl <= 0.0)
      return MAX_FUEL;
   else if (s == PRACTICE)
      retval = fpl * 10.0;
   else if (s == QUALIFYING)
      retval = fpl * 2.5;
   else
      retval = fpl * (double)lapsleft;

   retval -= cur_fuel;
   if (retval < 0.0)
      return 0.0;
   else if (retval > MAX_FUEL)
      return MAX_FUEL;
   return retval;
}
//-----END spd_func.cpp-----//

Driver * getSBv3r0Instance()
{
   static int i = -SB3_USE_DEFAULT_NAME;
   
   return new SBv3r0(i++);
}
#undef SB3_car_mass
#undef SB3_drag
#undef SB3_drag_accel
#undef SB3_CARS
#undef SB3_USE_DEFAULT_NAME
